/* Le calibrage des valeurs optimales à affecter aux constantes de détection des seils indiqués
   en Fig.22 et Fig.23 de la page 8 du didacticiel seront plus faciles à déterminer avec ce
   petit outil logiciel qui n'est développé que dans ce but.
   Son utilisation est élémentaire :
   • Chaque fois que l'on clique sur un BP sa valeur numérisée est affichée sur l'écran OLED.
   • Chaque fois que l'on tourne le codeur rotatif se sont les vaurs à vide qui sont indiquées.
   
Branchements à effectuer :
--------------------------
 A0 : Ligne à trois touches du clavier.
 A1 : Ligne à deux touches du clavier. Le schéma du clavier est donné en Fig.19 de la page 8 du didacticiel.
 
 A2 : Bouton central du codeur rotatif.
 A4 : SDA OLED.
 A5 : SCL OLED.
 Codeur rotatif B : Broche D2. >>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
 Codeur rotatif A : Broche D3. >>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
                               >>>>>>>>>>>>>>>>>> obtenir un sens de variation correct.
 D4 : BUZZER actif.
 D5 : Anode LED rouge.
 D6 : Anode LED verte.
 D7 : Anode LED bleue.
                                                   PGM               RAM dynamique
Codeur rotatif + BUZZER BP central :         10348 Octets (33%)     460 Octets (22%)    */

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)

//------------- Définition des broches d'entrées analogiques ----------------
#define Ligne_trois_touches 14 // A0 pour lire le clavier à 3 BP. 
#define Ligne_deux_touches 15  // A1 pour lire le clavier à 2 BP. 
#define BP_central 16          // A2 pour lire le bouton central du codeur rotatif.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 3    // D2 : DT
#define Codeur_B 2    // D3 : CLK 
#define BUZZER 4      // D4 : BUZZER de type actif ou passif.
#define ROUGE 5       // Anode LED ROUGE.
#define VERT 6        // Anode LED Verte.
#define BLEU 7        // Anode LED bleue.

//------------------- Déclaration des constante du programme ----------------
#define NB_tests_antirebonds 30 // Nombre de tests de stabilisation B.P.

//----------------------- Déclaration des variables -------------------------
boolean BP_central_a_traiter;
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.

byte Num_BP;

int U_sur_A0, U_sur_A1;

volatile boolean Rotation;  // Indique une activation du codeur rotatif.

//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  pinMode(ROUGE, OUTPUT); pinMode(VERT, OUTPUT); pinMode(BLEU, OUTPUT); Eteindre_les_LEDs();
  BP_central_a_traiter = false; digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
  Front_sur_A(); Front_sur_B();}
  
void loop() {
  //---------------- Traite le B.P. central du codeur incrémental -------------
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
  //-------- Attendre le relacher du B.P. central. --------
  Rouge(); BIP(); while (analogRead(BP_central) < 512); // Attendre le relâché.
  Eteindre_les_LEDs();}
  //---------------------------------------------------------------------------
  Traite_le_clavier();
  //---------------------------------------------------------------------------
  if (Rotation) {TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL();}}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
  void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
  void Front_sur_A(){ // Interruption quand A change d'état avant B.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
    if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
      A_set = !A_set;
      if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
  void Front_sur_B(){ // Interruption quand B change d'état avant A.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
    if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
      B_set = !B_set;
      // Décrémenter le compteur si B avant A.
      if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
  //--------------------------------------------------------------------------------------------------

void Delay_60() {delay(60);}

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}
        
void Rouge() {digitalWrite(ROUGE, HIGH);}
void Verte() {digitalWrite(VERT, HIGH);}
void Bleue() {digitalWrite(BLEU, HIGH);}
void Eteindre_les_LEDs() {digitalWrite(ROUGE, LOW); digitalWrite(VERT, LOW); digitalWrite(BLEU, LOW);}

void ATTENDRE_le_relacher() {
  byte COMPTEUR = 0; digitalWrite(VERT, HIGH); Delay_60();// Attendre_la_stabilisation du travail.
  digitalWrite(VERT, HIGH);
  while (COMPTEUR < NB_tests_antirebonds) {// Attendre_la_stabilisation du repos :
        if ((analogRead(Ligne_trois_touches) < 813) || (analogRead(Ligne_deux_touches) < 776)) COMPTEUR = 0;
        else COMPTEUR++;}
  digitalWrite(VERT, LOW);}

void Tester_le_Clavier() {// Les valeurs tiennent compte de VCC = 5V car fourni par la ligne USB.
  U_sur_A1 = 0; Num_BP = 0; // Signale aucun B.P. utilisé.
  U_sur_A0 = analogRead(Ligne_trois_touches);
  if (U_sur_A0 < 813) {U_sur_A0 = 0; Delay_60(); 
      for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_A0 = U_sur_A0 + analogRead(Ligne_trois_touches);
      U_sur_A0 = U_sur_A0 / NB_tests_antirebonds;
      if (U_sur_A0 < 813) {Num_BP = 1;
      if (U_sur_A0 < 478)  Num_BP = 2; if (U_sur_A0 < 223)  Num_BP = 3;}}
  U_sur_A1 = analogRead(Ligne_deux_touches);
  if (U_sur_A1 < 776) {U_sur_A1 = 0; Delay_60();
      for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_A1 = U_sur_A1 + analogRead(Ligne_deux_touches);
      U_sur_A1 = U_sur_A1 / NB_tests_antirebonds;
      if (U_sur_A1 < 776) {Num_BP = 4; if (U_sur_A1 < 302)  Num_BP = 5;}}
  if (Num_BP != 0) ATTENDRE_le_relacher();}

void Traite_le_clavier() {
  Tester_le_Clavier();
  if (Num_BP > 0) {
     u8g.firstPage(); do {
     u8g.setPrintPos(5,38); u8g.print("B.P. "); u8g.print(Num_BP); u8g.print(" Valeur = ");
     if (Num_BP > 3) u8g.print(U_sur_A1); else u8g.print(U_sur_A0);}
     while(u8g.nextPage());}}
     
void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL() {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  U_sur_A0 = analogRead(Ligne_trois_touches); U_sur_A1 = analogRead(Ligne_deux_touches);
  u8g.firstPage(); do {  
      u8g.setPrintPos(5,18); u8g.print("CAN 3 BP = "); u8g.print(U_sur_A0);
      u8g.setPrintPos(5,58); u8g.print("CAN 2 BP = "); u8g.print(U_sur_A1);}
  while(u8g.nextPage());} 

 
 
 
