/* Programme qui met en place le MENU de BASE et les aiguillages pour les items.
   Si Clic central valide l'item indexé.
   Tant que le B.P. central est appuyé, la LED d'arduino s'illumine.
   Le sous_menu activé ouvre un écran noir avec sa nature et génère un BIP sonore.
   Cliquer à nouveau sur le bouton ramène au MENU de BASE.
   Le Menu BARILLET est en place. Valider chaque Item engendre un BIP sonore et
   ramène au MENU de BASE.
   
   >>>> Lire les ommentaires sur ce programme en fin de listage.

Branchements à effectuer :
--------------------------
 A2 : Bouton central du codeur rotatif.
 A4 : SDA OLED.
 A5 : SCL OLED.
 Codeur rotatif B : Broche D2. >>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
 Codeur rotatif A : Broche D3. >>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
                               >>>>>>>>>>>>>>>>>> obtenir un sens de variation correct.
 D4 : BUZZER actif.
                                                   PGM               RAM dynamique
Codeur rotatif + BUZZER BP central :          9682 Octets (31%)     376 Octets (18%) */

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)
//------------- Définition des broches d'entrées analogiques ----------------
#define BP_central 16 // A2 pour lire le bouton central.

//------------- Définition des broches de sorties analogiques ---------------
#define LED_Clavier 13 // LED rouge d'Arduino.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 3    // D2 : DT  >>>>>>>>>>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
#define Codeur_B 2    // D3 : CLK >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
#define BUZZER 4      // D4 : BUZZER actif.

//----------------------- Déclaration des variables -------------------------
boolean BP_central_a_traiter;
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.

boolean FINI;

byte Position_Index;

volatile boolean Rotation;  // Indique une activation du codeur rotatif.

//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  BP_central_a_traiter = false; digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
  Position_Index = 4; Front_sur_A(); Front_sur_B();} // Voir informations en fin de listage.

void loop() {MENU_de_BASE();}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
  void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
  void Front_sur_A(){ // Interruption quand A change d'état avant B.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
    if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
      A_set = !A_set;
      if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
  void Front_sur_B(){ // Interruption quand B change d'état avant A.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
    if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
      B_set = !B_set;
      // Décrémenter le compteur si B avant A.
      if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
  //--------------------------------------------------------------------------------------------------
void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

void Positionne_INDEX(byte X, byte Y) {u8g.setPrintPos(X,Y); u8g.print('>');}

void Affiche_INDEX_10() {
  byte X, Y;
  X = 52; if (Position_Index > 4) X = 117;
  Y = 11 + Position_Index * 12; if (Position_Index > 4) Y = 11 + (Position_Index-5) * 12;
  u8g.setPrintPos(X,Y); u8g.print('<');}

void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(byte MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if (Sens_positif) {Position_Index++; if (Position_Index == (MAX+1)) Position_Index = 0;}
  else {if (Position_Index > 0) Position_Index--; else Position_Index = MAX;}}

void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     digitalWrite(LED_Clavier, HIGH); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     digitalWrite(LED_Clavier, LOW);
     // ===========================================================
     BP_central_a_traiter = true;}}

void Traite_BP_central() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {FINI = true;
  BP_central_a_traiter = false; Position_Index = 0;}}

void MENU_du_RUN() {
  u8g.firstPage(); do {u8g.setPrintPos(40, 11); u8g.print("RUN");} while(u8g.nextPage());
  BIP(); Position_Index = 0;
  while (!FINI) {Traite_BP_central();
         if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}}
  
void MENU_BARILLET() {
  Position_Index = 0;
  while (!FINI) {
  //------------------ Afficher le menu du BARILLET : ------------------
  u8g.firstPage();
  do {u8g.setPrintPos(6, 11); u8g.print("Affiche    PLATEAU");
      u8g.setPrintPos(18, 23); u8g.print("SAUVE   Recharge");
      u8g.setPrintPos(6, 35); u8g.print("Arrange   Init.BIT");
      u8g.setPrintPos(6, 48); u8g.print("Origine   Pos.TETE");
      u8g.setPrintPos(0, 60); u8g.print("Tourne G   Tourne D"); Affiche_INDEX_10();}
      while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) {TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9); Affiche_INDEX_10;}
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {BP_central_a_traiter = false;
  switch (Position_Index) {
     case 0 : {BIP(); FINI = true; break;}
     case 1 : {BIP(); FINI = true; break;}
     case 2 : {BIP(); FINI = true; break;}
     case 3 : {BIP(); FINI = true; break;}
     case 4 : {BIP(); FINI = true; break;}
     case 5 : {BIP(); FINI = true; break;}
     case 6 : {BIP(); FINI = true; break;}
     case 7 : {BIP(); FINI = true; break;}
     case 8 : {BIP(); FINI = true; break;}
     case 9 : {BIP(); FINI = true; break;}}}}
  Position_Index = 0;}
   
void MENU_EEPROM() {
  u8g.firstPage();
  do {u8g.setPrintPos(40,35); u8g.print("EEPROM");} while(u8g.nextPage());
  BIP(); Position_Index = 0;
  while (!FINI) {Traite_BP_central();
         if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}}

void MENU_des_OPTIONS() {
  u8g.firstPage();
  do {u8g.setPrintPos(40,48); u8g.print("OPTIONS");} while(u8g.nextPage());
  BIP(); Position_Index = 0;
  while (!FINI) {Traite_BP_central();
        if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}}

void MENU_PROGRAMME() {
  u8g.firstPage();
  do {u8g.setPrintPos(40,60); u8g.print("PROGRAMME");} while(u8g.nextPage());
  BIP(); Position_Index = 0;
  while (!FINI) {Traite_BP_central();
        if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}}

void MENU_de_BASE() {
  byte X,Y; FINI = false;
  //-------------------- Afficher le MENU de BASE : --------------------
  u8g.firstPage(); do {
    u8g.drawRFrame(0,0,128,64,10);
    u8g.setPrintPos(40,11); u8g.print("RUN");
    u8g.setPrintPos(40,23); u8g.print("BARILLET");
    u8g.setPrintPos(40,35); u8g.print("EEPROM");
    u8g.setPrintPos(40,48); u8g.print("OPTIONS");
    u8g.setPrintPos(40,60); u8g.print("PROGRAMME");
    Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {
     BP_central_a_traiter = false; FINI = false;
  switch (Position_Index) {
     case 0 : {MENU_du_RUN(); break;}
     case 1 : {MENU_BARILLET(); break;}
     case 2 : {MENU_EEPROM(); break;}
     case 3 : {MENU_des_OPTIONS(); break;}
     case 4 : MENU_PROGRAMME();}}}
     
/* Problèmes liés au codeur rotatif :
   ----------------------------------
   Ce type de codeur engendre une rotation fantôme positive et négative à sa première utilisation.
   Pour parer cette difficulté, les instructions suivantes sont ajoutées dans les initialisations :
   Front_sur_A(); Front_sur_B(); Num_PGM_indexee = 4;
   Cette ligne de code effectue une "rotation positive artificielle". Du coup Num_PGM_indexee prend
   la valeur 0 en tête de listage et le fonctionnement du codeur rotatif est correct. 
  
   En fonction du codeur rotatif approvisionné, il peut s'avérer nécessaire d'inverser D2 et D3
   pour obtenir le sens de déplacement correct.

   Ayant approvisionné ce type de codeur chez divers fournisseurs, les tarifs ne sont pas les mêmes,
   mais la qualité non plus. Les moins onéreux sont victimes de faux contacts qui se traduisent par
   des déplacements aléatoires du curseur. Aussi, un moyen radical d'éliminer ce problème
   particulièrement pénalisant consiste à augmenter "la constante de tems" de l'établissement du
   niveau électrique logique en plaçant un condendateur entre A, B et GND.                         */
  
  
