/* Programme qui visualise le contenu fictif de l'EEPROM qui contiendrait 10 algorithmes au maximum.
   Intègre le codeur rotatif et ajoute l'indexation des dix fichiers potentiels.
   Si Clic du bouton central sur fichier "---" on obtient un BIP d'alerte.
   Si Clic central sur fichier "programme" pas d'action.
   Tant que le B.P. central est appuyé, la LED d'arduino s'illumine.
   
   >>>> Lire les commentaires sur ce programme en fin de listage.

Branchements à effectuer :
--------------------------
 A2 : Bouton central du codeur rotatif.
 A4 : SDA OLED.
 A5 : SCL OLED.
 Codeur rotatif B : Broche D2. >>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
 Codeur rotatif A : Broche D3. >>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
                               >>>>>>>>>>>>>>>>>> obtenir un sens de variation correct.
 D4 : BUZZER actif.
                                                   PGM               RAM dynamique
Codeur rotatif + BUZZER BP central :          8406 Octets (27%)     270 Octets (13%) */

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)
//------------- Définition des broches d'entrées analogiques ----------------
#define BP_central 16 // A2 pour lire le bouton central.

//------------- Définition des broches de sorties analogiques ---------------
#define LED_Clavier 13 // LED rouge d'Arduino.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 3    // D2 : DT  >>>>>>>>>>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
#define Codeur_B 2    // D3 : CLK >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
#define BUZZER 4      // D4 : BUZZER actif.
//----------------------- Déclaration des variables -------------------------
boolean BP_central_a_traiter;
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.
boolean Existe; // L'emplacement simulé contient un algorithme.
volatile boolean Rotation;  // Indique une activation du codeur rotatif.

byte Num_PGM_indexee;

byte Emplacement[] = {0,12,0,45,0,123,0,32,0,255}; // Initialisation de références simulées.
//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  BP_central_a_traiter = false; digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
   Num_PGM_indexee = 9; Front_sur_A(); Front_sur_B();
  Liste_contenu_EEPROM();}

void loop() {
  //======================== Traite le B.P. central ======================
  Teste_activation_BP_central(); if (BP_central_a_traiter) {BP_central_a_traiter = false; Traite_BP_Central();}
  //--------------------- Traite la rotation du codeur incrémental ---------------------
  if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL();}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
  void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
  void Front_sur_A(){ // Interruption quand A change d'état avant B.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
    if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
      A_set = !A_set;
      if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
  void Front_sur_B(){ // Interruption quand B change d'état avant A.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
    if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
      B_set = !B_set;
      // Décrémenter le compteur si B avant A.
      if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
  //--------------------------------------------------------------------------------------------------

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

void Debut_Nom(byte X, byte Y, byte N) {u8g.setPrintPos(X, Y); u8g.print("Pgm"); u8g.print(N);}

void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL() {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  Existe = true;
  if (Sens_positif) {Num_PGM_indexee++; if (Num_PGM_indexee == 10) Num_PGM_indexee = 0;}
  else {if (Num_PGM_indexee > 0) Num_PGM_indexee--; else Num_PGM_indexee = 9;}
  if (Emplacement[Num_PGM_indexee] == 0) Existe = false; Liste_contenu_EEPROM();}

void Valide_le_fichier_indexe() {} // Provisoirement ne fait rien.

void Traite_BP_Central() {// B.P. central activé dans la boucle de base.
     if (Existe) Valide_le_fichier_indexe(); else BIP();}

void Liste_contenu_EEPROM() {
  byte X,Y;
  // Afficher la liste :
  u8g.firstPage(); do {
     Y = 10; X = 0;
     for (byte I=0; I<10; I++) {
          if (I > 4) X = 66; if (I == 5) Y = 10;
          Debut_Nom(X,Y,I); if (Emplacement[I] > 0) {u8g.print(" "); if (Emplacement[I] < 10) u8g.print("0"); 
                               if (Emplacement[I] < 100) u8g.print("0"); u8g.print(Emplacement[I]);}
                            else {u8g.print(" ---");}  Y = Y + 13;
          if (I == Num_PGM_indexee) u8g.print(" <");}} 
  while(u8g.nextPage());}

void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     digitalWrite(LED_Clavier, HIGH); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     digitalWrite(LED_Clavier, LOW);
     // ===========================================================
     BP_central_a_traiter = true;}}

/* Commentaires sur ce programme.
   ==============================

Problèmes liés au codeur rotatif :
----------------------------------
  Ce type de codeur engendre une rotation fantôme positive et négative à sa première utilisation.
  Pour parer cette difficulté, les instructions suivantes sont ajoutées dans les initialisations :
  Num_PGM_indexee = 9; Front_sur_A(); Front_sur_B();
  Cette ligne de code effectue une "rotation positive artificielle". Du coup Num_PGM_indexee prend
  la valeur 0 en tête de listage et le fonctionnement du codeur rotatif est correct. 
  
  En fonction du codeur rotatif approvisionné, il peut s'avérer nécessaire d'inverser D2 et D3
  pour obtenir le sens de déplacement correct.

  Ayant approvisionné ce type de codeur chez divers fournisseurs, les tarifs ne sont pas les mêmes,
  mais la qualité non plus. Les moins onéreux sont victimes de faux contacts qui se traduisent par
  des déplacements aléatoires du curseur. Aussi, un moyen radical d'éliminer ce problème
  particulièrement pénalisant consiste à augmenter "la constante de tems" de l'établissement du
  niveau électrique logique en plaçant un condendateur entre A, B et GND.*/
  
  


