/*    >>>>>> ATTENTION : Avant de téléverser ce croquis il faut au préalable utiliser P00 <<<<<<

Programme strictement identique à P07 auquel on a ajouté les routines de lecture et d'écriture en EEPROM
et la procédure d'affichage de textes puisés en EEPROM.

Branchements à effectuer :
--------------------------
 A2 : Bouton central du codeur rotatif.
 A4 : SDA OLED.
 A5 : SCL OLED.
 Codeur rotatif B : Broche D2. >>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
 Codeur rotatif A : Broche D3. >>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
                               >>>>>>>>>>>>>>>>>> obtenir un sens de variation correct.
 D4 : BUZZER actif.
                                                   PGM               RAM dynamique
Codeur rotatif + BUZZER BP central :         10422 Octets (33%)     262 Octets (12%)    */

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)
#include <EEPROM.h>
//------------- Définition des broches d'entrées analogiques ----------------
#define BP_central 16 // A2 pour lire le bouton central.

//------------- Définition des broches de sorties analogiques ---------------
#define LED_Clavier 13 // LED rouge d'Arduino.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 3    // D2 : DT
#define Codeur_B 2    // D3 : CLK 
#define BUZZER 4      // D4 : BUZZER de type actif ou passif.

//----------------------- Déclaration des variables -------------------------
boolean BP_central_a_traiter;
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.

boolean FINI;

byte Position_Index, Num_PGM_indexee;

byte Emplacement[] = {0,12,0,45,0,123,0,32,0,255}; // Initialisation de références simulées.

volatile boolean Rotation;  // Indique une activation du codeur rotatif.

//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  BP_central_a_traiter = false; digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
  Position_Index = 4; Front_sur_A(); Front_sur_B();} // Voir informations en fin de listage.

void loop() {MENU_de_BASE();}

  //---------------------- Routines de servitude pour l'utilisation de l'EEPROM ----------------------

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {// Routines pour les affichages textes EEPROM.
    for (byte I=0; I<Longueur; I++) {u8g.print(char(EEPROM.read(PTR))); PTR++;}; }

void Afficher_un_texte(byte X, byte Y, int PTR, byte Longueur) {
  u8g.setPrintPos(X, Y); Aff_TEXTE_EEPROM(PTR,Longueur);}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
  void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
  void Front_sur_A(){ // Interruption quand A change d'état avant B.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
    if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
      A_set = !A_set;
      if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
  void Front_sur_B(){ // Interruption quand B change d'état avant A.
    if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
    if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
      B_set = !B_set;
      // Décrémenter le compteur si B avant A.
      if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
  //--------------------------------------------------------------------------------------------------

void Debut_Nom(byte X, byte Y, byte N) {u8g.setPrintPos(X, Y); u8g.print("Pgm"); u8g.print(N);}

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

void Positionne_INDEX(byte X, byte Y) {u8g.setPrintPos(X,Y); u8g.print('>');}

void Affiche_INDEX_10() {
  byte X, Y;
  X = 52; if (Position_Index > 4) X = 117;
  Y = 11 + Position_Index * 12; if (Position_Index > 4) Y = 11 + (Position_Index-5) * 12;
  u8g.setPrintPos(X,Y); u8g.print('<');}

void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(byte MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if (Sens_positif) {Position_Index++; if (Position_Index == (MAX+1)) Position_Index = 0;}
  else {if (Position_Index > 0) Position_Index--; else Position_Index = MAX;}}

void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     digitalWrite(LED_Clavier, HIGH); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     digitalWrite(LED_Clavier, LOW);
     // ===========================================================
     BP_central_a_traiter = true;}}

void Traite_BP_central() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {FINI = true;
  BP_central_a_traiter = false; Position_Index = 0;}}

void MENU_du_RUN() {
  u8g.firstPage(); do {u8g.setPrintPos(40, 11); Aff_TEXTE_EEPROM(149,3);} while(u8g.nextPage());
  BIP();
  while (!FINI) {Traite_BP_central();
         if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}}
  
void MENU_BARILLET() {
  while (!FINI) {
  //------------------ Afficher le menu du BARILLET : ------------------
  u8g.firstPage();                       
  do {Afficher_un_texte(6, 11, 0, 18);
      Afficher_un_texte(18, 23, 18, 16);
      Afficher_un_texte(6, 35, 34, 18);
      Afficher_un_texte(6, 48, 52, 18);
      Afficher_un_texte(30, 60, 70, 3); u8g.setPrintPos(78,63); Aff_TEXTE_EEPROM(83,6);
  Affiche_INDEX_10();}
      while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) {TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9);}
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {BP_central_a_traiter = false;
  switch (Position_Index) {
     case 0 : {BIP(); FINI = true; break;}
     case 1 : {BIP(); FINI = true; break;}
     case 2 : {BIP(); FINI = true; break;}
     case 3 : {BIP(); FINI = true; break;}
     case 4 : {BIP(); FINI = true; break;}
     case 5 : {BIP(); FINI = true; break;}
     case 6 : {BIP(); FINI = true; break;}
     case 7 : {BIP(); FINI = true; break;}
     case 8 : {BIP(); FINI = true; break;}
     case 9 : {BIP(); FINI = true; break;}}}}
  Position_Index = 0;}

void Liste_contenu_EEPROM() {
  byte X,Y;
  // Afficher la liste :
  u8g.firstPage(); do {
     Y = 10; X = 0;
     for (byte I=0; I<10; I++) {
          if (I > 4) X = 66; if (I == 5) Y = 10;
          Debut_Nom(X,Y,I); if (Emplacement[I] > 0) {u8g.print(" "); if (Emplacement[I] < 10) u8g.print("0"); 
                               if (Emplacement[I] < 100) u8g.print("0"); u8g.print(Emplacement[I]);}
                            else {Aff_TEXTE_EEPROM(177, 4);}  Y = Y + 13;} Affiche_INDEX_10();}     
  while(u8g.nextPage());}

void MENU_EEPROM() {
  boolean fini = false;  byte Cible;
  //------------------------ Indexer un algorithme -----------------------
  while (!fini) {
    Liste_contenu_EEPROM(); if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9);
    Teste_activation_BP_central(); if (BP_central_a_traiter) fini = true;}
  Cible = Position_Index; BP_central_a_traiter = false; Position_Index = 0;
  if (Emplacement[Cible] == 0) {BIP(); FINI = true;}
  else {FINI = false; // Instruction à laisser juste avant le "while".
        while (!FINI) {
          u8g.firstPage(); do {
          Afficher_un_texte(40, 11, 89, 7);
          Afficher_un_texte(40, 23, 96, 11);
          Afficher_un_texte(40, 35, 107, 7);
          Afficher_un_texte(40, 48, 114, 5);
          Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
          //--------------- Traite la rotation du codeur incrémental -----------
          if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(3);
          //---------------------- Traite le B.P. central ----------------------
          Teste_activation_BP_central();
          if (BP_central_a_traiter) {
              BP_central_a_traiter = false;
              switch (Position_Index) {
                 case 0 : {FINI = true; break;}
                 case 1 : {FINI = true; break;}
                 case 2 : {FINI = true; break;}
                 case 3 : FINI = true;}}}}
    Position_Index = 0;}

void MENU_des_OPTIONS() {
  u8g.firstPage();
  do Afficher_un_texte(40, 48, 119, 7); while(u8g.nextPage());
  BIP();
  while (!FINI) {Traite_BP_central();
        if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}}

void MENU_PROGRAMME() {
  byte X,Y; FINI = false;
  while (!FINI) {
    //-------------------- Afficher le MENU PROGRAMME --------------------
    u8g.firstPage(); do {
      Afficher_un_texte(40, 11, 126, 7);
      Afficher_un_texte(40, 23, 135, 8);
      Afficher_un_texte(40, 35, 107, 7);
      Afficher_un_texte(40, 48, 143, 8);
      Afficher_un_texte(40, 60, 114, 5);
      Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
    //--------------- Traite la rotation du codeur incrémental -----------
    if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
    //---------------------- Traite le B.P. central ----------------------
    Teste_activation_BP_central();
    if (BP_central_a_traiter) {
       BP_central_a_traiter = false; FINI = false;
    switch (Position_Index) {
       case 0 : {BIP(); FINI = true; break;}
       case 1 : {BIP(); FINI = true; break;}
       case 2 : {BIP(); FINI = true; break;}
       case 3 : {BIP(); FINI = true; break;}
       case 4 : FINI = true;}}}
  Position_Index = 0;}

void MENU_de_BASE() {
  byte X,Y; FINI = false;
  //--------------------- Afficher le MENU de BASE ---------------------
  u8g.firstPage(); do {
    u8g.drawRFrame(0,0,128,64,10);
    Afficher_un_texte(40, 11, 151, 3);
    Afficher_un_texte(40, 23, 154, 8);
    Afficher_un_texte(40, 35, 162, 6);
    Afficher_un_texte(40, 48, 119, 7);
    Afficher_un_texte(40, 60, 168, 9);
    Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {
     BP_central_a_traiter = false; FINI = false;
  switch (Position_Index) {
     case 0 : {Position_Index = 0; MENU_du_RUN(); break;}
     case 1 : {Position_Index = 0; MENU_BARILLET(); break;}
     case 2 : {Position_Index = 0; MENU_EEPROM(); break;}
     case 3 : {Position_Index = 0; MENU_des_OPTIONS(); break;}
     case 4 : {Position_Index = 0; MENU_PROGRAMME();}}}}

 
