
/*    >>>>>> ATTENTION : Avant de téléverser ce croquis il faut au préalable avoir utilisé P00 <<<<<<

      >>>>>> Quand le matériel sera définitif, optimiser les valeur de comparaison   <<<<<<
      >>>>>> dans la procédure Tester_le_Clavier() en utilisant P09 qui les affiche. <<<<<<
      
                                                       **************************************
   Pour simplifier la gestion glolale, le symbole pour * Représenter "F" sera la valeur 21. *
                                                       **************************************
 Programme ajoutant à l'exploration des MENUS les ajouts suivants :
   • Valider OPTION inverse le mode MACHINE ETENDUE pour tester les autres fonctions.
   • Transfert d'un ALGORITHME en zone programme dédiée, (Tient compte de l'option PGM ETENDU.)
     Si emplacement vide fait 3 BIPs. N'utilise pas le B.P.C. mais le clavier pour valider ou fuir.
   • EFFACE l'Algorithme en zone programme si confirmation. (Tient compte de l'option PGM ETENDU.)
   • Listage des instructions du programme en zone dédiée.
   • La fonction REDIGER demande la référence du programme et efface ce dernier si le
     mode Machine ÉTENDUE n'est pas actif et que l'on confirme. Il sera impossible de saisir la référence 255.
     Si on donne 0 comme référence il y a fuite avec génération de 3 BIPs sonores.
     Pour renommer le programme il suffit de REDIGER sans effacer.
   • Efface un programme en EEPROM. Si mode ETENDU se contente de faire 3 BIPs car effacer la zone
     concernée ne présente aucune utilité. Si vraiment on veut le faire : RESET suivi du passage en
     Machine ÉTENDUE et enfin sauvegarder.
   • Sauvegarde le PGM en EEPROM.
   • Formate le programme pour recherche standard de la donnée. (Gère le mode Machine Étendue.)

Le comportement de l'écran OLED et du capteur rotatif restent inchangés mis à part le fait que
le capteur central allume la LED triple en rouge au lieu de la LED d'ARDUINO.
N'importe quel B.P. du chavier allume la LED en vert tant que la touche est apputée.
Puis il y a affichage des données du capteur qui a été sollicité. Il faut alors cliquer sur le
B.P. du codeur roratif pour revenir au menu de base.
   
Branchements à effectuer :  Identique à ceux décrits dans le croquis du démonstrateur P09.
--------------------------
                                         PGM               RAM dynamique
Compilateur version 1.7.9           15868 Octets (51%)     463 Octets (22%)    */

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)
#include <EEPROM.h>
//------------- Définition des broches d'entrées analogiques ----------------
#define Ligne_trois_touches 14 // A0 pour lire le clavier à 3 BP. 
#define Ligne_deux_touches 15  // A1 pour lire le clavier à 2 BP. 
#define BP_central 16          // A2 pour lire le bouton central du codeur rotatif.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 3    // D2 : DT
#define Codeur_B 2    // D3 : CLK 
#define BUZZER 4      // D4 : BUZZER de type actif ou passif.
#define ROUGE 5       // Anode LED ROUGE.
#define VERT 6        // Anode LED Verte.
#define BLEU 7        // Anode LED bleue.

//------------------- Déclaration des constante du programme ----------------
#define NB_tests_antirebonds 30 // Nombre de tests de stabilisation B.P.

//----------------------- Déclaration des variables -------------------------
boolean BP_central_a_traiter;
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.

boolean FINI, Machine_ETENDUE, Place_possible, Surbrillance;

byte Emplacement_Cible, Etendue, Increment, Index, MAX, Nb_TRN, Nombre, Num_BP, Num_PGM, Num_Tr, OCTET,
     OCTET1, OCTET2, Pos_grille, Pointeur, Position_Index, X, Y;

int PTR, U_sur_D0, U_sur_D1;

byte Ref_en_EEPROM[10]; // Contenu de l'EEPROM.
byte TRANSITION[6]; // Contient les 6 octets de la transition pointée;
byte PROGRAMME[120]; // 120 au lieu de 66 pour compatibilité si Machine ETENDUE virtuelle à 20 transitions.

volatile boolean Rotation;  // Indique une activation du codeur rotatif.
//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  pinMode(ROUGE, OUTPUT); pinMode(VERT, OUTPUT); pinMode(BLEU, OUTPUT); Eteindre_la_LED();
  digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
  //------------------------------- Initialisation des variables   //-------------------------------
  BP_central_a_traiter = false; Machine_ETENDUE = false; MAX = 11; Num_PGM = 0; Num_Tr = 1; Surbrillance = true;
  Front_sur_A(); Front_sur_B(); Position_Index = 4;}

void loop() {MENU_de_BASE();}

  //---------------------- Routines de servitude pour l'utilisation de l'EEPROM ----------------------

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

void AFF_0() {u8g.print('0');}

void SPC() {u8g.print(' ');}

void SPC_Interrogation() {u8g.print(" ?");}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {// Routines pour les affichages textes EEPROM.
    for (byte I=0; I<Longueur; I++) {u8g.print(char(EEPROM.read(PTR))); PTR++;}; }

void Afficher_un_texte(byte X, byte Y, int PTR, byte Longueur) {
  u8g.setPrintPos(X, Y); Aff_TEXTE_EEPROM(PTR,Longueur);}

void Utilisation_EEPROM() {// Analyse le contenu de l'EEPROM et met à jour Ref_en_EEPROM[10].
  PTR = 351; 
  for(byte I = 0; I < 10; I++) {Ref_en_EEPROM[I] = EEPROM.read(PTR); PTR = PTR + 67;}}

void Eteindre_la_LED() {digitalWrite(ROUGE, LOW); digitalWrite(VERT, LOW); digitalWrite(BLEU, LOW);}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------

void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
void Front_sur_A(){ // Interruption quand A change d'état avant B.
 if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
 if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
 
  //----------------------- Routines de servitude pour traiter le petit clavier ----------------------

void Eteindre_les_LEDs() {digitalWrite(ROUGE, LOW); digitalWrite(VERT, LOW); digitalWrite(BLEU, LOW);}

void ATTENDRE_le_relacher() {
  byte COMPTEUR = 0; digitalWrite(VERT, HIGH); Delay_60();// Attendre_la_stabilisation du travail.
  digitalWrite(VERT, HIGH);
  while (COMPTEUR < NB_tests_antirebonds) {// Attendre_la_stabilisation du repos :
        if ((analogRead(Ligne_trois_touches) < 813) || (analogRead(Ligne_deux_touches) < 776)) COMPTEUR = 0;
        else COMPTEUR++;}
  digitalWrite(VERT, LOW);}

void Tester_le_Clavier() {// Les valeurs tiennent compte de VCC = 4V car fourni par la carte NANO.
  U_sur_D0 = 0; U_sur_D1 = 0; Num_BP = 0; // Signale aucun B.P. utilisé.
  for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_D0 = U_sur_D0 + analogRead(Ligne_trois_touches);
  U_sur_D0 = U_sur_D0 / NB_tests_antirebonds;
  if (U_sur_D0 < 813) {Num_BP = 1;
      if (U_sur_D0 < 478)  Num_BP = 2; if (U_sur_D0 < 223)  Num_BP = 3;}
  for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_D1 = U_sur_D1 + analogRead(Ligne_deux_touches);
  U_sur_D1 = U_sur_D1 / NB_tests_antirebonds; 
  if (U_sur_D1 < 776) {Num_BP = 4; if (U_sur_D1 < 302)  Num_BP = 5;}
  if (Num_BP != 0) ATTENDRE_le_relacher();}

void Attendre_un_BP() {Num_BP = 0;
   while (Num_BP == 0) {digitalWrite(VERT, HIGH); Tester_le_Clavier(); Delay_60(); Eteindre_la_LED(); Delay_60();}}

  //--------------------------------------------------------------------------------------------------
void Delay_60() {delay(60);}

void Delay_200() {delay(200);}

void Alerte() {BIP(); Delay_200(); BIP(); Delay_200(); BIP();}

void Pointe_la_transition_dans_le_programme() {Index = (Num_Tr - 1) * 6;}

void Calcule_coordonnees_Grille() {
  switch (Pos_grille) {
    case 1 : {X = 52; Y = 17; break;}
    case 2 : {X = 76; Y = 17; break;}
    case 3 : {X = 100; Y = 17; break;}
    case 4 : {X = 52; Y = 32; break;}
    case 5 : {X = 76; Y = 32; break;}
    case 6 : {X = 100; Y = 32; break;}
    case 7 : {X = 52; Y = 47; break;}
    case 8 : {X = 76; Y = 47; break;}
    case 9 : X = 100; Y = 47;}}

void Affiche_TRANSITION() {
  switch (Pos_grille) {
    case 3 : {OCTET = TRANSITION[1]; Valeur_Tr(); break;}
    case 6 : {OCTET = TRANSITION[3]; Valeur_Tr(); break;}
    case 9 : {OCTET = TRANSITION[5]; Valeur_Tr(); break;}}}
    
void Transcode_ROT() {
  switch (OCTET) {
    case 8 : {OCTET = 71; break;}
    case 16 : OCTET = 68;}}

void Affiche_ROTATION() {
  switch (Pos_grille) {
    case 2 : {OCTET = TRANSITION[0] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));} break;}
    case 5 : {OCTET = TRANSITION[2] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));} break;}
    case 8 : {OCTET = TRANSITION[4] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));}}}}

void Transcode_ECR() {
  switch (OCTET) {
    case 1 : {OCTET = 66; break;}
    case 2 : {OCTET = 48; break;}
    case 4 : OCTET = 49;}}

void Affiche_ECRITURE() {
  switch (Pos_grille) {
    case 1 : {OCTET = TRANSITION[0] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));} break;}
    case 4 : {OCTET = TRANSITION[2] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));} break;}
    case 7 : {OCTET = TRANSITION[4] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));}}}}

void Valeur_Tr() {
  boolean FIN = false;
  if (OCTET > 0) {
      if (OCTET == 21) {u8g.print('F'); FIN = true;}
  if (!FIN) {if (OCTET > 9) u8g.setPrintPos((X + 6),Y + 11); u8g.print(OCTET);}}}

void Affiche_Item(){
  switch (X) {
    case 52 : {Affiche_ECRITURE(); break;}
    case 76 : {Affiche_ROTATION(); break;}
    case 100 : Affiche_TRANSITION();}}

void Affiche_les_Items_de_la_transition() {
  byte SAV_Pos_grille = Pos_grille;
  //---------------- Affiche les 9 cellules dans la grille ----------------
  for (Pos_grille = 1; Pos_grille < 10; Pos_grille++) {
      Calcule_coordonnees_Grille(); u8g.setPrintPos(X + 9, Y + 11); Affiche_Item();}
  //-----------------------------------------------------------------------
  Pos_grille = SAV_Pos_grille;}

void Aff_Grille_transition() {
  u8g.firstPage();
  do {u8g.drawRFrame(3,1,121,61,0); if(Surbrillance) {Calcule_coordonnees_Grille(); u8g.drawRFrame(X,Y,23,14,0);}
      Affiche_les_Items_de_la_transition();
      u8g.drawHLine(4,16,119); u8g.drawHLine(27,31,96); u8g.drawHLine(27,46,96);
      u8g.drawVLine(27,17,44); for (byte Y = 51; Y<100; Y = Y+24) u8g.drawVLine(Y,2,59);
      u8g.setPrintPos(22,13); u8g.print("TR"); u8g.setPrintPos(61,13); u8g.print('E'); u8g.setPrintPos(85,13);
      u8g.print('R'); u8g.setPrintPos(109,13); u8g.print('T'); u8g.setPrintPos(37,28); u8g.print('1');
      u8g.setPrintPos(37,43); AFF_0(); u8g.setPrintPos(37,58); u8g.print('B');
      u8g.setPrintPos(10,42); if (Num_Tr < 10) AFF_0(); u8g.print(Num_Tr);} while(u8g.nextPage());}

void Pointer_algorithme_cible() {PTR = 351 + ((Emplacement_Cible) * 67);}

void Tranfere_le_PGM_depuis_EEPROM() {
  byte Nb = 67;
  if (Machine_ETENDUE) {PTR = 229; Nb = 121;}
  else {Pointer_algorithme_cible(); Num_PGM = EEPROM.read(PTR); PTR++;}
  Pointeur = 0; 
  for (byte I = 1; I < Nb; I++) {
       PROGRAMME[Pointeur] = EEPROM.read(PTR); PTR++; Pointeur++;}}

void Charger_un_emplacement() {
  if (Machine_ETENDUE) {Tranfere_le_PGM_depuis_EEPROM();}
  else {Utilisation_EEPROM(); // Analyse le contenu de l'EEPROM et met à jour Ref_en_EEPROM[10].
        if (Ref_en_EEPROM[Emplacement_Cible] == 0) Alerte();
        else {u8g.firstPage(); do {Afficher_un_texte(10, 35, 184, 10); SPC(); // "Transferer "
              u8g.print(Ref_en_EEPROM[Emplacement_Cible]); SPC_Interrogation();} // " ?"
              while(u8g.nextPage());
              Attendre_un_BP(); if (Num_BP == 3) Tranfere_le_PGM_depuis_EEPROM();}}}

void Purge_la_zone_programme() {// Efface toute la zone y compris si mode Machine ÉTENDUE.
   for (byte I = 0; I < 119; I++) PROGRAMME[I] = 0;}

void le_PGM_Interrogation() {Aff_TEXTE_EEPROM(216,4); Aff_TEXTE_EEPROM(133,3); SPC_Interrogation();}

void Effacer_le_programme() {
  u8g.firstPage(); do {Afficher_un_texte(10, 35, 107, 7); le_PGM_Interrogation();} // "EFFACER le PGM ?"
  while(u8g.nextPage());
  Attendre_un_BP();
  if (Num_BP == 3) {if (!Machine_ETENDUE) Num_PGM = 0; Purge_la_zone_programme();}}
  
void Debut_Nom(byte X, byte Y, byte N) {Afficher_un_texte(X, Y, 181, 3); u8g.print(N);} // "Pgm"

void Positionne_INDEX(byte X, byte Y) {u8g.setPrintPos(X,Y); u8g.print('>');}

void Affiche_INDEX_10() {
  byte X, Y;
  X = 52; if (Position_Index > 4) X = 117;
  Y = 11 + Position_Index * 13; if (Position_Index > 4) Y = 11 + (Position_Index-5) * 13;
  u8g.setPrintPos(X,Y); u8g.print('<');}

void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(byte MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if (Sens_positif) {Position_Index++; if (Position_Index == (MAX+1)) Position_Index = 0;}
  else {if (Position_Index > 0) Position_Index--; else Position_Index = MAX;}}

void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     digitalWrite(ROUGE, HIGH); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     digitalWrite(ROUGE, LOW);
     // ===========================================================
     BP_central_a_traiter = true;}}

void Traite_BP_central() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {FINI = true;
  BP_central_a_traiter = false; Position_Index = 0;}} 

void Liste_contenu_EEPROM() {
  byte X,Y; Utilisation_EEPROM();
  // Afficher la liste :
  u8g.firstPage(); do {
     Y = 10; X = 0;
     for (byte I=0; I<10; I++) {
          if (I > 4) X = 66; if (I == 5) Y = 10;
          Debut_Nom(X,Y,I); if (Ref_en_EEPROM[I] > 0) {SPC(); if (Ref_en_EEPROM[I] < 10) AFF_0(); 
                               if (Ref_en_EEPROM[I] < 100) AFF_0(); u8g.print(Ref_en_EEPROM[I]);}
                            else {Aff_TEXTE_EEPROM(177, 4);}  Y = Y + 13;} Affiche_INDEX_10();}     
  while(u8g.nextPage());}
 
void AFF_Num_PGM(byte X, byte Y) {
  u8g.setPrintPos(X,Y); if (Num_PGM < 100) AFF_0(); if (Num_PGM < 10) AFF_0(); u8g.print(Num_PGM);}

void Saisir_Nombre(byte MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if ((Sens_positif) && (Nombre >= Increment)) Nombre = Nombre - Increment; 
  else {Nombre = Nombre + Increment; if (Nombre == (MAX+1)) Nombre = 1;}}

void Modifie_Ecriture(){
  byte Octet;
  byte Cible = 0; if (Pos_grille == 4) Cible = 2; if (Pos_grille == 7) Cible = 4;
  Octet = TRANSITION[Cible] & 7; TRANSITION[Cible] = TRANSITION[Cible] & 24;
  if (!Sens_positif)
        switch (Octet) {
          case 0 : {Octet = 1; break;}
          case 1 : {Octet = 2; break;}
          case 2 : {Octet = 4; break;}
          case 4 : Octet = 0;}
  else {switch (Octet) {
          case 0 : {Octet = 4; break;}
          case 1 : {Octet = 0; break;}
          case 2 : {Octet = 1; break;}
          case 4 : Octet = 2;}}
  TRANSITION[Cible] = TRANSITION[Cible] | Octet;}

void Modifie_Mouvement(){
  byte Octet;
  byte Cible = 0; if (Pos_grille == 5) Cible = 2; if (Pos_grille == 8) Cible = 4;
  Octet = TRANSITION[Cible] & 24; TRANSITION[Cible] = TRANSITION[Cible] & 7;
  switch (Octet) {
     case 0 : {Octet = 8; break;}
     case 8 : {Octet = 16; break;}
     case 16 : Octet = 0;}
  TRANSITION[Cible] = TRANSITION[Cible] | Octet;}

void Modifie_Transition(){
  byte Octet;
  byte Cible = 1; if (Pos_grille == 6) Cible = 3; if (Pos_grille == 9) Cible = 5; Octet = TRANSITION[Cible];
  if (Machine_ETENDUE) {
      if (Sens_positif) {if (Octet > 0) Octet--;  else Octet = 21;}
      else {Octet++; if (Octet == 22) Octet = 0;}}
  else {if (Sens_positif) {if (Octet == 21) Octet = 12; if (Octet > 0) Octet--; else Octet = 21;}
        else {Octet++; if (Octet == 22) Octet = 0; if(Octet == 12) Octet = 21;}}      
  TRANSITION[Cible] = Octet;}          

void Modifier_Item() {
  switch (Pos_grille) {
    case 1 : {Modifie_Ecriture(); break;}
    case 2 : {Modifie_Mouvement(); break;}
    case 3 : {Modifie_Transition(); break;}
    case 4 : {Modifie_Ecriture(); break;}
    case 5 : {Modifie_Mouvement(); break;}
    case 6 : {Modifie_Transition(); break;}
    case 7 : {Modifie_Ecriture(); break;}
    case 8 : {Modifie_Mouvement(); break;}
    case 9 : Modifie_Transition();}}

void Recopie_les_6_octets_de_la_transition() {
  Pointe_la_transition_dans_le_programme(); // Index = (Num_Tr - 1) * 6;
  for (byte I = 0; I < 6; I++) {TRANSITION[I] = PROGRAMME[Index]; Index++;}}

void Sauvegarde_les_6_octets_de_la_transition() {
  Pointe_la_transition_dans_le_programme(); // Index = (Num_Tr - 1) * 6;
  for (byte I = 0; I < 6; I++) {PROGRAMME[Index] = TRANSITION[I]; Index++;}}

void Saisir_Les_Items_dans_la_grille() {
   boolean Fini = false;
   Recopie_les_6_octets_de_la_transition(); // Pour travailler sur des données initiales.
   //--------------------------- Saisir les items dans la grille --------------------------
    Fini = false;
    while (!Fini) {Aff_Grille_transition(); 
       //------------- Traite le B.P. central du codeur incrémental ---------
       Teste_activation_BP_central();
       if (BP_central_a_traiter) {
           Fini = true; BP_central_a_traiter = false; Sauvegarde_les_6_octets_de_la_transition();}
       //--------------- Traite la rotation du codeur incrémental -----------
       if (Rotation) {
          INT_active = true; // Armer un nouveau comptage.
          Rotation = false; // Prise en compte du traitement.
          Modifier_Item();}
       //------------------------- Traite le petit clavier ------------------
       Tester_le_Clavier(); 
       switch (Num_BP) {
         case 1 : {Sauvegarde_les_6_octets_de_la_transition(); if (Num_Tr < MAX) {Num_Tr++;
                   Pos_grille = 1; Recopie_les_6_octets_de_la_transition();}
                   else Fini = true; break;}
         case 2 : {Surbrillance = !Surbrillance; break;}
         case 3 : {Sauvegarde_les_6_octets_de_la_transition(); if (Num_Tr > 1) Num_Tr--;
                   Pos_grille = 1; Recopie_les_6_octets_de_la_transition(); break;}
         case 4 : {Pos_grille--; if (Pos_grille == 0) Pos_grille = 9; break;}
         case 5 : {Pos_grille++; if (Pos_grille == 10) Pos_grille = 1;}}}}

void MODIFIER_le_programme() {
  boolean Fini = false; Increment = 1; Nombre = 1; Pos_grille = 1;
  //------------------------ Saisir la transition de début d'édition ------------------------
  while (!Fini) {
    u8g.firstPage(); do {u8g.setPrintPos(10,38); Aff_TEXTE_EEPROM(194, 13); // "Transition : "
    u8g.print(Nombre);} while(u8g.nextPage());
    if (Rotation) Saisir_Nombre(MAX);
  //------------------- Traite le B.P. central du codeur incrémental ------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false; Num_Tr = Nombre; if (Num_Tr == 0) Num_Tr = 1;}}
  Recopie_les_6_octets_de_la_transition();
  //-----------------------------------------------------------------------  
  Saisir_Les_Items_dans_la_grille();}
  
void REDIGER_le_programme() {
  boolean Fini = false; Increment = 1; Nombre = 1; Num_Tr = 1;   Pos_grille = 1;
  if (!Machine_ETENDUE) {
     while (!Fini) {
           //------------------- Saisir la référence du programme ---------------
           u8g.firstPage(); do {u8g.setPrintPos(10,38); Aff_TEXTE_EEPROM(207,10); Aff_TEXTE_EEPROM(133,3);
           Aff_TEXTE_EEPROM(204, 3); u8g.print(Nombre);} // "Référence PGM : "
           while(u8g.nextPage());
           //--------------- Traite la rotation du codeur incrémental -----------
           if (Rotation) Saisir_Nombre(254);
           //------------- Traite le B.P. central du codeur incrémental ---------
           Teste_activation_BP_central(); if (BP_central_a_traiter) Fini = true; BP_central_a_traiter = false;
           //------------------------- Traite le petit clavier ------------------
           Tester_le_Clavier();
           switch (Num_BP) {
             case 1 : {Increment = 1; break;}
             case 2 : {Fini = true; break;}
             case 3 : {Increment = 100; break;}
             case 4 : {Increment =5; break;}
             case 5 : Increment = 10;}}
      //---------------- Ne passe en EDITION que si la Référence est correcte ----------------
      if (Nombre > 0) {Num_PGM = Nombre; Effacer_le_programme(); Saisir_Les_Items_dans_la_grille();}
      else Alerte();}
  else Saisir_Les_Items_dans_la_grille();}

void Liste_EEPROM_et_Attente_clavier() {Liste_contenu_EEPROM(); Attendre_un_BP();}

void Effacer_un_PGM_en_EEPROM() {
  if (Machine_ETENDUE) Alerte();
  else {u8g.firstPage(); do {Afficher_un_texte(20, 38, 107, 7); SPC(); Aff_TEXTE_EEPROM(181, 3);
        u8g.print(Emplacement_Cible); SPC_Interrogation();} // " EFFACER PgmN ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {Pointer_algorithme_cible(); EEPROM.write(PTR,0); Liste_EEPROM_et_Attente_clavier();}}}

void Sauvegarder_dans_un_emplacement() {
  byte Nb = 67; Pointeur = 0; 
  if (Machine_ETENDUE) {
      u8g.firstPage(); do {Afficher_un_texte(30, 30, 96, 11); Afficher_un_texte(42, 43, 222, 6);
      SPC_Interrogation();} while(u8g.nextPage());// "SAUVEGARDER ÉTENDU ?"
      Attendre_un_BP();
      if (Num_BP == 3) {PTR = 229; Nb = 121; 
          for (byte I = 1; I < Nb; I++) {
               EEPROM.write(PTR,PROGRAMME[Pointeur]); PTR++; Pointeur++;}}}
  else {u8g.firstPage(); do {Afficher_un_texte(15, 38, 96, 11); SPC(); Aff_TEXTE_EEPROM(181, 3);
        u8g.print(Emplacement_Cible); SPC_Interrogation();} // " SAUVEGARDER PgmN ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {Pointer_algorithme_cible(); EEPROM.write(PTR,Num_PGM); PTR++;
           for (byte I = 1; I < Nb; I++) {EEPROM.write(PTR,PROGRAMME[Pointeur]); PTR++; Pointeur++;}
           Position_Index = 0; Liste_EEPROM_et_Attente_clavier();}}}

void Verifier_place_possible() {
  Place_possible = true;
  if (Machine_ETENDUE) {
      OCTET1 = PROGRAMME[114]; OCTET2 = PROGRAMME[115]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[116]; OCTET2 = PROGRAMME[117]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[118]; OCTET2 = PROGRAMME[119]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}   
  else {OCTET1 = PROGRAMME[60]; OCTET2 = PROGRAMME[61]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[62]; OCTET2 = PROGRAMME[63]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[64]; OCTET2 = PROGRAMME[65]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}
  if (!Place_possible) {BIP(); u8g.firstPage(); do {
      u8g.setPrintPos(10,30); u8g.print("Place insuffisante.");} while(u8g.nextPage()); Attendre_un_BP();}}

void Modifier_le_PGM_pour_recherche_standard_de_la_donnee() {
  byte Debut_translation = 59; Etendue = 66; Nb_TRN = 11;
  if (Num_PGM == 0) Alerte();
  else {if (Machine_ETENDUE) {Debut_translation = 113; Etendue = 120; Nb_TRN = 20;};
        BIP(); u8g.firstPage(); do {Afficher_un_texte(20, 30, 143, 8); le_PGM_Interrogation();} // "FORMATER le PGM ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {
            Verifier_place_possible();
            if (Place_possible) {
                //----------------- Décaler d'une transition les instructions -----------------             
                    for (byte I = Debut_translation; I > 0; I--) PROGRAMME[I + 6] = PROGRAMME[I];
                    PROGRAMME[7] = PROGRAMME[1]; PROGRAMME[6] = PROGRAMME[0];         
                //--------- Remplacer les trois premières lignes par la vectorisation ---------
                    PROGRAMME[0] = 0; PROGRAMME[1] = 2; PROGRAMME[2] = 0; PROGRAMME[3] = 2;
                    PROGRAMME[4] = 8; PROGRAMME[5] = 0;
                //------------------ Recalculer les adresse des translations ------------------
                for (byte I = 7; I < Etendue; I = I + 2) {
                     OCTET = PROGRAMME[I]; if((OCTET > 0) && (OCTET != (Nb_TRN + 1)))
                                              {OCTET++; PROGRAMME[I] = OCTET;}}}}}}

void MENU_des_OPTIONS() {
  Position_Index = 0;
  u8g.firstPage(); do Afficher_un_texte(35, 30, 220, 6); while(u8g.nextPage()); BIP();
  Machine_ETENDUE = !Machine_ETENDUE; 
  if (Machine_ETENDUE) {Num_PGM = 255; MAX = 20;} else {Num_PGM = 0; MAX = 11;} FINI = false;
  while (!FINI) {Traite_BP_central();
        if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}
  Position_Index = 0;}

void MENU_du_RUN() {
  u8g.firstPage(); do {u8g.setPrintPos(40, 11); Aff_TEXTE_EEPROM(149,3);} while(u8g.nextPage());
  BIP();
  while (!FINI) {Traite_BP_central();
         if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}
  Position_Index = 0;}  

void MENU_BARILLET() {
  Position_Index = 0; FINI = false;
  while (!FINI) {
  //------------------ Afficher le menu du BARILLET : ------------------
  u8g.firstPage();                       
  do {Afficher_un_texte(6, 11, 0, 18);   // "Affiche     PLATEAU"
      Afficher_un_texte(18, 23, 18, 16); // "SAUVE   RECHARGE"
      Afficher_un_texte(6, 35, 34, 18);  // "Arrange   Init.BIT"
      Afficher_un_texte(6, 48, 52, 18);  // "Origine   Pos.TETE"
      u8g.setPrintPos(0,60); u8g.print("Tourne G   Tourne D"); Affiche_INDEX_10();}
      while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) {TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9);}
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {BP_central_a_traiter = false;
  switch (Position_Index) {
     case 0 : {BIP(); FINI = true; break;}
     case 1 : {BIP(); FINI = true; break;}
     case 2 : {BIP(); FINI = true; break;}
     case 3 : {BIP(); FINI = true; break;}
     case 4 : {BIP(); FINI = true; break;}
     case 5 : {BIP(); FINI = true; break;}
     case 6 : {BIP(); FINI = true; break;}
     case 7 : {BIP(); FINI = true; break;}
     case 8 : {BIP(); FINI = true; break;}
     case 9 : {BIP(); FINI = true;}}}}
  Position_Index = 0;}

void MENU_EEPROM() {
  boolean fini = false; Position_Index = 0; FINI = false;
  //------------------------ Indexer un algorithme -----------------------
  if (!Machine_ETENDUE) {
      //------------ Liste le contenu de l'EEPROM et indexe un fichier ------------
      while (!fini) {
         Liste_contenu_EEPROM(); if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9);
         Teste_activation_BP_central(); if (BP_central_a_traiter) fini = true;}
      //---------------------------------------------------------------------------
      Emplacement_Cible = Position_Index; BP_central_a_traiter = false; Position_Index = 0;}
  while (!FINI) {
     u8g.firstPage(); do {
     Afficher_un_texte(40, 11, 89, 7);  // "CHARGER"
     Afficher_un_texte(40, 23, 96, 11); // "SAUVEGARDER"
     Afficher_un_texte(40, 35, 107, 7); // "EFFACER"
     Afficher_un_texte(40, 48, 114, 5); // "Fuite"
     //-------------------- Affiche le texte >>> ÉTENDU <<< ---------------
     if (Machine_ETENDUE) {u8g.setPrintPos(16,61); u8g.print(">>> "); Aff_TEXTE_EEPROM(220, 8); u8g.print("<<<");}
     //-------------------- Visualise l'index à sa position ---------------
     Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
     //--------------- Traite la rotation du codeur incrémental -----------
     if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(3);
     //---------------------- Traite le B.P. central ----------------------
     Teste_activation_BP_central();
     if (BP_central_a_traiter) {
         BP_central_a_traiter = false;
         switch (Position_Index) {
            case 0 : {Charger_un_emplacement(); break;}
            case 1 : {Sauvegarder_dans_un_emplacement(); break;}
            case 2 : {Effacer_un_PGM_en_EEPROM(); break;}
            case 3 : FINI = true;} FINI = true;}}
  Position_Index = 0;}

void MENU_PROGRAMME() {
  FINI = false; Position_Index = 0;  
  while (!FINI) {
    //-------------------- Afficher le MENU PROGRAMME --------------------
    u8g.firstPage(); do {
      Afficher_un_texte(30, 11, 126, 7); // "RÉDIGER"
      Afficher_un_texte(30, 23, 135, 8); // "MODIFIER"
      Afficher_un_texte(30, 35, 107, 7); // "EFFACER"
      Afficher_un_texte(30, 48, 143, 8); // "FORMATER"
      Afficher_un_texte(30, 60, 114, 5); // "Fuite"
      Afficher_un_texte(93, 27, 133, 3); // "PGM"
      AFF_Num_PGM(93,39);  u8g.drawRFrame(88,16,27,26,0);
      Positionne_INDEX(21, 11 + Position_Index * 12 );} while(u8g.nextPage());
    //--------------- Traite la rotation du codeur incrémental -----------
    if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
    //---------------------- Traite le B.P. central ----------------------
    Teste_activation_BP_central();
    if (BP_central_a_traiter) {
       BP_central_a_traiter = false; FINI = false;
       switch (Position_Index) {
          case 0 : {REDIGER_le_programme(); break;}
          case 1 : {MODIFIER_le_programme(); break;}
          case 2 : {Effacer_le_programme(); break;}
          case 3 : {Modifier_le_PGM_pour_recherche_standard_de_la_donnee(); break;}}
          FINI = true;}}
  Position_Index = 0;}

void MENU_de_BASE() {
  //--------------------- Afficher le MENU de BASE ---------------------
  u8g.firstPage(); do {
    u8g.drawRFrame(0,0,128,64,10); AFF_Num_PGM(100, 16);
    Afficher_un_texte(40, 11, 151, 3); // "RUN"
    Afficher_un_texte(40, 23, 154, 8); // "BARILLET"
    Afficher_un_texte(40, 35, 162, 6); // "EEPROM"
    Afficher_un_texte(40, 47, 119, 7); // "OPTIONS"
    Afficher_un_texte(40, 59, 168, 9); // "PROGRAMME"
    Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {
     BP_central_a_traiter = false;
     switch (Position_Index) {
        case 0 : {MENU_du_RUN(); break;}
        case 1 : {MENU_BARILLET(); break;}
        case 2 : {MENU_EEPROM(); break;}
        case 3 : {MENU_des_OPTIONS(); break;}
        case 4 : {MENU_PROGRAMME();}}}}
 
 //---------------------------- Routine de développement ---------------------------- 
 //void Tracer_le_PGM(byte N) {u8g.firstPage();
 // do {u8g.setPrintPos(22,30); u8g.print("Pt"); u8g.print(N);} while(u8g.nextPage());
 // Attendre_un_BP();}
 
 
 
