
/*    >>>>>> ATTENTION : Avant de téléverser ce croquis il faut au préalable avoir utilisé P00 <<<<<<

      >>>>>> Quand le matériel sera définitif, optimiser les valeur de comparaison   <<<<<<
      >>>>>> dans la procédure Tester_le_Clavier() en utilisant P09 qui les affiche. <<<<<<
 
 * NOTE sur la séquence de d'affichage sur OLED :
 * ==============================================
 *    u8g.firstPage(); do {
 *    Attention : Chaque affichage balaye huit fois cette séquence de type "do".  
 *    while(u8g.nextPage());
    
 Programme ajoutant aux fonctions émulées dans P10 les ajouts suivants :
   • PLATEAU : Fonction d'Affichage de la GRILLE des PIONS.
   • SAUVE : Fonction qui sauvegarde le BARILLET_initial en EEPROM. (Mémorise Origine et Pos_tete.)
   • Arrange : Saisie dans la GRILLE de l'initialisation des PIONS. L'entrée dans cette procédure n'efface pas
     l'état initial du BARILLET. Il faut faire appel à l'item RAZ.
     D fait passer la Pas à 5 et G fait passer le Pas à 1.
   • Origine : Impose la valeur de ce repère arbitraire.
   • RAZ : Force tous les pions à "B" si confirmation.
   • Sature : Fonction qui force tous les PIONS du plateau à "B", "0" ou "1".
   • Impose la position du BARILLET sous la tête de L/E.
   • Run->PLT : Recopie le contenu du PLATEAU dans le BARILLET.
   • RECHARGE : Fonction qui recharge le BARILLET_initial depuis l'EEPROM.
   • Motif : Impose au BARILLET un motif de base simple, suite de "B", "0" et "1" sur tout le BARILLET.
   • Dans l'affichage du BARILLET le temps réel est calculé et affiché mis en page.
   
   • Dans le menu de base RUN fait un BIP et charge PLATEAU a moitié avec des "0" et l'autre moitié avec
     des "1". Cette configuration est commode pour expérimenter les rotations G et D et utile également pour
     tester la fonction Run->PLT. De plus, l'item RUN replace à zéro la valeur de Nb_Cycles_Horloge.
   • Pour tester l'affichage de la durée réelle de traitement, le nombre de Cycles d'Horloge est augmenté de
     50.000.000 dans Affiche_le_BARILLET() chafois que l'on clique sur l'un des boutons du clavier quand on
     affiche le BARILLET en mode graphique. La duré consommée par le µP est simulée en supposant 4400 cycles
     par secondes. Cette durée simulée est calculée dans la procédure Affiche_type_Ecran().
 
   • MANU de BASE : les cinq touches du clavier permettent de passer l'afficheur en veille.
   
   • La procédure de gestion du clavier a été améliorée en vus d'augmenter l'anti-rebonds pour parer l'éventualité
     d'utilisation de B.P. de qualité médiocre qui présentent un "temps de stabilisation un peu long".

Branchements à effectuer :  Identique à ceux décrits dans le croquis du démonstrateur P09.
--------------------------
                                        PGM               RAM dynamique
Avant en compilateur 1.8.0        15626 Octets (34%)     439 Octets (21%) 
Après en compilateur 1.8.0        21616 Octets (70%)     730 Octets (35%)*/

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)
#include <EEPROM.h>
//------------- Définition des broches d'entrées analogiques ----------------
#define Ligne_trois_touches 14 // A0 pour lire le clavier à 3 BP. 
#define Ligne_deux_touches 15  // A1 pour lire le clavier à 2 BP. 
#define BP_central 16          // A2 pour lire le bouton central du codeur rotatif.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 3    // D2 : DT
#define Codeur_B 2    // D3 : CLK 
#define BUZZER 4      // D4 : BUZZER de type actif ou passif.
#define ROUGE 5       // Anode LED ROUGE.
#define VERT 6        // Anode LED Verte.
#define BLEU 7        // Anode LED bleue.

//------------------- Déclaration des constante du programme ----------------
#define NB_tests_antirebonds 30 // Nombre de tests de stabilisation B.P.

//----------------------- Déclaration des variables -------------------------
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.

boolean Aff_pas, BP_central_a_traiter, Barillet_Initial, Ecran_noir, FINI, Aff_Origine, Machine_ETENDUE,
        Mode_RUN, OUI, Place_possible, Renommer, Surbrillance, T_MicroP;

byte Emplacement_Cible, Etendue, Increment, Index, MAX, Nb_Pas, Nb_TRN, Nombre, Num_BP, Num_PGM, Num_Tr, OCTET,
     OCTET1, OCTET2, ORG_fenetre, Origine, Pos_grille, Pos_ORG, POS_TETE, Pos_tete, Pointeur, Pos_FENETRE,
     Position_Index, X, Y;

int PTR, U_sur_A0, U_sur_A1;

unsigned long Nb_Cycles_Horloge, T_microP;

float Duree_du_RUN; // Duree calculée en secondes pour la machine électromécanique.

byte Ref_en_EEPROM[10]; // Contenu de l'EEPROM.
byte TRANSITION[6]; // Contient les 6 octets de la transition pointée;
byte PROGRAMME[120]; // 120 au lieu de 66 pour compatibilité si Machine ETENDUE virtuelle à 20 transitions.
byte BARILLET_Initial[56]; // Si BIT == 0 : Etat "B"; Si BIT == 1 : Etat "0"; Si BIT == 2 : Etat "1";
byte BARILLET[56];
byte FENETRE[11];

volatile boolean Rotation;  // Indique une activation du codeur rotatif.
//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  pinMode(ROUGE, OUTPUT); pinMode(VERT, OUTPUT); pinMode(BLEU, OUTPUT); Eteindre_les_LEDs();
  digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
  //------------------------------- Initialisation des variables   //-------------------------------
  for (byte I=0; I<56; I++) BARILLET[I] = 0; // Tous les pions du PLATEAU sont forcés à "B".
  Aff_pas = false; Aff_Origine = true; Barillet_Initial = true; BP_central_a_traiter = false;
  Ecran_noir = false; Machine_ETENDUE = false; MAX = 11; Mode_RUN = false; Nb_Pas = 1; Num_PGM = 0; Num_Tr = 1;
  Origine = 1; Pos_tete = 1; POS_TETE = 20; Surbrillance = true; T_MicroP = true; Position_Index = 4;
  Front_sur_A(); Front_sur_B();}

void loop() {MENU_de_BASE();}

  //---------------------- Routines de servitude pour l'utilisation de l'EEPROM ----------------------

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

void AFF_0() {u8g.print('0');}

void SPC() {u8g.print(' ');}

void SPC_Interrogation() {u8g.print(" ?");}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {// Routines pour les affichages textes EEPROM.
    for (byte I=0; I<Longueur; I++) {u8g.print(char(EEPROM.read(PTR))); PTR++;}; }

void Afficher_un_texte(byte X, byte Y, int PTR, byte Longueur) {
  u8g.setPrintPos(X, Y); Aff_TEXTE_EEPROM(PTR,Longueur);}

void Utilisation_EEPROM() {// Analyse le contenu de l'EEPROM et met à jour Ref_en_EEPROM[10].
  PTR = 351; 
  for(byte I = 0; I < 10; I++) {Ref_en_EEPROM[I] = EEPROM.read(PTR); PTR = PTR + 67;}}

void Rouge() {digitalWrite(ROUGE, HIGH);}
void Verte() {digitalWrite(VERT, HIGH);}
void Bleue() {digitalWrite(BLEU, HIGH);}
void Eteindre_les_LEDs() {digitalWrite(ROUGE, LOW); digitalWrite(VERT, LOW); digitalWrite(BLEU, LOW);}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------

void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
void Front_sur_A(){ // Interruption quand A change d'état avant B.
 if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
 if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
 
  //----------------------- Routines de servitude pour traiter le petit clavier ----------------------

void ATTENDRE_le_relacher() {
  byte COMPTEUR = 0; digitalWrite(VERT, HIGH); Delay_60();// Attendre_la_stabilisation du travail.
  digitalWrite(VERT, HIGH);
  while (COMPTEUR < NB_tests_antirebonds) {// Attendre_la_stabilisation du repos :
        if ((analogRead(Ligne_trois_touches) < 813) || (analogRead(Ligne_deux_touches) < 776)) COMPTEUR = 0;
        else COMPTEUR++;}
  digitalWrite(VERT, LOW);}

void Tester_le_Clavier() {// Les valeurs tiennent compte de VCC = 5V car fourni par la ligne USB.
  U_sur_A1 = 0; Num_BP = 0; // Signale aucun B.P. utilisé.
  U_sur_A0 = analogRead(Ligne_trois_touches);
  if (U_sur_A0 < 813) {U_sur_A0 = 0; Delay_60(); 
      for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_A0 = U_sur_A0 + analogRead(Ligne_trois_touches);
      U_sur_A0 = U_sur_A0 / NB_tests_antirebonds;
      if (U_sur_A0 < 813) {Num_BP = 1;
      if (U_sur_A0 < 478)  Num_BP = 2; if (U_sur_A0 < 223)  Num_BP = 3;}}
  U_sur_A1 = analogRead(Ligne_deux_touches);
  if (U_sur_A1 < 776) {U_sur_A1 = 0; Delay_60();
      for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_A1 = U_sur_A1 + analogRead(Ligne_deux_touches);
      U_sur_A1 = U_sur_A1 / NB_tests_antirebonds;
      if (U_sur_A1 < 776) {Num_BP = 4; if (U_sur_A1 < 302)  Num_BP = 5;}}
  if (Num_BP != 0) ATTENDRE_le_relacher();}

  //--------------------------------------------------------------------------------------------------
void Delay_60() {delay(60);}

void Delay_200() {delay(200);}

void Alerte() {BIP(); Delay_200(); BIP(); Delay_200(); BIP();}

void Attendre_un_BP() {Num_BP = 0;
   while (Num_BP == 0) {Verte(); Tester_le_Clavier(); Delay_60(); Eteindre_les_LEDs(); Delay_60();}}

void Pointe_la_transition_dans_le_programme() {Index = (Num_Tr - 1) * 6;}

void Calcule_coordonnees_Grille() {
  switch (Pos_grille) {
    case 1 : {X = 52; Y = 17; break;}
    case 2 : {X = 76; Y = 17; break;}
    case 3 : {X = 100; Y = 17; break;}
    case 4 : {X = 52; Y = 32; break;}
    case 5 : {X = 76; Y = 32; break;}
    case 6 : {X = 100; Y = 32; break;}
    case 7 : {X = 52; Y = 47; break;}
    case 8 : {X = 76; Y = 47; break;}
    case 9 : X = 100; Y = 47;}}

void Affiche_TRANSITION() {
  switch (Pos_grille) {
    case 3 : {OCTET = TRANSITION[1]; Valeur_Tr(); break;}
    case 6 : {OCTET = TRANSITION[3]; Valeur_Tr(); break;}
    case 9 : {OCTET = TRANSITION[5]; Valeur_Tr(); break;}}}
    
void Transcode_ROT() {
  switch (OCTET) {
    case 8 : {OCTET = 71; break;}
    case 16 : OCTET = 68;}}

void Affiche_ROTATION() {
  switch (Pos_grille) {
    case 2 : {OCTET = TRANSITION[0] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));} break;}
    case 5 : {OCTET = TRANSITION[2] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));} break;}
    case 8 : {OCTET = TRANSITION[4] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));}}}}

void Transcode_ECR() {
  switch (OCTET) {
    case 1 : {OCTET = 66; break;}
    case 2 : {OCTET = 48; break;}
    case 4 : OCTET = 49;}}

void Affiche_ECRITURE() {
  switch (Pos_grille) {
    case 1 : {OCTET = TRANSITION[0] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));} break;}
    case 4 : {OCTET = TRANSITION[2] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));} break;}
    case 7 : {OCTET = TRANSITION[4] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));}}}}

void Valeur_Tr() {
  boolean FIN = false;
  if (OCTET > 0) {
      if (OCTET == 21) {u8g.print('F'); FIN = true;}
  if (!FIN) {if (OCTET > 9) u8g.setPrintPos((X + 6),Y + 11); u8g.print(OCTET);}}}

void Affiche_Item(){
  switch (X) {
    case 52 : {Affiche_ECRITURE(); break;}
    case 76 : {Affiche_ROTATION(); break;}
    case 100 : Affiche_TRANSITION();}}

void Affiche_les_Items_de_la_transition() {
  byte SAV_Pos_grille = Pos_grille;
  //---------------- Affiche les 9 cellules dans la grille ----------------
  for (Pos_grille = 1; Pos_grille < 10; Pos_grille++) {
      Calcule_coordonnees_Grille(); u8g.setPrintPos(X + 9, Y + 11); Affiche_Item();}
  //-----------------------------------------------------------------------
  Pos_grille = SAV_Pos_grille;}

void Aff_Grille_transition() {
  u8g.firstPage();
  do {u8g.drawRFrame(3,1,121,61,0); if(Surbrillance) {Calcule_coordonnees_Grille(); u8g.drawRFrame(X,Y,23,14,0);}
      Affiche_les_Items_de_la_transition();
      u8g.drawHLine(4,16,119); u8g.drawHLine(27,31,96); u8g.drawHLine(27,46,96);
      u8g.drawVLine(27,17,44); for (byte Y = 51; Y<100; Y = Y+24) u8g.drawVLine(Y,2,59);
      u8g.setPrintPos(22,13); u8g.print("TR"); u8g.setPrintPos(61,13); u8g.print('E'); u8g.setPrintPos(85,13);
      u8g.print('R'); u8g.setPrintPos(109,13); u8g.print('T'); u8g.setPrintPos(37,28); u8g.print('1');
      u8g.setPrintPos(37,43); AFF_0(); u8g.setPrintPos(37,58); u8g.print('B');
      u8g.setPrintPos(10,42); if (Num_Tr < 10) AFF_0(); u8g.print(Num_Tr);} while(u8g.nextPage());}

void Pointer_algorithme_cible() {PTR = 351 + ((Emplacement_Cible) * 67);}

void AFF_OUI_NON(boolean OUI) {SPC(); if (OUI) Aff_TEXTE_EEPROM(73,3); else Aff_TEXTE_EEPROM(76,3);}

void le_PGM() {Aff_TEXTE_EEPROM(216,4); Aff_TEXTE_EEPROM(133,3);}

void le_PGM_Interrogation() {le_PGM(); SPC_Interrogation();}

void Saisir_OUI_NON(byte TEXTE) {
  boolean Fini = false; OUI = true;
  while (!Fini) {
    u8g.firstPage(); do {u8g.setPrintPos(10,38); // Position du texte par défaut vers le centre de l'écran.
    switch(TEXTE) {
      case 1 : {Afficher_un_texte(5, 38, 89, 7); Aff_TEXTE_EEPROM(216,4);
                Aff_TEXTE_EEPROM(13,5); break;} // "CHARGER le Motif "
      case 2 : {Afficher_un_texte(20, 38, 96, 11); break;} // "SAUVEGARDER"
      case 3 : {Afficher_un_texte(20, 38, 89, 7); break;} // "CHARGER"
      case 4 : {u8g.setPrintPos(22,30); u8g.print("RAZ : "); break;}
      case 5 : {u8g.setPrintPos(3,38); u8g.print("Tout le plateau"); u8g.print(char(224)); break;} // "Sature" ...
      case 6 : {Afficher_un_texte(10, 35, 184, 10); SPC(); // "Transferer "
                u8g.print(Ref_en_EEPROM[Emplacement_Cible]); break;}
      case 7 : {Afficher_un_texte(10, 35, 107, 7); le_PGM(); break;} // "EFFACER le PGM"
      case 8 : {Afficher_un_texte(10, 30, 143, 8); le_PGM(); break;} // "FORMATER le PGM ?"
      case 9 : {u8g.setPrintPos(3,38); u8g.print("PLT vers Initial");}}
    AFF_OUI_NON(OUI);}
    while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) OUI = !OUI; INT_active = true; // Armer un nouveau comptage.
                Rotation = false; // Prise en compte du traitement.
  //------------- Traite le B.P. central du codeur incrémental ---------
  Teste_activation_BP_central(); if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}}}

void Modifier_le_PGM_pour_recherche_standard_de_la_donnee() {
  byte Debut_translation = 59; Etendue = 66; Nb_TRN = 11;
  if (Num_PGM == 0) Alerte();
  else {if (Machine_ETENDUE) {Debut_translation = 113; Etendue = 120; Nb_TRN = 20;};
        BIP(); Saisir_OUI_NON(8);
        if (OUI) {
            Verifier_place_possible();
            if (Place_possible) {
                //----------------- Décaler d'une transition les instructions -----------------             
                    for (byte I = Debut_translation; I > 0; I--) PROGRAMME[I + 6] = PROGRAMME[I];
                    PROGRAMME[7] = PROGRAMME[1]; PROGRAMME[6] = PROGRAMME[0];         
                //--------- Remplacer les trois premières lignes par la vectorisation ---------
                    PROGRAMME[0] = 0; PROGRAMME[1] = 2; PROGRAMME[2] = 0; PROGRAMME[3] = 2;
                    PROGRAMME[4] = 8; PROGRAMME[5] = 0;
                //------------------ Recalculer les adresse des translations ------------------
                for (byte I = 7; I < Etendue; I = I + 2) {
                     OCTET = PROGRAMME[I]; if((OCTET > 0) && (OCTET != 21))
                                              {OCTET++; PROGRAMME[I] = OCTET;}}}}}}

void Charger_un_emplacement() {
  if (Machine_ETENDUE) {Tranfere_le_PGM_depuis_EEPROM();}
  else {Utilisation_EEPROM(); // Analyse le contenu de l'EEPROM et met à jour Ref_en_EEPROM[10].
        if (Ref_en_EEPROM[Emplacement_Cible] == 0) Alerte();
        else {Saisir_OUI_NON(6); if (OUI) Tranfere_le_PGM_depuis_EEPROM();}}}

void Tranfere_le_PGM_depuis_EEPROM() {
  byte Nb = 67;
  if (Machine_ETENDUE) {PTR = 229; Nb = 121;}
  else {Pointer_algorithme_cible(); Num_PGM = EEPROM.read(PTR); PTR++;}
  Pointeur = 0; 
  for (byte I = 1; I < Nb; I++) {
       PROGRAMME[Pointeur] = EEPROM.read(PTR); PTR++; Pointeur++;}}

void Purge_la_zone_programme() {// Efface toute la zone y compris si mode Machine ÉTENDUE.
   for (byte I = 0; I < 119; I++) PROGRAMME[I] = 0;}

void Effacer_le_programme() {
  Saisir_OUI_NON(7); if (OUI){if (!Machine_ETENDUE && Renommer) Num_PGM = 0; Purge_la_zone_programme();}}
  
void Debut_Nom(byte X, byte Y, byte N) {Afficher_un_texte(X, Y, 181, 3); u8g.print(N);} // "Pgm"

void Positionne_INDEX(byte X, byte Y) {u8g.setPrintPos(X,Y); u8g.print('>');}

void Affiche_INDEX_10() {
  byte X, Y;
  X = 52; if (Position_Index > 4) X = 117;
  Y = 11 + Position_Index * 13; if (Position_Index > 4) Y = 11 + (Position_Index-5) * 13;
  u8g.setPrintPos(X,Y); u8g.print('<');}

void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(byte MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if (Sens_positif) {Position_Index++; if (Position_Index == (MAX+1)) Position_Index = 0;}
  else {if (Position_Index > 0) Position_Index--; else Position_Index = MAX;}}

void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     Rouge(); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     Eteindre_les_LEDs();
     // ===========================================================
     BP_central_a_traiter = true;}}

void Traite_BP_central() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {FINI = true;
  BP_central_a_traiter = false; Position_Index = 0;}} 

void Liste_contenu_EEPROM() {
  byte X,Y; Utilisation_EEPROM();
  // Afficher la liste :
  u8g.firstPage(); do {
     Y = 10; X = 0;
     for (byte I=0; I<10; I++) {
          if (I > 4) X = 66; if (I == 5) Y = 10;
          Debut_Nom(X,Y,I); if (Ref_en_EEPROM[I] > 0) {SPC(); if (Ref_en_EEPROM[I] < 10) AFF_0(); 
                               if (Ref_en_EEPROM[I] < 100) AFF_0(); u8g.print(Ref_en_EEPROM[I]);}
                            else {Aff_TEXTE_EEPROM(177, 4);}  Y = Y + 13;} Affiche_INDEX_10();}     
  while(u8g.nextPage());}
 
void AFF_Num_PGM(byte X, byte Y) {
  u8g.setPrintPos(X,Y); if (Num_PGM < 100) AFF_0(); if (Num_PGM < 10) AFF_0(); u8g.print(Num_PGM);}

void Modifie_Ecriture(){
  byte Octet;
  byte Cible = 0; if (Pos_grille == 4) Cible = 2; if (Pos_grille == 7) Cible = 4;
  Octet = TRANSITION[Cible] & 7; TRANSITION[Cible] = TRANSITION[Cible] & 24;
  if (Sens_positif)
        switch (Octet) {
          case 0 : {Octet = 1; break;}
          case 1 : {Octet = 2; break;}
          case 2 : {Octet = 4; break;}
          case 4 : Octet = 0;}
  else {switch (Octet) {
          case 0 : {Octet = 4; break;}
          case 1 : {Octet = 0; break;}
          case 2 : {Octet = 1; break;}
          case 4 : Octet = 2;}}
  TRANSITION[Cible] = TRANSITION[Cible] | Octet;}

void Modifie_Mouvement(){
  byte Octet;
  byte Cible = 0; if (Pos_grille == 5) Cible = 2; if (Pos_grille == 8) Cible = 4;
  Octet = TRANSITION[Cible] & 24; TRANSITION[Cible] = TRANSITION[Cible] & 7;
  switch (Octet) {
     case 0 : {Octet = 8; break;}
     case 8 : {Octet = 16; break;}
     case 16 : Octet = 0;}
  TRANSITION[Cible] = TRANSITION[Cible] | Octet;}

void Modifie_Transition(){
  byte Octet;
  byte Cible = 1; if (Pos_grille == 6) Cible = 3; if (Pos_grille == 9) Cible = 5; Octet = TRANSITION[Cible];
  if (Machine_ETENDUE) {
      if (Sens_positif) {if (Octet > 0) Octet--;  else Octet = 21;}
      else {Octet++; if (Octet == 22) Octet = 0;}}
  else {if (Sens_positif) {if (Octet == 21) Octet = 12; if (Octet > 0) Octet--; else Octet = 21;}
        else {Octet++; if (Octet == 22) Octet = 0; if(Octet == 12) Octet = 21;}}      
  TRANSITION[Cible] = Octet;}          

void Modifier_Item() {
  switch (Pos_grille) {
    case 1 : {Modifie_Ecriture(); break;}
    case 2 : {Modifie_Mouvement(); break;}
    case 3 : {Modifie_Transition(); break;}
    case 4 : {Modifie_Ecriture(); break;}
    case 5 : {Modifie_Mouvement(); break;}
    case 6 : {Modifie_Transition(); break;}
    case 7 : {Modifie_Ecriture(); break;}
    case 8 : {Modifie_Mouvement(); break;}
    case 9 : Modifie_Transition();}}

void Recopie_les_6_octets_de_la_transition() {
  Pointe_la_transition_dans_le_programme(); // Index = (Num_Tr - 1) * 6;
  for (byte I = 0; I < 6; I++) {TRANSITION[I] = PROGRAMME[Index]; Index++;}}

void Sauvegarde_les_6_octets_de_la_transition() {
  Pointe_la_transition_dans_le_programme(); // Index = (Num_Tr - 1) * 6;
  for (byte I = 0; I < 6; I++) {PROGRAMME[Index] = TRANSITION[I]; Index++;}}

void Saisir_Les_Items_dans_la_grille() {
   boolean Afficher = true; // Pour n'afficher la grille que si modification du contenu;
   boolean Fini = false;
   Recopie_les_6_octets_de_la_transition(); // Pour travailler sur des données initiales.
   //--------------------------- Saisir les items dans la grille --------------------------
    Fini = false;
    while (!Fini) {if (Afficher) {Aff_Grille_transition(); Afficher = false;} 
       //------------- Traite le B.P. central du codeur incrémental ---------
       Teste_activation_BP_central();
       if (BP_central_a_traiter) {
           Fini = true; BP_central_a_traiter = false; Sauvegarde_les_6_octets_de_la_transition();}
       //--------------- Traite la rotation du codeur incrémental -----------
       if (Rotation) {Afficher = true;
          INT_active = true; // Armer un nouveau comptage.
          Rotation = false; // Prise en compte du traitement.
          Modifier_Item();}
       //------------------------- Traite le petit clavier ------------------
       Tester_le_Clavier(); if (Num_BP>0) Afficher = true; 
       switch (Num_BP) {
         case 1 : {Sauvegarde_les_6_octets_de_la_transition(); if (Num_Tr < MAX) {Num_Tr++;
                   Pos_grille = 1; Recopie_les_6_octets_de_la_transition();}
                   else Fini = true; break;}
         case 2 : {Surbrillance = !Surbrillance; break;}
         case 3 : {Sauvegarde_les_6_octets_de_la_transition(); if (Num_Tr > 1) Num_Tr--;
                   Pos_grille = 1; Recopie_les_6_octets_de_la_transition(); break;}
         case 4 : {Pos_grille--; if (Pos_grille == 0) Pos_grille = 9; break;}
         case 5 : {Pos_grille++; if (Pos_grille == 10) Pos_grille = 1;}}}}

void Saisir_Nombre(byte MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if ((Sens_positif) && (Nombre >= Increment)) Nombre = Nombre - Increment; 
  else {Nombre = Nombre + Increment; if (Nombre > MAX) Nombre = 1;}}

void Saisir_un_nombre(byte TEXTE, byte MAX) {
  boolean Fini = false; Increment = 1; Nombre = 1;
  //------------------------ Saisir la transition de début d'édition ------------------------
  while (!Fini) {
    u8g.firstPage(); do {u8g.setPrintPos(10,38); // Position du texte par défaut vers le centre de l'écran.
    switch(TEXTE) {
      case 1 : {Aff_TEXTE_EEPROM(207,10); Aff_TEXTE_EEPROM(133,3);
                Aff_TEXTE_EEPROM(204,3); break;} // "Référence PGM : "
      case 2 : {u8g.print("Origine en : "); break;}
      case 3 : {Aff_TEXTE_EEPROM(194, 13); break;} // "Transition : "
      case 4 : {u8g.print("Position "); Aff_TEXTE_EEPROM(66,4); Aff_TEXTE_EEPROM(204,3);}}
    u8g.print(Nombre);}
    while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) Saisir_Nombre(MAX);
  //------------- Traite le B.P. central du codeur incrémental ---------
  Teste_activation_BP_central(); if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}
  //------------------------- Traite le petit clavier ------------------
  Tester_le_Clavier();
  switch (Num_BP) {
     case 1 : {Increment = 1; break;}
     case 2 : {if ((TEXTE == 2) || (TEXTE == 4)) BIP(); else Fini = true; break;}
     case 3 : {if ((TEXTE == 2) || (TEXTE == 4)) BIP(); else Increment = 100; break;}
     case 4 : {Increment = 5; break;}
     case 5 : Increment = 10;}}}

void REDIGER_le_programme() {
  Num_Tr = 1;   Pos_grille = 1;
  if (!Machine_ETENDUE) {
     Saisir_un_nombre(1,254); // Saisir la Référence du programme
     //---------------- Ne passe en EDITION que si la Référence est correcte ----------------
     if (Nombre > 0) {Num_PGM = Nombre; Renommer = false; Effacer_le_programme(); Saisir_Les_Items_dans_la_grille();}
     else Alerte();}
  else Saisir_Les_Items_dans_la_grille();}

void MODIFIER_le_programme() {
  byte MAX = 11; if (Machine_ETENDUE) MAX = 20;
  boolean Fini = false; Increment = 1; Nombre = 1; Pos_grille = 1;
  //------------------------ Saisir la transition de début d'édition ------------------
  Saisir_un_nombre(3, MAX); // 3 pour le texte "Transition : "
  //-----------------------------------------------------------------------
  Num_Tr = Nombre; if (Num_Tr == 0) Num_Tr = 1; Recopie_les_6_octets_de_la_transition();
  Saisir_Les_Items_dans_la_grille();}

void Liste_EEPROM_et_Attente_clavier() {Liste_contenu_EEPROM(); Attendre_un_BP();}

void Effacer_un_PGM_en_EEPROM() {
  if (Machine_ETENDUE) Alerte();
  else {u8g.firstPage(); do {Afficher_un_texte(20, 38, 107, 7); SPC(); Aff_TEXTE_EEPROM(181, 3);
        u8g.print(Emplacement_Cible); SPC_Interrogation();} // " EFFACER PgmN ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {Pointer_algorithme_cible(); EEPROM.write(PTR,0); Liste_EEPROM_et_Attente_clavier();}}}

void Sauvegarder_dans_un_emplacement() {
  byte Nb = 67; Pointeur = 0;
  if (Machine_ETENDUE) {
      u8g.firstPage(); do {Afficher_un_texte(30, 30, 96, 11); Afficher_un_texte(42, 43, 220, 6);
      SPC_Interrogation();} while(u8g.nextPage());// "SAUVEGARDER ÉTENDU ?"
      Attendre_un_BP();
      if (Num_BP == 3) {PTR = 229; Nb = 121; 
          for (byte I = 1; I < Nb; I++) {
               EEPROM.write(PTR,PROGRAMME[Pointeur]); PTR++; Pointeur++;}}}
  else {u8g.firstPage(); do {Afficher_un_texte(15, 38, 96, 11); SPC(); Aff_TEXTE_EEPROM(181, 3);
        u8g.print(Emplacement_Cible); SPC_Interrogation();} // " SAUVEGARDER PgmN ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {Pointer_algorithme_cible(); EEPROM.write(PTR,Num_PGM); PTR++;
           for (byte I = 1; I < Nb; I++) {EEPROM.write(PTR,PROGRAMME[Pointeur]); PTR++; Pointeur++;}
           Position_Index = 0; Liste_EEPROM_et_Attente_clavier();}}}

void Verifier_place_possible() {
  Place_possible = true;
  if (Machine_ETENDUE) {
      OCTET1 = PROGRAMME[114]; OCTET2 = PROGRAMME[115]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[116]; OCTET2 = PROGRAMME[117]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[118]; OCTET2 = PROGRAMME[119]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}   
  else {OCTET1 = PROGRAMME[60]; OCTET2 = PROGRAMME[61]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[62]; OCTET2 = PROGRAMME[63]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[64]; OCTET2 = PROGRAMME[65]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}
  if (!Place_possible) {BIP(); u8g.firstPage(); do {
      u8g.setPrintPos(10,30); u8g.print("Place insuffisante.");} while(u8g.nextPage()); Attendre_un_BP();}}
  
//=============================== Routines de gestion de l'affichage du BARILLET ===============================
void Affiche_Origine(byte X, byte Y) {u8g.drawRFrame(X-2,Y-2,5,5,1); u8g.drawPixel(X,Y);}

void Affiche_Repere_5(byte X) {u8g.drawVLine(X,26,7);}

void Affiche_5(byte X, byte Y) {
  u8g.drawHLine(X-3,Y-2,3); u8g.drawHLine(X-3,Y,2); u8g.drawHLine(X-3,Y+2,2); u8g.drawPixel(X-3,Y-1);
  u8g.drawPixel(X-1,Y+1);}

void Affiche_1(byte X, byte Y) {
     u8g.drawVLine(X+1,Y-2,5); u8g.drawPixel(X,Y-1); u8g.drawPixel(X,Y+2); u8g.drawPixel(X+2,Y+2);}

void Affiche_0(byte X, byte Y) {
     u8g.drawVLine(X,Y-1,3); u8g.drawVLine(X+2,Y-1,3); u8g.drawPixel(X+1,Y+2); u8g.drawPixel(X+1,Y-2);}

void Affiche_Repere(byte X, byte Y, byte Valeur) {
  switch(Valeur) {
    case 1  : {Affiche_1(X,Y); break;}
    case 5  : {Affiche_Repere_5(X); break;}  
    case 10 : {Affiche_1(X-3,Y); Affiche_0(X+1,Y); break;}
    case 15 : {Affiche_Repere_5(X); break;}
    case 20 : {u8g.drawHLine(X-3,Y-2,2); u8g.drawHLine(X-3,Y+2,3); u8g.drawPixel(X-1,Y-1); u8g.drawPixel(X-2,Y);
               u8g.drawPixel(X-3,Y+1); Affiche_0(X+1,Y); break;}      
    case 25 : {Affiche_Repere_5(X); break;}
    case 30 : {u8g.drawHLine(X-3,Y-2,2); u8g.drawHLine(X-3,Y+2,2); u8g.drawPixel(X-1,Y-1);
               u8g.drawPixel(X-2,Y); u8g.drawPixel(X-1,Y+1); Affiche_0(X+1,Y); break;}
    case 35 : {Affiche_Repere_5(X); break;}
    case 40 : {u8g.drawVLine(X-1,Y-2,5); u8g.drawHLine(X-3,Y+1,4); u8g.drawPixel(X-3,Y);
               u8g.drawPixel(X-2,Y-1); Affiche_0(X+2, Y); break;}
    case 45 : {Affiche_Repere_5(X); break;}
    case 50 : {Affiche_5(X,Y); Affiche_0(X+1,Y); break;}
    case 55 : Affiche_Repere_5(X);}}

void Affiche_les_reperes() {
  byte Xorg = X;
  byte Pos_FENETRE = 0;
  byte Position = Pos_tete - 1; if (!Barillet_Initial) Position = POS_TETE - 1;
  for (byte X=4; X <130; X=X+12) {
      if (Position != Origine) Affiche_Repere(X,29,Position);
      Pos_FENETRE++; Position++; if (Position==57) Position = 1;}}

 void Afficher_le_temps_reel_pour_executer() {
  byte Cas_a_afficher = 3;
  float Temps_qui_reste, Quotient;
  int Annees = 0; byte Heures = 0; byte Minutes = 0; int Jours = 0;
  Temps_qui_reste = Duree_du_RUN;
  //---------------------- Déterminer si dépasse un AN soit 31536000 s ---------------------- 
  if (Temps_qui_reste > 31536000) {Quotient = Temps_qui_reste / 31536000; Annees = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 31536000);}
  //----------------------- Déterminer si dépasse un JOUR soit 86400 s ----------------------
  if (Temps_qui_reste > 86400) {
     Quotient = Temps_qui_reste / 86400; Jours = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 86400);}
  //---------------------- Déterminer si dépasse une HEURE soit 3600 s ----------------------
  if (Temps_qui_reste > 3600) {
     Quotient = Temps_qui_reste / 3600; Heures = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 3600);}
  //----------------------- Déterminer si dépasse une minute soit 60 s ----------------------
  if (Temps_qui_reste > 60) {
     Quotient = Temps_qui_reste / 60; Minutes = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 60);}
  //-----------------------------------------------------------------------------------------
  if (Annees > 0) Cas_a_afficher = 1;
  if ((Annees == 0) && (Jours > 0)) Cas_a_afficher = 2;
  switch(Cas_a_afficher) {
    case 1 : {u8g.print(Annees); u8g.print("A "); u8g.print(Jours); u8g.print('J'); break;}
    case 2 : {u8g.print(Jours); u8g.print("J "); u8g.print(Heures); u8g.print('H'); break;}
    case 3 : {if (Heures>0) {u8g.print(Heures); u8g.print("H ");} u8g.print(Minutes); u8g.print("' ");
              if (Duree_du_RUN > 60) {u8g.print(Temps_qui_reste,0); u8g.print('"');}}}}

void Affiche_type_Ecran() {
  u8g.setPrintPos(35,8);
  if (Mode_RUN) {Aff_TEXTE_EEPROM(181, 3); AFF_Num_PGM(63, 8); u8g.setPrintPos(95,8); u8g.print("PAUSE");}
  else {if (Barillet_Initial) u8g.print("INITIAL"); else Aff_TEXTE_EEPROM(0,7);
        Affiche_Origine(94,3); u8g.setPrintPos(101,8); u8g.print("= "); u8g.print(Origine); u8g.setPrintPos(15,49);
        if (T_MicroP) {
            Aff_TEXTE_EEPROM(79,4); u8g.print(char(181)); u8g.print("P ");
            u8g.print((Nb_Cycles_Horloge / 4400.0),2); // Simulation en supposant 4400 cycles / S.
            u8g.print(" S");}
        else {u8g.print("Tr"); u8g.print(char(233)); u8g.print("el "); Afficher_le_temps_reel_pour_executer();}}}

void Affiche_PION(byte X, byte Etat) {
  if (Etat == 0) u8g.drawHLine(X-2,24,5);
  if (Etat == 1) u8g.drawRBox(X-2,18,5,7,0);
  if (Etat == 2) u8g.drawRBox(X-2,12,5,13,0);}

void Affiche_NB_Pas() {
  if (Aff_pas) {u8g.drawRFrame(1,53,7,9,0); if (Nb_Pas == 5) {Affiche_5(6,57);} else Affiche_Repere(3,57,1);}}

void Tourne_a_gauche() {// Fait tourner le plateau Machine ou INITIAL en fonction de Barillet_Initial.
  // Faire tourner le plateau à gauche consiste à incrémenter la position de la Tête de L/E.
  byte OCTET; if(Barillet_Initial) {OCTET = BARILLET_Initial[0]; Pos_tete++; if (Pos_tete == 57) Pos_tete = 1;} 
              else {OCTET = BARILLET[0]; POS_TETE++; if (POS_TETE == 57) POS_TETE = 1;}}

void Tourne_a_droite() {// Fait tourner le plateau Machine ou INITIAL en fonction de Barillet_Initial.
  // Faire tourner le plateau à droite consiste à décrémenter la position de la Tête de L/E.
  byte OCTET; if(Barillet_Initial) {OCTET = BARILLET_Initial[55]; Pos_tete--; if (Pos_tete == 0) Pos_tete = 56;}
              else {OCTET = BARILLET[55];  POS_TETE--; if (POS_TETE == 0) POS_TETE = 56;}}

void Affiche_la_GRILLE() {
  byte Position_Tete = Pos_tete; if (!Barillet_Initial) Position_Tete = POS_TETE;
  byte Index, X, Y ;
  //--------------------------------- Affiche la grille ----------------------------------
  u8g.firstPage(); do {
  u8g.drawRFrame(13,2,101,61,0);
  for (byte I= 4; I<63; I = I+8) u8g.drawHLine(16,I,95);
  for (byte I= 15; I<120; I = I+12) u8g.drawVLine(I,4,57);
  //------------------------------ Affiche la position tête ------------------------------
  Index = 0; X = 0; Y = 1;
  while (Index != Position_Tete) {Index++; X++; if (X == 9) {X = 1; Y++;}}
  X = (X * 12) + 3; Y = (Y * 8) - 4;
  u8g.drawHLine(X,0,13); u8g.drawHLine(X,1,13); u8g.drawVLine(12,Y,9); u8g.drawVLine(11,Y,9);
  //--------------------------------- Affiche l'origine ----------------------------------
  if (Aff_Origine) {// Ne pas afficher en mode "Arrange" le BARILLET.
      Index = 0; X = 0; Y = 1; // Barillet_Initial a déjà été pris en compte à la déclaration de Position_Tete.
      while (Index != Origine) {Index++; X++; if (X == 9) {X = 1; Y++;}}
      X = (X * 12) + 3; Y = (Y * 8) - 4;
      u8g.drawHLine(X,61,13); u8g.drawHLine(X,63,13);u8g.drawVLine(112,Y,9); u8g.drawVLine(114,Y,9);}
  //------------------------ Affiche les PIXELs des repères en 5 -------------------------
  u8g.drawPixel(69,8); u8g.drawPixel(93,16); u8g.drawPixel(21,32);
  u8g.drawPixel(45,40); u8g.drawPixel(69,48); u8g.drawPixel(93,56);
  //--------------------------- Affiche les repèrères décimaux ---------------------------
  Affiche_reperes_decimaux(); // N'afficher le repère que si le pion est à l'état "B".
  //----------------------------- Affiche l'état des 56 BITs -----------------------------
  X = 21; Y = 8;
  for (byte I = 0; I < 56; I++) {
       if (Barillet_Initial)
          switch (BARILLET_Initial[I]) {// Si valeur 0 : "B" donc ne rien afficher et valeur non prise en compte.
             case 1 :  {Affiche_BIT_0 (X,Y); break;}
             case 2 :   Affiche_BIT_1 (X,Y);}
       else switch (BARILLET[I]) {// Si valeur 0 : "B" donc ne rien afficher et valeur non prise en compte.
               case 1 :  {Affiche_BIT_0 (X,Y); break;}
               case 2 :   Affiche_BIT_1 (X,Y);}
       X = X + 12; if (X == 117) {X = 21; Y = Y + 8;}}
  //------------------------- Affiche le symbole Origine ou le P -------------------------
  u8g.drawRFrame(0,28,9,9,0);
  if (Barillet_Initial) Affiche_Origine(4,32);
  else {u8g.drawVLine(2,30,5); u8g.drawHLine(3,30,3); u8g.drawHLine(3,32,3); u8g.drawPixel(6,31);}
  //------------- Affiche les repères et le nombre de Pas pour G et D---------------------
  Affiche_NB_Pas();} while(u8g.nextPage());}

void Initialise_le_BARILLET_Initial() {
  boolean Afficher = true; // Pour n'afficher la grille que si modification du contenu;
  boolean Fini = false; byte SAV_tete = Pos_tete; Pos_tete = 1;
  Barillet_Initial = true; Aff_Origine = false;  Nb_Pas = 1;
  while (!Fini) {
        if (Afficher) {Affiche_la_GRILLE(); Afficher = false;}
        //--------------- Traite la rotation du codeur incrémental -----------
        if (Rotation) {Afficher = true;
           if (Sens_positif) {Pos_tete = Pos_tete - Nb_Pas; if ((Pos_tete > 200) || (Pos_tete == 0)) Pos_tete = 56;}
                       else {Pos_tete = Pos_tete + Nb_Pas; if (Pos_tete > 56) Pos_tete = 1;} Rotation = false;}
        //---------------------- Traite le B.P. central ----------------------
        Teste_activation_BP_central(); if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}           
        //------------------------- Traite le petit clavier ------------------
        Tester_le_Clavier();
        if (Num_BP > 0) 
            switch (Num_BP) {
              case 1 : {BARILLET_Initial[Pos_tete - 1] = 0; Afficher = true; break;}
              case 2 : {BARILLET_Initial[Pos_tete - 1] = 1; Afficher = true; break;}
              case 3 : {BARILLET_Initial[Pos_tete - 1] = 2; Afficher = true; break;}
              case 4 : {Nb_Pas = 1; Afficher = true; break;}
              case 5 : {Nb_Pas = 5; Afficher = true;}}}
  Pos_tete = SAV_tete; Aff_Origine = true;}

void Sauvegarder_le_BARILLET() {
  PTR = 295;
  for (byte I = 0; I < 56; I++) {EEPROM.write(PTR,BARILLET_Initial[I]); PTR++;}
  EEPROM.write(227,Pos_tete); EEPROM.write(228,Origine);}

void Recharger_le_BARILLET() {
  PTR = 295;
  for (byte I = 0; I < 56; I++) {BARILLET_Initial[I] = EEPROM.read(PTR); PTR++;}
  Pos_tete = EEPROM.read(227); Origine = EEPROM.read(228);}

void Sauvegarde_le_BARILLET() {Saisir_OUI_NON(2); if (OUI) Sauvegarder_le_BARILLET();}
  
void Transfere_le_BARILLET() {Saisir_OUI_NON(3);
  if (OUI) {Recharger_le_BARILLET(); Barillet_Initial = true; Affiche_la_GRILLE(); Attendre_un_BP();}}

void Saisie_Origine() {Saisir_un_nombre(2,56); Origine = Nombre;}

void Affiche_BIT_0(byte X, byte Y) {u8g.drawRFrame(X-2, Y-2,5,5,0);}

void Affiche_BIT_1(byte X, byte Y) {
  Affiche_BIT_0(X,Y); u8g.drawRFrame(X-3, Y-3,7,7,0); u8g.drawVLine(X-4,Y-3,7); u8g.drawVLine(X+4,Y-3,7);
  u8g.drawVLine(X-4,Y-3,7); u8g.drawVLine(X-5,Y-3,7); u8g.drawVLine(X+5,Y-3,7);}

void Affiche_reperes_decimaux() {
  // N'afficher le repère que si le pion est à l'état "B".
  if (Barillet_Initial) {if (BARILLET_Initial[0] == 0)Affiche_Repere(21, 8, 1);
     if (BARILLET_Initial[9] == 0) Affiche_Repere(33, 16, 10);
     if (BARILLET_Initial[19] == 0) Affiche_Repere(57, 24, 20);
     if (BARILLET_Initial[29] == 0) Affiche_Repere(81, 32, 30);
     if (BARILLET_Initial[39] == 0) Affiche_Repere(105, 40, 40);
     if (BARILLET_Initial[49] == 0) Affiche_Repere(33, 56, 50);}
  else {if (BARILLET[0] == 0)Affiche_Repere(21, 8, 1); if (BARILLET[9] == 0) Affiche_Repere(33, 16, 10);
     if (BARILLET[19] == 0) Affiche_Repere(57, 24, 20); if (BARILLET[29] == 0) Affiche_Repere(81, 32, 30);
     if (BARILLET[39] == 0) Affiche_Repere(105, 40, 40); if (BARILLET[49] == 0) Affiche_Repere(33, 56, 50);}}

void Recopie_les_pions_dans_la_fenetre() {
  if (Barillet_Initial) {
       Pos_FENETRE = Pos_tete; Pos_FENETRE--; if (Pos_FENETRE == 0) Pos_FENETRE = 55;
       for (byte I=0; I<11; I++) {
            FENETRE[I] = BARILLET_Initial[Pos_FENETRE]; Pos_FENETRE++;
            if (Pos_FENETRE > 55) Pos_FENETRE = 0;}}
            
  else {Pos_FENETRE = POS_TETE; Pos_FENETRE--; if (Pos_FENETRE == 0) Pos_FENETRE = 55;
       Pos_FENETRE--; if (Pos_FENETRE == 0) Pos_FENETRE = 55;
       for (byte I=0; I<11; I++) {
            FENETRE[I] = BARILLET[Pos_FENETRE]; Pos_FENETRE++;
            if (Pos_FENETRE > 55) Pos_FENETRE = 0;}}}

void Determine_la_position_origine_dans_FENETRE() {
  byte SAVorg = Origine;
  Pos_ORG = Pos_tete; 
  ORG_fenetre = 200; // Valeur arbitraire > 10;
  if (Barillet_Initial) {
      Pos_ORG = Pos_tete; Pos_ORG--; if(Pos_ORG == 0) Pos_ORG = 56;
      for (byte I=0; I<11; I++){
           if (Pos_ORG == Origine) ORG_fenetre = I; Pos_ORG++; if(Pos_ORG == 57) Pos_ORG = 1;}}
  else {Pos_ORG = POS_TETE; Pos_ORG--; if(Pos_ORG == 0) Pos_ORG = 56;
      for (byte I=0; I<11; I++){
           if (Pos_ORG == Origine) ORG_fenetre = I; Pos_ORG++; if(Pos_ORG == 57) Pos_ORG = 1;}}
  Origine = SAVorg; // Restituer la valeur initiale.
  X = 4 + (ORG_fenetre * 12);}

void Affiche_le_BARILLET(){
  Recopie_les_pions_dans_la_fenetre();
  //-------------------------------------------------------------------------------------- 
  u8g.firstPage(); do {// Attention : Chaque affichage balaye huit fois cette séquence de type "do".  
  //---------------------- Affiche les lignes du plateau et la tête ----------------------
  u8g.drawHLine(0,25,128); u8g.drawHLine(0,33,128); u8g.drawRFrame(12,-3,9,13,3); u8g.drawHLine(12,0,9);
  //--------------------------- Affiche le petit E de la tête ----------------------------
  u8g.drawHLine(15,2,3); u8g.drawPixel(15,3); u8g.drawHLine(15,4,2); u8g.drawPixel(15,5); u8g.drawHLine(15,6,3);
  //-------------------------- Affiche les pions sous le plateau -------------------------
  for (byte X=2; X <130; X=X+12) u8g.drawRBox(X,34,5,3,0);
  //--------------------------------------------------------------------------------------
  Affiche_type_Ecran(); u8g.setPrintPos(18,62); u8g.print(Nb_Cycles_Horloge); u8g.print(" Cycles.");
  //---------------------------- Affiche les pions de la fenêtre -------------------------
  Pos_FENETRE = 0; // Laisser cette initialisation ici et pas en tête de procédure dans la déclaration.
  for (byte X=4; X <130; X=X+12) {Affiche_PION(X, FENETRE[Pos_FENETRE]); Pos_FENETRE++;}
  //------------------- Affiche l'origine si dans la FENETRE visualisée ------------------
  Determine_la_position_origine_dans_FENETRE();
  if (ORG_fenetre < 12) Affiche_Origine(X,29);
  //------------- Affiche les repères et le nombre de Pas pour G et D --------------------
  Affiche_les_reperes(); Affiche_NB_Pas();}
  //--------------------------------------------------------------------------------------
  while(u8g.nextPage()); Nb_Cycles_Horloge = Nb_Cycles_Horloge + 50000000; Duree_du_RUN = 7.7 * Nb_Cycles_Horloge;}

void Affiche_le_plateau() {
  boolean Fini = false;
  boolean Afficher = true; // Pour n'afficher la grille que si modification du contenu;
  boolean Affiche_la_grille = false; // En entrée on affiche le carrousel graphique.
  Nb_Pas = 1; 
  while (!Fini) {
        if (Afficher) {if (Affiche_la_grille) Affiche_la_GRILLE(); else Affiche_le_BARILLET(); Afficher = false;}
        //--------------- Traite la rotation du codeur incrémental -----------
        if (Rotation) {INT_active = true; Rotation = false; Affiche_la_grille = !Affiche_la_grille; Afficher = true;}
        //---------------------- Traite le B.P. central ----------------------
        Teste_activation_BP_central();
        if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}
        //--------------------- Traite le petit clavier ----------------------
        Tester_le_Clavier();
        switch (Num_BP) {
           case 1 : {if (!Affiche_la_grille) {T_MicroP = !T_MicroP; Afficher = true;} // Temps µP ou temps réel.
                     else BIP(); break;}
           case 2 : {if (Nb_Pas == 5) Nb_Pas = 1; else Nb_Pas = 5; Afficher = true; break;}
           case 3 : {Barillet_Initial = !Barillet_Initial; Afficher = true; break;} // Plateau ou barillet initial.
           case 4 : {for (byte I = 0; I < Nb_Pas; I++) Tourne_a_gauche(); Afficher = true; break;}
           case 5 : for (byte I = 0; I < Nb_Pas; I++) Tourne_a_droite(); Afficher = true; }}}

void Force_tous_les_BITs(byte BIT) { // Force tout le plateau à "B", "0" ou "1".
  for (byte I = 0; I < 56; I++) BARILLET_Initial[I] = BIT;}

void Aff_Etat(char BIT) {u8g.print(char(34)); u8g.print(BIT); u8g.print(char(34));}

void Sature_le_plateau() {
  boolean Fini = false; byte OCTET = 0;
  while (!Fini) {
     u8g.firstPage();
     do {if (OCTET != 3) {u8g.setPrintPos(3,38); u8g.print("Tout le plateau "); u8g.print(char(224)); SPC();}
     switch (OCTET) {
       case 0 : {Aff_Etat('B'); break;}
       case 1 : {Aff_Etat('0'); break;}
       case 2 : {Aff_Etat('1'); break;}
       case 3 : Afficher_un_texte(35, 38, 114, 5);}} while(u8g.nextPage());
        //--------------- Traite la rotation du codeur incrémental -----------
        if (Rotation) {if (Sens_positif) {OCTET--; if (OCTET == 255) OCTET = 3;}
                       else {OCTET++; if (OCTET == 4) OCTET = 0;} Rotation = false;}
        //---------------------- Traite le B.P. central ----------------------
        Teste_activation_BP_central();
        if (BP_central_a_traiter) {
            Fini = true; BP_central_a_traiter = false; if (OCTET < 3) Force_tous_les_BITs(OCTET);}}          
        //------------------------- Traite le petit clavier ------------------
  Affiche_la_GRILLE(); Attendre_un_BP();}

void Saisie_Positionne_la_TETE() {
  Saisir_un_nombre(4, 56); // 4 pour le texte "Position Tête : "
  Pos_tete = Nombre; if (Pos_tete == 0) Pos_tete = 1;}

void RAZ() {Saisir_OUI_NON(4); if (OUI) {Force_tous_les_BITs(0); Affiche_la_GRILLE(); Attendre_un_BP();}}

void Recopier_le_plateau_dans_le_BARILLET() {
  Saisir_OUI_NON(9);
  if (OUI) for (byte I=0; I<56; I++) BARILLET_Initial[I] = BARILLET[I];
  Affiche_la_GRILLE(); Attendre_un_BP();}

void MENU_des_OPTIONS() {
  Position_Index = 0;
  u8g.firstPage(); do Afficher_un_texte(35, 30, 220, 6); while(u8g.nextPage()); BIP();
  Machine_ETENDUE = !Machine_ETENDUE; 
  if (Machine_ETENDUE) {Num_PGM = 255; MAX = 20;} else {Num_PGM = 0; MAX = 11;} FINI = false;
  while (!FINI) {Traite_BP_central();
        if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}
  Position_Index = 0;}

void MENU_du_RUN() {
 // for (byte I=0; I<56; I++) BARILLET[I] = BARILLET_Initial[I];
  for (byte I=0; I<56; I++) {if (I<28) BARILLET[I] = 1; else BARILLET[I] = 2;}
 // Ci-avant : Charge un barillet machine provisoire. 
  BIP(); Nb_Cycles_Horloge = 0;}

void Construire_le_Motif() {
  byte BIT = 0; Saisir_OUI_NON(1);
  if (OUI) {for (byte I=0; I<56; I++) {BARILLET_Initial[I] = BIT; BIT++; if (BIT==3) BIT = 0;}
     Barillet_Initial = true; Affiche_la_GRILLE(); Attendre_un_BP();}}

void MENU_BARILLET() {
  Position_Index = 0; FINI = false;
  while (!FINI) {
  //------------------ Afficher le menu du BARILLET : ------------------
  u8g.firstPage();                       
  do {Afficher_un_texte(6, 11, 0, 18);   // "Affiche     PLATEAU"
      Afficher_un_texte(18, 24, 18, 16); // "SAUVE   RECHARGE"
      Afficher_un_texte(6, 36, 34, 18);  // "Arrange   Run >PLT"
      u8g.drawHLine(85,31,9); Afficher_un_texte(6, 49, 52, 18);  // "Origine   Pos.TETE"
      u8g.setPrintPos(30,63); Aff_TEXTE_EEPROM(70,3); u8g.setPrintPos(78,63); Aff_TEXTE_EEPROM(83,6); // "RAZ   Sature"
      Affiche_INDEX_10();} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) {TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9);}
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;
  switch (Position_Index) {
     case 0 : {Aff_pas = true; Barillet_Initial = true; Affiche_le_plateau(); Aff_pas = false; break;}
     case 1 : {Sauvegarde_le_BARILLET(); break;} // Sauvegarde le BARILLET.
     case 2 : {Aff_pas = true; Initialise_le_BARILLET_Initial(); Aff_pas = false; break;}
     case 3 : {Saisie_Origine(); break;}
     case 4 : {RAZ(); break;}
     case 5 : {Construire_le_Motif(); break;}
     case 6 : {Transfere_le_BARILLET(); break;} // Recharge le BARILLET.
     case 7 : {Recopier_le_plateau_dans_le_BARILLET(); break;}
     case 8 : {Saisie_Positionne_la_TETE(); break;}
     case 9 : Sature_le_plateau();}}}
  Position_Index = 0;}

void MENU_EEPROM() {
  boolean fini = false; Position_Index = 0; FINI = false;
  //------------------------ Indexer un algorithme -----------------------
  if (!Machine_ETENDUE) {
      //------------ Liste le contenu de l'EEPROM et indexe un fichier ------------
      while (!fini) {
         Liste_contenu_EEPROM(); if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(9);
         Teste_activation_BP_central(); if (BP_central_a_traiter) fini = true;}
      //---------------------------------------------------------------------------
      Emplacement_Cible = Position_Index; BP_central_a_traiter = false; Position_Index = 0;}
  while (!FINI) {
     u8g.firstPage(); do {
     Afficher_un_texte(40, 11, 89, 7);  // "CHARGER"
     Afficher_un_texte(40, 23, 96, 11); // "SAUVEGARDER"
     Afficher_un_texte(40, 35, 107, 7); // "EFFACER"
     Afficher_un_texte(40, 48, 114, 5); // "Fuite"
     //-------------------- Affiche le texte >>> ÉTENDU <<< ---------------
     if (Machine_ETENDUE) {u8g.setPrintPos(16,61); u8g.print(">>>"); Aff_TEXTE_EEPROM(219, 8); u8g.print("<<<");}
     //-------------------- Visualise l'index à sa position ---------------
     Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
     //--------------- Traite la rotation du codeur incrémental -----------
     if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(3);
     //---------------------- Traite le B.P. central ----------------------
     Teste_activation_BP_central();
     if (BP_central_a_traiter) {
         BP_central_a_traiter = false;
         switch (Position_Index) {
            case 0 : {Charger_un_emplacement(); break;}
            case 1 : {Sauvegarder_dans_un_emplacement(); break;}
            case 2 : {Effacer_un_PGM_en_EEPROM(); break;}
            case 3 : FINI = true;} FINI = true;}}
  Position_Index = 0;}

void MENU_PROGRAMME() {
  FINI = false; Position_Index = 0;  
  while (!FINI) {
    //-------------------- Afficher le MENU PROGRAMME --------------------
    u8g.firstPage(); do {
      Afficher_un_texte(30, 11, 126, 7); // "RÉDIGER"
      Afficher_un_texte(30, 23, 135, 8); // "MODIFIER"
      Afficher_un_texte(30, 35, 107, 7); // "EFFACER"
      Afficher_un_texte(30, 48, 143, 8); // "FORMATER"
      Afficher_un_texte(30, 60, 114, 5); // "Fuite"
      Afficher_un_texte(93, 27, 133, 3); // "PGM"
      AFF_Num_PGM(93,39);  u8g.drawRFrame(88,16,27,26,0);
      Positionne_INDEX(21, 11 + Position_Index * 12 );} while(u8g.nextPage());
    //--------------- Traite la rotation du codeur incrémental -----------
    if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
    //---------------------- Traite le B.P. central ----------------------
    Teste_activation_BP_central();
    if (BP_central_a_traiter) {
       BP_central_a_traiter = false; FINI = false;
       switch (Position_Index) {
          case 0 : {REDIGER_le_programme(); break;}
          case 1 : {MODIFIER_le_programme(); break;}
          case 2 : {Renommer = true; Effacer_le_programme(); break;}
          case 3 : {Modifier_le_PGM_pour_recherche_standard_de_la_donnee(); break;}}
          FINI = true;}}
  Position_Index = 0;}

void MENU_de_BASE() {
  //--------------------- Afficher le MENU de BASE ---------------------
  u8g.firstPage(); do {
    u8g.drawRFrame(0,0,128,64,10); AFF_Num_PGM(100, 16);
    Afficher_un_texte(40, 11, 151, 3); // "RUN"
    Afficher_un_texte(40, 23, 154, 8); // "BARILLET"
    Afficher_un_texte(40, 35, 162, 6); // "EEPROM"
    Afficher_un_texte(40, 47, 119, 7); // "OPTIONS"
    Afficher_un_texte(40, 59, 168, 9); // "PROGRAMME"
    Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL(4);
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {
     BP_central_a_traiter = false;
     switch (Position_Index) {
        case 0 : {MENU_du_RUN(); break;}
        case 1 : {MENU_BARILLET(); break;}
        case 2 : {MENU_EEPROM(); break;}
        case 3 : {MENU_des_OPTIONS(); break;}
        case 4 : {MENU_PROGRAMME();}}}
   //--------------------- Traite le petit clavier ----------------------
        Tester_le_Clavier();
        if (Num_BP>0) {Ecran_noir = !Ecran_noir;
                       if (Ecran_noir) {u8g.sleepOn(); Rouge(); Verte(); Bleue();}
                       else {u8g.sleepOff(); Eteindre_les_LEDs();}}}


