#include <EEPROM.h>


/*    >>>>> Programme sur lequel : 
 *          * Les affichages RUN ont été simplifiés ce qui libère 46 octets.
 *          * RAZ du BARILLET n'éffiche plus la GRILLE résultat. (Inutile et gain de place.)
 *          * Fonction "Vérifier les lignes non utilisées" enlevée. (Libère un Item dans Analyse. Gain : 258 octets.)
 *          * Ajouté la fonction "Permuter G et D". Il reste 190 octets.
 *          * Doublé le cadre de la référence du programme. Il reste 58 octets.
 *          * Rajouté deux accentués pour "Vérif ...". Il reste 34 octets.
 *          * Rajouté un accent sur "Increment". Il reste 20 octets.
 *          =========================================================================================
 *          * Le 20 mars 2023 après publication, les "sens de rotation" du codeur incrémental ont été
 *            inversés car A et B ont été permutés au soudage de la ligne de liaison.
 *          * La rotation du codeur incrémental a été revue en fonction des menus. Du coup il y a
 *            deux procédure, une pour le sens DIRECT et une pour le sens RETROGRADE.
 *          * Pour sortir de l'affichage du portrait il faut cliquer sur le clavier.
 *          * Les touches 1, 3 et 4 dans le MENU de BASE sont sans effet pour gagner de la place pour :
 *            Dans le MENU de BASE le nombre de trous est affiché si il est > 0, c'est à dire si un
 *            programme standard ou un programme étendu est présent en mémoire.
 *          
      
      >>>>>> ATTENTION : Avant de téléverser ce croquis il faut au préalable avoir utilisé P00 <<<<<<

      >>>>>> Quand le matériel sera définitif, optimiser les valeur de comparaison   <<<<<<
      >>>>>> dans la procédure Tester_le_Clavier() en utilisant Calibrage_des_CAN.   <<<<<<
 
 Programme ajoutant aux fonctions de P12 les OPTIONS suivantes :
   • Affichage en option le portrait à la mémoire d'Alan TURING.
   • Dans le MENU de BASE BP4 ne fait plus afficher la place qui reste en mémoire dynamique.

Branchements à effectuer :  Identique à ceux décrits dans le croquis du démonstrateur P09.
--------------------------
                                        PGM               RAM dynamique
 Compilateur Version 1.8.0        30606 Octets (99%)     1501 Octets (74%)  */

//--------------------- Déclarations pour le matériel -----------------------
#include "U8glib.h"
U8GLIB_SH1106_128X64 u8g(U8G_I2C_OPT_NONE); // Usage de la ligne I2C (Two Wire Interface)
#include <EEPROM.h>
//------------- Définition des broches d'entrées analogiques ----------------
#define Ligne_trois_touches 14 // A0 pour lire le clavier à 3 BP. 
#define Ligne_deux_touches 15  // A1 pour lire le clavier à 2 BP. 
#define BP_central 16          // A2 pour lire le bouton central du codeur rotatif.
#define Entree_Vmetre 21       // A7 pour mesurer la tension de l'accumulateur.

//-------------- Définition des broches de sorties Binaires -----------------
#define Codeur_A 2    // D2 : DT
#define Codeur_B 3    // D3 : CLK 
#define BUZZER 4      // D4 : BUZZER de type actif ou passif.
#define ROUGE 5       // Anode LED ROUGE.
#define VERT 6        // Anode LED Verte.
#define BLEU 7        // Anode LED bleue.

//------------------- Déclaration des constantes du programme ---------------
#define NB_tests_antirebonds 30 // Nombre de tests de stabilisation B.P.

#define Version "Rev 20-03-2023"   //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// L'image est une matrice de 64 lignes de 7 octets soit 3584 Pixels.
// Soit une taille de 448 octets.
byte VISAGE[] U8G_PROGMEM = { // Taille 56 x 30
     255,255,B01111111,0,B11100000,255,B01111111,255,255,B00001111,0,0,255,B01111111,255,255,B00000011,0,0,B11111100,
     B01111111,255,255,1,0,0,B11111000,B01111111,255,255,128,255,1,B11100000,B01111111,255,B01111111,B11000000,0,
     B00011110,B11100000,B01111111,255,B01111111,B11000000,0,B00011110,B11100000,B01111111,255,B00111111,B00110000,0,
     B01100000,B11000000,B01111111,255,B00011111,B00001000,0,0,B11000000,B01111111,255,B00001111,B00000110,B11111100,
     B00000011,128,B01111111,255,B10000111,B00000011,B10011111,B00001111,B11000000,B01111111,255,B11000111,B11000000,1,
     B00001100,B00100000,B01111111,255,3,B01100000,0,0,B00010000,B01111111,255,3,B00010001,0,0,B00010000,B01111110,
     255,B11000001,B00001000,B11110000,B00001111,B00001000,B01111100,255,B00110001,B00000110,00001110,0,B00000100,B01111100,
     255,B00001001,B10000010,1,0,B00000011,B01111100,255,B00000100,0,0,128,B01100001,B01111000,255,B00000100,0,0,128,
     B01100001,B01111000,255,B00000110,0,0,254,B01000011,B01111000,255,253,0,B11110000,255,B01000111,B01111000,255,
     B11100010,B00001111,B11100000,255,B01000111,B01111000,127,252,255,255,255,B10000111,B01111000,127,254,255,255,255,
     B10000111,B01111000,127,254,255,255,255,B10000111,B01111000,127,254,255,255,255,B10000111,B01111000,255,254,255,255,
     255,B10000111,B01111000,255,254,255,255,255,B10000111,B01111000,127,B11000000,255,255,255,B11000111,B01111000,
     B00111111,128,255,255,255,B01000011,B01111000,B00111111,0,255,255,255,B01000011,B01111000,B01011111,0,254,B00000011,
     255,B01100111,B01111000,B01101111,B00010000,B11111100,0,B11111100,B00101111,B01111100,B00101111,B00110110,B11111100,0,
     B11111100,B00001111,B01111100,B00101111,B00101110,B11111100,B11000001,B11111100,B00000111,B01111110,B00001111,B00001111,
     254,1,B11111000,B00000111,B01111110,B11001111,255,255,3,B11110000,B00000111,B01111101,B11011111,255,255,255,255,
     B11110111,B01111011,B10011111,255,255,255,255,B11110111,B01111011,B10011111,255,255,255,255,B11111011,B01111011,
     B10111111,255,255,255,255,B10000011,B01111011,B00111111,255,255,255,255,B10011011,B01111011,B00111111,B00111111,254,255,
     255,B10111011,B01111101,127,B00111110,B11000000,255,255,B11111001,B01111110,127,B01111110,B11110000,255,255,B11111101,
     B01111110,127,254,B11111100,255,255,B01111101,127,255,254,255,255,255,B10111100,127,255,254,255,255,255,B11001100,127,
     255,B11000110,255,255,255,B11110000,127,255,B00000110,254,255,255,254,127,255,B00001110,B11000000,255,127,255,127,     
     255,B10011101,B00111111,B11111000,B00111111,255,127,255,B00011101,254,B11111100,B01011111,255,127,255,B00111101,
     B00011000,255,B10101111,255,127,255,B11111011,B11100000,255,B10110011,255,127,255,B11110011,255,255,B00011001,255,127,
     255,B11100000,255,255,B00011110,B11111100,127,B00111111,B10011000,255,127,B01001111,B11110000,127,B01001111,B01110010,
     254,B10011111,B01000111,B11001000,127,B00110111,B11000110,B11111001,B11101111,B01100011,B10100000,127,B00011011,
     B00001110,B00000111,B11110000,B11100000,B01000000,B01111110,B00000101,B00011110,B11111100,B00011111,B01111000,0,
     B01111101,B00000010,B00111110,B11000000,B00000111,B01111100,0,B01111010,1,B01111110,0,0,B00111110,0,B01110100};

//----------------------- Déclaration des variables -------------------------
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif; // Sens de rotation.

char Etat;

boolean Aff_pas, Aff_Origine, BP_central_a_traiter, Barillet_Initial, Ecran_noir, F_existe, FINI, Grand_sauts,
        Infos_sur_RESET, Ignorer_non_FIN, Machine_ETENDUE, OUI, PAS_a_PAS, Place_possible, Ralenti, Renommer,
        Surbrillance, Retour_direct, T_MicroP, T_mP;

byte Emplacement_Cible, DEBUT, Etendue, FIN, Index, Nb_Pas, Nb_TRN, Num_BP, Num_emplacement, Num_PGM, Num_Tr, OCTET,
     OCTET1, OCTET2, ORG_fenetre, Origine, PION, Pos_grille, Pos_ORG, POS_TETE, Pos_tete, Pointeur, Pos_FENETRE,
     Position_Index, SAV_OCTET1, SAV_OCTET2, SAV_PION, SAV_Transition, Temps_reel, Theme, Transition, X, Y;

int Nb_Trous, PTR, Tension_lue, U_sur_A0, U_sur_A1;

unsigned long BORNE, Compteur_de_PAUSE, Increment, Nb_Cycles_Horloge, Nombre, T_microP, SURVEILLE;

float Debut_du_RUN, Duree_du_RUN, Duree_MicroP, Duree_pour_un_cycle, MAX;

byte Ref_en_EEPROM[10]; // Contenu de l'EEPROM.
byte TRANSITION[6]; // Contient les 6 octets de la transition pointée;
byte PROGRAMME[120]; // 120 au lieu de 66 pour compatibilité si Machine ETENDUE virtuelle à 20 transitions.
byte BARILLET_Initial[56]; // Si BIT == 0 : Etat "B"; Si BIT == 1 : Etat "0"; Si BIT == 2 : Etat "1";
byte BARILLET[56];
byte FENETRE[11];
boolean Utilisee[60]; // Pour compatibilité avec PGM standards et PGM ÉTENDUS.

volatile boolean Rotation;  // Indique une activation du codeur rotatif.
//---------------------------------------------------------------------------

void setup() {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  u8g.setFont(u8g_font_profont12); // Police de matrices de 6x10 pixels lettres MAJ/min comprises.
  pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW);
  pinMode(ROUGE, OUTPUT); pinMode(VERT, OUTPUT); pinMode(BLEU, OUTPUT); Eteindre_les_LEDs();
  digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A2.
  //-------------------------- Initialisation PLATEAU machine et INITIAL----------------------------
  for (byte I=0; I<56; I++) BARILLET[I] = 0; // Tous les pions du PLATEAU sont forcés à "B".
  for (byte I=0; I<56; I++) {if (I<28) BARILLET_Initial[I] = 1; else BARILLET_Initial[I] = 2;}
  //------------------------------- Initialisation des variables   ---------------------------------
  Aff_pas = false; Aff_Origine = true; Barillet_Initial = true; Booleens_en_mode_Aveugle(); BORNE = 0;
  BP_central_a_traiter = false; Duree_MicroP = 0; Ecran_noir = false; Grand_sauts = false; Ignorer_non_FIN = true;
  Machine_ETENDUE = false; Nb_Pas = 1; Nb_TRN = 11; Num_Tr = 1; Origine = 1; Pos_tete = 1;
  POS_TETE = 1; Surbrillance = true; T_MicroP = true; Position_Index = 0; Impose_standard();
  if (EEPROM.read(1021) & 2) Afficher_le_portrait();
  if (EEPROM.read(1021) & 1) Recharger_le_BARILLET();
  if (EEPROM.read(1023) == 255) {
     if (EEPROM.read(1022) == 255) {Machine_ETENDUE = true; MAX = 20; Etendue = 120; Tranfere_le_PGM_depuis_EEPROM();}
     else {Emplacement_Cible = EEPROM.read(1022); Tranfere_le_PGM_depuis_EEPROM();}}
  if (EEPROM.read(1021) & 4) {Num_emplacement =  EEPROM.read(1022); if (Num_emplacement == 255) Num_emplacement = 10;
     Affiche_Etat_RESET(); Attendre_un_BP(); Front_sur_A(); Front_sur_B();}
  if (Rotation) {INT_active = true; Rotation = false;}}

void loop() {MENU_de_BASE();}

//=========================================== Routines de servitude ===========================================

void Impose_standard() {Num_PGM = 0; MAX = 11; Etendue = 66;}

void Traite_BP_central() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {FINI = true;
  BP_central_a_traiter = false; Position_Index = 0;}} 

void Attendre_le_BPC() {FINI = false; while (!FINI) {Traite_BP_central();}}

void Afficher_le_portrait() {
  u8g.firstPage(); do {
  u8g.setPrintPos(60,17); u8g.print("Alan TURING"); u8g.setPrintPos(66,34); u8g.print("23 7 1912");
  u8g.setPrintPos(72,54); u8g.print("7 7 1954");
  u8g.drawXBMP(0, 0, 56, 64, VISAGE);}  // Portrait : 64 lignes de 7 octets.
  while(u8g.nextPage()); Attendre_un_BP(); }

void Affiche_la_version() {
  FINI = false;
  while (!FINI) {
     u8g.firstPage(); do {u8g.drawRFrame(0,0,128,64,10); u8g.drawRFrame(4,4,120,56,5);
     u8g.setPrintPos(11,25); u8g.print(Version);
     u8g.setPrintPos(21,44); u8g.print("Infos RUN :"); AFF_OUI_NON(EEPROM.read(1021) & 8);}
     while(u8g.nextPage());
     if (Rotation) {Rotation = false;
        OCTET = EEPROM.read(1021); if (OCTET & 8) OCTET = OCTET & 247; else OCTET = OCTET | 8; EEPROM.write(1021,OCTET);}
        Traite_BP_central();}}

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

void Eteindre_OLED() {u8g.firstPage(); do {} while(u8g.nextPage());}

void AFF_0() {u8g.print('0');}

void SPC() {u8g.print(' ');}

void SPC_Interrogation() {u8g.print(" ?");}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {// Routines pour les affichages textes EEPROM.
    for (byte I=0; I<Longueur; I++) {u8g.print(char(EEPROM.read(PTR))); PTR++;}; }

void Afficher_un_texte(byte X, byte Y, int PTR, byte Longueur) {
  u8g.setPrintPos(X, Y); Aff_TEXTE_EEPROM(PTR,Longueur);}

void Utilisation_EEPROM() {// Analyse le contenu de l'EEPROM et met à jour Ref_en_EEPROM[10].
  PTR = 351; 
  for(byte I = 0; I < 10; I++) {Ref_en_EEPROM[I] = EEPROM.read(PTR); PTR = PTR + 67;}}

void Rouge() {digitalWrite(ROUGE, HIGH);}
void Verte() {digitalWrite(VERT, HIGH);}
void Bleue() {digitalWrite(BLEU, HIGH);}
void Eteindre_les_LEDs() {digitalWrite(ROUGE, LOW); digitalWrite(VERT, LOW); digitalWrite(BLEU, LOW);}

  //-------------------- Routines de servitude pour traiter le capteur incrémental -------------------

void Delay_antirebonds() {for(int T=0; T<1000; T++);} // Valeur de 100 expérimentale.
  
void Front_sur_A(){ // Interruption quand A change d'état avant B.
 if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
 if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }
  
void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
 
  //----------------------- Routines de servitude pour traiter le petit clavier ----------------------

void ATTENDRE_le_relacher() {
  byte COMPTEUR = 0; digitalWrite(VERT, HIGH); Delay_60();// Attendre_la_stabilisation du travail.
  digitalWrite(VERT, HIGH);
  while (COMPTEUR < NB_tests_antirebonds) {// Attendre_la_stabilisation du repos :
        if ((analogRead(Ligne_trois_touches) < 815) || (analogRead(Ligne_deux_touches) < 767)) COMPTEUR = 0;
        else COMPTEUR++;}
  digitalWrite(VERT, LOW);}

void Tester_le_Clavier() {// Les valeurs tiennent compte de VCC = 5V car fourni par la ligne USB.
  U_sur_A1 = 0; Num_BP = 0; // Signale aucun B.P. utilisé.
  U_sur_A0 = analogRead(Ligne_trois_touches);
  if (U_sur_A0 < 813) {U_sur_A0 = 0; Delay_60(); 
      for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_A0 = U_sur_A0 + analogRead(Ligne_trois_touches);
      U_sur_A0 = U_sur_A0 / NB_tests_antirebonds;
      if (U_sur_A0 < 815) {Num_BP = 1;
      if (U_sur_A0 < 465)  Num_BP = 2; if (U_sur_A0 < 161)  Num_BP = 3;}}
  U_sur_A1 = analogRead(Ligne_deux_touches);
  if (U_sur_A1 < 767) {U_sur_A1 = 0; Delay_60();
      for (byte I = 0; I < NB_tests_antirebonds; I++) U_sur_A1 = U_sur_A1 + analogRead(Ligne_deux_touches);
      U_sur_A1 = U_sur_A1 / NB_tests_antirebonds;
      if (U_sur_A1 < 767) {Num_BP = 4; if (U_sur_A1 < 255)  Num_BP = 5;}}
  if (Num_BP != 0) ATTENDRE_le_relacher();}

  //--------------------------------------------------------------------------------------------------
  
void Delay_60() {delay(60);}

void Delay_200() {delay(200);}

void Alerte() {BIP(); Delay_200(); BIP(); Delay_200(); BIP();}

void Attendre_un_BP() {Num_BP = 0;
   while (Num_BP == 0) {Verte(); Tester_le_Clavier(); Delay_60(); Eteindre_les_LEDs(); Delay_60();}}

void Pointe_la_transition_dans_le_programme() {Index = (Num_Tr - 1) * 6;}

void Calcule_coordonnees_Grille() {
  switch (Pos_grille) {
    case 1 : {X = 52; Y = 17; break;}
    case 2 : {X = 76; Y = 17; break;}
    case 3 : {X = 100; Y = 17; break;}
    case 4 : {X = 52; Y = 32; break;}
    case 5 : {X = 76; Y = 32; break;}
    case 6 : {X = 100; Y = 32; break;}
    case 7 : {X = 52; Y = 47; break;}
    case 8 : {X = 76; Y = 47; break;}
    case 9 : X = 100; Y = 47;}}

void Affiche_TRANSITION() {
  switch (Pos_grille) {
    case 3 : {OCTET = TRANSITION[1]; Valeur_Tr(); break;}
    case 6 : {OCTET = TRANSITION[3]; Valeur_Tr(); break;}
    case 9 : {OCTET = TRANSITION[5]; Valeur_Tr(); break;}}}
    
void Transcode_ROT() {
  switch (OCTET) {
    case 8 : {OCTET = 71; break;}
    case 16 : OCTET = 68;}}

void Affiche_ROTATION() {
  switch (Pos_grille) {
    case 2 : {OCTET = TRANSITION[0] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));} break;}
    case 5 : {OCTET = TRANSITION[2] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));} break;}
    case 8 : {OCTET = TRANSITION[4] & 24; if (OCTET > 0) {Transcode_ROT(); u8g.print(char(OCTET));}}}}

void Transcode_ECR() {
  switch (OCTET) {
    case 1 : {OCTET = 66; break;}
    case 2 : {OCTET = 48; break;}
    case 4 : OCTET = 49;}}

void Affiche_ECRITURE() {
  switch (Pos_grille) {
    case 1 : {OCTET = TRANSITION[0] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));} break;}
    case 4 : {OCTET = TRANSITION[2] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));} break;}
    case 7 : {OCTET = TRANSITION[4] & 7; if (OCTET > 0) {Transcode_ECR(); u8g.print(char(OCTET));}}}}

void Valeur_Tr() {
  boolean FIN = false;
  if (OCTET > 0) {
      if (OCTET == 21) {u8g.print('F'); FIN = true;}
  if (!FIN) {if (OCTET > 9) u8g.setPrintPos((X + 6),Y + 11); u8g.print(OCTET);}}}

void Affiche_Item(){
  switch (X) {
    case 52 : {Affiche_ECRITURE(); break;}
    case 76 : {Affiche_ROTATION(); break;}
    case 100 : Affiche_TRANSITION();}}

void Affiche_les_Items_de_la_transition() {
  byte SAV_Pos_grille = Pos_grille;
  //---------------- Affiche les 9 cellules dans la grille ----------------
  for (Pos_grille = 1; Pos_grille < 10; Pos_grille++) {
      Calcule_coordonnees_Grille(); u8g.setPrintPos(X + 9, Y + 11); Affiche_Item();}
  //-----------------------------------------------------------------------
  Pos_grille = SAV_Pos_grille;}

void Aff_Grille_transition() {
  u8g.firstPage();
  do {u8g.drawRFrame(3,1,121,61,0); if(Surbrillance) {Calcule_coordonnees_Grille(); u8g.drawRFrame(X,Y,23,14,0);}
      Affiche_les_Items_de_la_transition();
      u8g.drawHLine(4,16,119); u8g.drawHLine(27,31,96); u8g.drawHLine(27,46,96);
      u8g.drawVLine(27,17,44); for (byte Y = 51; Y<100; Y = Y+24) u8g.drawVLine(Y,2,59);
      u8g.setPrintPos(22,13); u8g.print("TR"); u8g.setPrintPos(61,13); u8g.print('E'); u8g.setPrintPos(85,13);
      u8g.print('R'); u8g.setPrintPos(109,13); u8g.print('T'); u8g.setPrintPos(37,28); u8g.print('1');
      u8g.setPrintPos(37,43); AFF_0(); u8g.setPrintPos(37,58); u8g.print('B');
      u8g.setPrintPos(10,42); if (Num_Tr < 10) AFF_0(); u8g.print(Num_Tr);} while(u8g.nextPage());}

void Pointer_algorithme_cible() {PTR = 351 + ((Emplacement_Cible) * 67);}

void AFF_OUI_NON(boolean OUI) {SPC(); if (OUI) Aff_TEXTE_EEPROM(73,3); else Aff_TEXTE_EEPROM(76,3);}

void le_PGM() {Aff_TEXTE_EEPROM(216,4); Aff_TEXTE_EEPROM(133,3);}

void le_PGM_Interrogation() {le_PGM(); SPC_Interrogation();}

void Aff_PAS_a_PAS() {u8g.print("Pas "); u8g.print(char(224)); u8g.print(" PAS :");}

void Saisir_OUI_NON(byte TEXTE) {
  boolean Fini = false; OUI = true;
  while (!Fini) {
    u8g.firstPage(); do {u8g.setPrintPos(10,38); // Position du texte par défaut vers le centre de l'écran.
    switch(TEXTE) {
      case 1 : {Afficher_un_texte(5, 38, 89, 7); Aff_TEXTE_EEPROM(216,4);
                Aff_TEXTE_EEPROM(13,5); break;} // "CHARGER le Motif "
      case 2 : {Afficher_un_texte(20, 38, 96, 11); break;} // "SAUVEGARDER"
      case 3 : {Afficher_un_texte(20, 38, 89, 7); break;} // "CHARGER"
      case 4 : {u8g.setPrintPos(22,30); u8g.print("RAZ : "); break;}
      // Il n'y a plus de case 5 suite à optimisation du PGM.
      case 6 : {Afficher_un_texte(10, 35, 184, 10); SPC(); // "Transferer "
                u8g.print(Ref_en_EEPROM[Emplacement_Cible]); break;}
      case 7 : {Afficher_un_texte(10, 35, 107, 7); le_PGM(); break;} // "EFFACER le PGM"
      case 8 : {Afficher_un_texte(10, 30, 143, 8); le_PGM(); break;} // "FORMATER le PGM ?"
      case 9 : {u8g.setPrintPos(3,38); u8g.print("PLT vers Initial"); break;}
      case 10 : {Aff_PAS_a_PAS(); break;}
      case 11 : {u8g.print("T R"); u8g.print(char(233)); u8g.print("el :"); break;}   
      case 12 : {u8g.print("Ralenti :"); break;}     
      case 13 : {u8g.print("V Maximale :"); break;}
      case 14 : {u8g.print("Mode "); Aff_TEXTE_EEPROM(220,7);u8g.print(':'); break;}
      case 15 : {u8g.print("Pg d'Accueil :"); break;}
      case 16 : u8g.print("Permuter G et D");} AFF_OUI_NON(OUI);} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) OUI = !OUI; INT_active = true; // Armer un nouveau comptage.
                Rotation = false; // Prise en compte du traitement.
  //------------- Traite le B.P. central du codeur incrémental ---------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false; if (OUI) FINI = true;}}}

void Modifier_le_PGM_pour_recherche_standard_de_la_donnee() {
  byte Debut_translation = 59; Etendue = 66; Nb_TRN = 11;
  if (Num_PGM == 0) Alerte();
  else {if (Machine_ETENDUE) {Debut_translation = 113; Etendue = 120; Nb_TRN = 20;};
        BIP(); Saisir_OUI_NON(8);
        if (OUI) {
            Verifier_place_possible();
            if (Place_possible) {
                //----------------- Décaler d'une transition les instructions -----------------             
                    for (byte I = Debut_translation; I > 0; I--) PROGRAMME[I + 6] = PROGRAMME[I];
                    PROGRAMME[7] = PROGRAMME[1]; PROGRAMME[6] = PROGRAMME[0];         
                //--------- Remplacer les trois premières lignes par la vectorisation ---------
                    PROGRAMME[0] = 0; PROGRAMME[1] = 2; PROGRAMME[2] = 0; PROGRAMME[3] = 2;
                    PROGRAMME[4] = 8; PROGRAMME[5] = 0;
                //------------------ Recalculer les adresse des translations ------------------
                for (byte I = 7; I < Etendue; I = I + 2) {
                     OCTET = PROGRAMME[I]; if((OCTET > 0) && (OCTET != 21))
                                              {OCTET++; PROGRAMME[I] = OCTET;}}}}}}

void Tranfere_le_PGM_depuis_EEPROM() {
  byte Nb = 67;
  if (Machine_ETENDUE) {PTR = 229; Nb = 121; Num_PGM = 255;}
  else {Pointer_algorithme_cible(); Num_PGM = EEPROM.read(PTR); PTR++;}
  Pointeur = 0; 
  for (byte I = 1; I < Nb; I++) {
  PROGRAMME[Pointeur] = EEPROM.read(PTR); PTR++; Pointeur++;}}
       
void Charger_un_emplacement() {
  if (Machine_ETENDUE) {Tranfere_le_PGM_depuis_EEPROM();}
  else {Utilisation_EEPROM(); // Analyse le contenu de l'EEPROM et met à jour Ref_en_EEPROM[10].
        if (Ref_en_EEPROM[Emplacement_Cible] == 0) Alerte();
        else {Saisir_OUI_NON(6); if (OUI) Tranfere_le_PGM_depuis_EEPROM();}}}

void Purge_la_zone_programme() {// Efface toute la zone y compris si mode Machine ÉTENDUE.
   for (byte I = 0; I < 119; I++) PROGRAMME[I] = 0;}

void Effacer_le_programme() {
  Saisir_OUI_NON(7); if (OUI){if (!Machine_ETENDUE && Renommer) Num_PGM = 0; Purge_la_zone_programme();}}
  
void Debut_Nom(byte X, byte Y, byte N) {Afficher_un_texte(X, Y, 181, 3); u8g.print(N);} // "Pgm"

void Positionne_INDEX(byte X, byte Y) {u8g.setPrintPos(X,Y); u8g.print('>');}

void Positionne_Curseur(byte X, byte Y) {u8g.setPrintPos(X,Y); u8g.print('<');}

void Affiche_INDEX_10() {
  byte X, Y;
  X = 52; if (Position_Index > 4) X = 117;
  Y = 11 + Position_Index * 13; if (Position_Index > 4) Y = 11 + (Position_Index-5) * 13;
  u8g.setPrintPos(X,Y); u8g.print('<');}

void TRAITE_LA_ROTATION_en_DIRECT(byte MAX) {
  INT_active = true;   Rotation = false; // Armer un nouveau comptage et Prise en compte du traitement.
  if (Sens_positif) {Position_Index++; if (Position_Index == (MAX+1)) Position_Index = 0;}
  else {if (Position_Index > 0) Position_Index--; else Position_Index = MAX;}}

void TRAITE_LA_ROTATION_en_RETROGRADE(byte MAX) {
  INT_active = true;   Rotation = false; // Armer un nouveau comptage et Prise en compte du traitement.
  if (!Sens_positif) {Position_Index++; if (Position_Index == (MAX+1)) Position_Index = 0;}
  else {if (Position_Index > 0) Position_Index--; else Position_Index = MAX;}}

void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     Rouge(); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     Eteindre_les_LEDs();
     // ===========================================================
     BP_central_a_traiter = true;}}

void Liste_contenu_EEPROM() {
  byte X,Y; Utilisation_EEPROM();
  // Afficher la liste :
  u8g.firstPage(); do {
     Y = 10; X = 0;
     for (byte I=0; I<10; I++) {
          if (I > 4) X = 66; if (I == 5) Y = 10;
          Debut_Nom(X,Y,I); if (Ref_en_EEPROM[I] > 0) {SPC(); if (Ref_en_EEPROM[I] < 10) AFF_0(); 
                               if (Ref_en_EEPROM[I] < 100) AFF_0(); u8g.print(Ref_en_EEPROM[I]);}
                            else {Aff_TEXTE_EEPROM(177, 4);}  Y = Y + 13;} Affiche_INDEX_10();}     
  while(u8g.nextPage());}
 
void AFF_Num_PGM(byte X, byte Y) {
  u8g.setPrintPos(X,Y); if (Num_PGM < 100) AFF_0(); if (Num_PGM < 10) AFF_0(); u8g.print(Num_PGM);}

void Modifie_Ecriture(){
  byte Octet;
  byte Cible = 0; if (Pos_grille == 4) Cible = 2; if (Pos_grille == 7) Cible = 4;
  Octet = TRANSITION[Cible] & 7; TRANSITION[Cible] = TRANSITION[Cible] & 24;
  if (Sens_positif)
        switch (Octet) {    
          case 0 : {Octet = 4; break;}
          case 1 : {Octet = 0; break;}
          case 2 : {Octet = 1; break;}
          case 4 : Octet = 2;}
  else {switch (Octet) {
          case 0 : {Octet = 1; break;}
          case 1 : {Octet = 2; break;}
          case 2 : {Octet = 4; break;}
          case 4 : Octet = 0;}}
  TRANSITION[Cible] = TRANSITION[Cible] | Octet;}

void Modifie_Mouvement(){
  byte Octet;
  byte Cible = 0; if (Pos_grille == 5) Cible = 2; if (Pos_grille == 8) Cible = 4;
  Octet = TRANSITION[Cible] & 24; TRANSITION[Cible] = TRANSITION[Cible] & 7;
  switch (Octet) {
     case 0 : {Octet = 8; break;}
     case 8 : {Octet = 16; break;}
     case 16 : Octet = 0;}
  TRANSITION[Cible] = TRANSITION[Cible] | Octet;}

void Modifie_Transition(){
  byte Octet;
  byte Cible = 1; if (Pos_grille == 6) Cible = 3; if (Pos_grille == 9) Cible = 5; Octet = TRANSITION[Cible];
  if (Machine_ETENDUE) {
      if (Sens_positif) {if (Octet > 0) Octet--;  else Octet = 21;}
      else {Octet++; if (Octet == 22) Octet = 0;}}
  else {if (Sens_positif) {if (Octet == 21) Octet = 12; if (Octet > 0) Octet--; else Octet = 21;}
        else {Octet++; if (Octet == 22) Octet = 0; if(Octet == 12) Octet = 21;}}      
  TRANSITION[Cible] = Octet;}          

void Modifier_Item() {
  switch (Pos_grille) {
    case 1 : {Modifie_Ecriture(); break;}
    case 2 : {Modifie_Mouvement(); break;}
    case 3 : {Modifie_Transition(); break;}
    case 4 : {Modifie_Ecriture(); break;}
    case 5 : {Modifie_Mouvement(); break;}
    case 6 : {Modifie_Transition(); break;}
    case 7 : {Modifie_Ecriture(); break;}
    case 8 : {Modifie_Mouvement(); break;}
    case 9 : Modifie_Transition();}}

void Recopie_les_6_octets_de_la_transition() {
  Pointe_la_transition_dans_le_programme(); // Index = (Num_Tr - 1) * 6;
  for (byte I = 0; I < 6; I++) {TRANSITION[I] = PROGRAMME[Index]; Index++;}}

void Sauvegarde_les_6_octets_de_la_transition() {
  Pointe_la_transition_dans_le_programme(); // Index = (Num_Tr - 1) * 6;
  for (byte I = 0; I < 6; I++) {PROGRAMME[Index] = TRANSITION[I]; Index++;}}

void Saisir_Les_Items_dans_la_grille() {
   boolean Afficher = true; // Pour n'afficher la grille que si modification du contenu;
   boolean Fini = false; Surbrillance = true;
   Recopie_les_6_octets_de_la_transition(); // Pour travailler sur des données initiales.
   //--------------------------- Saisir les items dans la grille --------------------------
    Fini = false;
    while (!Fini) {if (Afficher) {Aff_Grille_transition(); Afficher = false;} 
       //------------- Traite le B.P. central du codeur incrémental ---------
       Teste_activation_BP_central();
       if (BP_central_a_traiter) {
           Fini = true; BP_central_a_traiter = false; Sauvegarde_les_6_octets_de_la_transition();}
       //--------------- Traite la rotation du codeur incrémental -----------
       if (Rotation) {Afficher = true;
          INT_active = true; Rotation = false; // Réarmer et prise en compte du traitement.
          Modifier_Item();}
       //------------------------- Traite le petit clavier ------------------
       Tester_le_Clavier(); if (Num_BP>0) Afficher = true; 
       switch (Num_BP) {
         case 1 : {Sauvegarde_les_6_octets_de_la_transition(); if (Num_Tr < MAX) {Num_Tr++;
                   Pos_grille = 1; Recopie_les_6_octets_de_la_transition();}
                   else Fini = true; break;}
         case 2 : {Surbrillance = !Surbrillance; break;}
         case 3 : {Sauvegarde_les_6_octets_de_la_transition(); if (Num_Tr > 1) Num_Tr--;
                   Pos_grille = 1; Recopie_les_6_octets_de_la_transition(); break;}
         case 4 : {Pos_grille--; if (Pos_grille == 0) Pos_grille = 9; break;}
         case 5 : {Pos_grille++; if (Pos_grille == 10) Pos_grille = 1;}}}}

void Saisir_Nombre(unsigned long MAX) {
  INT_active = true; // Armer un nouveau comptage.
  Rotation = false; // Prise en compte du traitement.
  if ((Sens_positif) && (Nombre >= Increment)) Nombre = Nombre - Increment; 
  else {Nombre = Nombre + Increment; if (Nombre > MAX) Nombre = 1;}}

void Saisir_un_nombre(byte TEXTE, unsigned long MAX) {
  boolean Fini = false; byte Coeficient = 1; Increment = 1; Nombre = 1;
  //------------------------ Saisir la transition de début d'édition ------------------------
  while (!Fini) {
    u8g.firstPage(); do {u8g.setPrintPos(10,18); u8g.print("Incr"); u8g.print(char(233)); u8g.print("ment = ");
    u8g.print(Increment); u8g.setPrintPos(10,38); // Position du texte par défaut vers le centre de l'écran.
    switch(TEXTE) {
      case 1 : {Aff_TEXTE_EEPROM(207,10); Aff_TEXTE_EEPROM(133,3);
                Aff_TEXTE_EEPROM(204,3); break;} // "Référence PGM : "
      case 2 : {u8g.print("Origine en : "); break;}
      case 3 : {u8g.setPrintPos(10,34); u8g.print("Commencer en");
                Afficher_un_texte(10,50,194,13); break;} // "Transition : "
      case 4 : {u8g.print("Position "); Aff_TEXTE_EEPROM(66,4); Aff_TEXTE_EEPROM(204,3); break;}
      case 5 : {u8g.print("BORNE en "); break;}
      case 6 : {u8g.print("JALONS = ");}}
    u8g.print(Nombre);}
    while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) Saisir_Nombre(MAX);
  //------------- Traite le B.P. central du codeur incrémental ---------
  Teste_activation_BP_central(); if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}
  //------------------------- Traite le petit clavier ------------------
  Tester_le_Clavier();
  switch (Num_BP) {
     case 1 : {Increment = 1; Coeficient = 1; break;}
     case 2 : {if (Grand_sauts) {Increment = 10; Coeficient = 1;}
               else {if ((TEXTE == 2) || (TEXTE == 4)) BIP(); else Fini = true;} break;}
     case 3 : {if ((TEXTE == 2) || (TEXTE == 4)) BIP(); else {Increment = 100; Coeficient = 1;} break;}
     case 4 : {Coeficient = 1; if (Grand_sauts) Increment = 1000; else Increment = 5; break;}
     case 5 : {if (Grand_sauts) switch (Coeficient) {
                                  case 1 : {Coeficient = 2; Increment = 10000; break;}
                                  case 2 : {Coeficient = 3; Increment = 100000; break;}
                                  case 3 : {Coeficient = 1; Increment = 1000000;}}
               else Increment = 10;}}}}

void REDIGER_le_programme() {
  Num_Tr = 1;   Pos_grille = 1;
  if (!Machine_ETENDUE) {
     Saisir_un_nombre(1,254); // Saisir la Référence du programme
     //---------------- Ne passe en EDITION que si la Référence est correcte ----------------
     if (Nombre > 0) {Num_PGM = Nombre; Renommer = false; Effacer_le_programme(); Saisir_Les_Items_dans_la_grille();}
     else Alerte();}
  else Saisir_Les_Items_dans_la_grille();}

void MODIFIER_le_programme() {
  byte MAX = 11; if (Machine_ETENDUE) MAX = 20;
  boolean Fini = false; Increment = 1; Nombre = 1; Pos_grille = 1;
  //------------------------ Saisir la transition de début d'édition ------------------
  Saisir_un_nombre(3, MAX); // 3 pour le texte "Transition : "
  //-----------------------------------------------------------------------
  Num_Tr = Nombre; if (Num_Tr == 0) Num_Tr = 1; Recopie_les_6_octets_de_la_transition();
  Saisir_Les_Items_dans_la_grille();}

void Liste_EEPROM_et_Attente_clavier() {Liste_contenu_EEPROM(); Attendre_un_BP();}

void Effacer_un_PGM_en_EEPROM() {
  if (Machine_ETENDUE) Alerte();
  else {u8g.firstPage(); do {Afficher_un_texte(20, 38, 107, 7); SPC(); Aff_TEXTE_EEPROM(181, 3);
        u8g.print(Emplacement_Cible); SPC_Interrogation();} // " EFFACER PgmN ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {Pointer_algorithme_cible(); EEPROM.write(PTR,0); Liste_EEPROM_et_Attente_clavier();}}}

void Sauvegarder_dans_un_emplacement() {
  byte Nb = 67; Pointeur = 0;
  if (Machine_ETENDUE) {
      u8g.firstPage(); do {u8g.setPrintPos(10,50); u8g.print("(Perte du "); Aff_TEXTE_EEPROM(154,8);u8g.print(".)");
      Afficher_un_texte(30, 20, 96, 11); Afficher_un_texte(42, 33, 220, 6); SPC_Interrogation();}
      while(u8g.nextPage());// "(Perte du BARILLET.)" et en dessous "SAUVEGARDER ÉTENDU ?"
      Attendre_un_BP();
      if (Num_BP == 3) {PTR = 229; Nb = 121; 
          for (byte I = 1; I < Nb; I++) {
               EEPROM.write(PTR,PROGRAMME[Pointeur]); PTR++; Pointeur++;}}}
  else {u8g.firstPage(); do {Afficher_un_texte(15, 38, 96, 11); SPC(); Aff_TEXTE_EEPROM(181, 3);
        u8g.print(Emplacement_Cible); SPC_Interrogation();} // " SAUVEGARDER PgmN ?"
        while(u8g.nextPage()); Attendre_un_BP();
        if (Num_BP == 3) {Pointer_algorithme_cible(); EEPROM.write(PTR,Num_PGM); PTR++;
           for (byte I = 1; I < Nb; I++) {EEPROM.write(PTR,PROGRAMME[Pointeur]); PTR++; Pointeur++;}
           Position_Index = 0; Liste_EEPROM_et_Attente_clavier();}}}

void Verifier_place_possible() {
  Place_possible = true;
  if (Machine_ETENDUE) {
      OCTET1 = PROGRAMME[114]; OCTET2 = PROGRAMME[115]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[116]; OCTET2 = PROGRAMME[117]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[118]; OCTET2 = PROGRAMME[119]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}   
  else {OCTET1 = PROGRAMME[60]; OCTET2 = PROGRAMME[61]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[62]; OCTET2 = PROGRAMME[63]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[64]; OCTET2 = PROGRAMME[65]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}
  if (!Place_possible) {BIP(); u8g.firstPage(); do {
      u8g.setPrintPos(10,30); u8g.print("Place insuffisante.");} while(u8g.nextPage()); Attendre_un_BP();}}
  
//=============================== Routines de gestion de l'affichage du BARILLET ===============================
void Affiche_Origine(byte X, byte Y) {u8g.drawRFrame(X-2,Y-2,5,5,1); u8g.drawPixel(X,Y);}

void Affiche_Repere_5(byte X) {u8g.drawVLine(X,26,7);}

void Affiche_5(byte X, byte Y) {
  u8g.drawHLine(X-3,Y-2,3); u8g.drawHLine(X-3,Y,2); u8g.drawHLine(X-3,Y+2,2); u8g.drawPixel(X-3,Y-1);
  u8g.drawPixel(X-1,Y+1);}

void Affiche_1(byte X, byte Y) {
     u8g.drawVLine(X+1,Y-2,5); u8g.drawPixel(X,Y-1); u8g.drawPixel(X,Y+2); u8g.drawPixel(X+2,Y+2);}

void Affiche_0(byte X, byte Y) {
     u8g.drawVLine(X,Y-1,3); u8g.drawVLine(X+2,Y-1,3); u8g.drawPixel(X+1,Y+2); u8g.drawPixel(X+1,Y-2);}

void Affiche_Repere(byte X, byte Y, byte Valeur) {
  switch(Valeur) {
    case 1  : {Affiche_1(X,Y); break;}
    case 5  : {Affiche_Repere_5(X); break;}  
    case 10 : {Affiche_1(X-3,Y); Affiche_0(X+1,Y); break;}
    case 15 : {Affiche_Repere_5(X); break;}
    case 20 : {u8g.drawHLine(X-3,Y-2,2); u8g.drawHLine(X-3,Y+2,3); u8g.drawPixel(X-1,Y-1); u8g.drawPixel(X-2,Y);
               u8g.drawPixel(X-3,Y+1); Affiche_0(X+1,Y); break;}      
    case 25 : {Affiche_Repere_5(X); break;}
    case 30 : {u8g.drawHLine(X-3,Y-2,2); u8g.drawHLine(X-3,Y+2,2); u8g.drawPixel(X-1,Y-1);
               u8g.drawPixel(X-2,Y); u8g.drawPixel(X-1,Y+1); Affiche_0(X+1,Y); break;}
    case 35 : {Affiche_Repere_5(X); break;}
    case 40 : {u8g.drawVLine(X-1,Y-2,5); u8g.drawHLine(X-3,Y+1,4); u8g.drawPixel(X-3,Y);
               u8g.drawPixel(X-2,Y-1); Affiche_0(X+2, Y); break;}
    case 45 : {Affiche_Repere_5(X); break;}
    case 50 : {Affiche_5(X,Y); Affiche_0(X+1,Y); break;}
    case 55 : Affiche_Repere_5(X);}}

void Affiche_les_reperes() {
  byte Xorg = X;
  byte Pos_FENETRE = 0;
  byte Position = Pos_tete - 1; if (!Barillet_Initial) Position = POS_TETE - 1;
  for (byte X=4; X <130; X=X+12) {
      if (Position != Origine) Affiche_Repere(X,29,Position);
      Pos_FENETRE++; Position++; if (Position==57) Position = 1;}}

 void Afficher_le_temps_reel_pour_executer() {
  byte Cas_a_afficher = 3;
  float Temps_qui_reste, Quotient;
  int Annees = 0; byte Heures = 0; byte Minutes = 0; int Jours = 0;
  Temps_qui_reste = Duree_du_RUN;
  //---------------------- Déterminer si dépasse un AN soit 31536000 s ---------------------- 
  if (Temps_qui_reste > 31536000) {Quotient = Temps_qui_reste / 31536000; Annees = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 31536000);}
  //----------------------- Déterminer si dépasse un JOUR soit 86400 s ----------------------
  if (Temps_qui_reste > 86400) {
     Quotient = Temps_qui_reste / 86400; Jours = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 86400);}
  //---------------------- Déterminer si dépasse une HEURE soit 3600 s ----------------------
  if (Temps_qui_reste > 3600) {
     Quotient = Temps_qui_reste / 3600; Heures = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 3600);}
  //----------------------- Déterminer si dépasse une minute soit 60 s ----------------------
  if (Temps_qui_reste > 60) {
     Quotient = Temps_qui_reste / 60; Minutes = Quotient;
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 60);}
  //-----------------------------------------------------------------------------------------
  if (Annees > 0) Cas_a_afficher = 1;
  if ((Annees == 0) && (Jours > 0)) Cas_a_afficher = 2;
  switch(Cas_a_afficher) {
    case 1 : {u8g.print(Annees); u8g.print("A "); u8g.print(Jours); u8g.print('J'); break;}
    case 2 : {u8g.print(Jours); u8g.print("J "); u8g.print(Heures); u8g.print('H'); break;}
    case 3 : {if (Heures>0) {u8g.print(Heures); u8g.print("H ");} u8g.print(Minutes); u8g.print("' ");
              if (Duree_du_RUN > 60) {u8g.print(Temps_qui_reste,0); u8g.print('"');}}}}

void Affiche_PION(byte X, byte Etat) {
  if (Etat == 0) u8g.drawHLine(X-2,24,5);
  if (Etat == 1) u8g.drawRBox(X-2,18,5,7,0);
  if (Etat == 2) u8g.drawRBox(X-2,12,5,13,0);}

void Affiche_NB_Pas() {
  if (Aff_pas) {u8g.drawRFrame(1,53,7,9,0); if (Nb_Pas == 5) {Affiche_5(6,57);} else Affiche_Repere(3,57,1);}}

void Tourne_a_gauche() {// Fait tourner le PLATEAU Machine ou INITIAL en fonction de "Barillet_Initial".
  // Faire tourner le carrousel à gauche consiste à incrémenter la position de la Tête de L/E.
  if(Barillet_Initial) {Pos_tete++; if (Pos_tete == 57) Pos_tete = 1;} 
  else {POS_TETE++; if (POS_TETE == 57) POS_TETE = 1;}}

void Tourne_a_droite() {// Fait tourner le PLATEAU Machine ou INITIAL en fonction de "Barillet_Initial".
  // Faire tourner le carrousel à droite consiste à décrémenter la position de la Tête de L/E.
  if(Barillet_Initial) {Pos_tete--; if (Pos_tete == 0) Pos_tete = 56;}
  else {POS_TETE--; if (POS_TETE == 0) POS_TETE = 56;}}

void Affiche_la_GRILLE() {
  byte Position_Tete = Pos_tete; if (!Barillet_Initial) Position_Tete = POS_TETE;
  byte Index, X, Y ;
  //--------------------------------- Affiche la grille ----------------------------------
  u8g.firstPage(); do {
  u8g.drawRFrame(13,2,101,61,0);
  for (byte I= 4; I<63; I = I+8) u8g.drawHLine(16,I,95);
  for (byte I= 15; I<120; I = I+12) u8g.drawVLine(I,4,57);
  //------------------------------ Affiche la position tête ------------------------------
  Index = 0; X = 0; Y = 1;
  while (Index != Position_Tete) {Index++; X++; if (X == 9) {X = 1; Y++;}}
  X = (X * 12) + 3; Y = (Y * 8) - 4;
  u8g.drawHLine(X,0,13); u8g.drawHLine(X,1,13); u8g.drawVLine(12,Y,9); u8g.drawVLine(11,Y,9);
  //--------------------------------- Affiche l'origine ----------------------------------
  if (Aff_Origine) {// Ne pas afficher en mode "Arrange" le BARILLET.
      Index = 0; X = 0; Y = 1; // Barillet_Initial a déjà été pris en compte à la déclaration de Position_Tete.
      while (Index != Origine) {Index++; X++; if (X == 9) {X = 1; Y++;}}
      X = (X * 12) + 3; Y = (Y * 8) - 4;
      u8g.drawHLine(X,61,13); u8g.drawHLine(X,63,13);u8g.drawVLine(112,Y,9); u8g.drawVLine(114,Y,9);}
  //------------------------ Affiche les PIXELs des repères en 5 -------------------------
  u8g.drawPixel(69,8); u8g.drawPixel(93,16); u8g.drawPixel(21,32);
  u8g.drawPixel(45,40); u8g.drawPixel(69,48); u8g.drawPixel(93,56);
  //--------------------------- Affiche les repèrères décimaux ---------------------------
  Affiche_reperes_decimaux(); // N'afficher le repère que si le pion est à l'état "B".
  //----------------------------- Affiche l'état des 56 BITs -----------------------------
  X = 21; Y = 8;
  for (byte I = 0; I < 56; I++) {
       if (Barillet_Initial)
          switch (BARILLET_Initial[I]) {// Si valeur 0 : "B" donc ne rien afficher et valeur non prise en compte.
             case 1 :  {Affiche_BIT_0 (X,Y); break;}
             case 2 :   Affiche_BIT_1 (X,Y);}
       else switch (BARILLET[I]) {// Si valeur 0 : "B" donc ne rien afficher et valeur non prise en compte.
               case 1 :  {Affiche_BIT_0 (X,Y); break;}
               case 2 :   Affiche_BIT_1 (X,Y);}
       X = X + 12; if (X == 117) {X = 21; Y = Y + 8;}}
  //------------------------- Affiche le symbole Origine ou le P -------------------------
  u8g.drawRFrame(0,28,9,9,0);
  if (Barillet_Initial) Affiche_Origine(4,32);
  else {u8g.drawVLine(2,30,5); u8g.drawHLine(3,30,3); u8g.drawHLine(3,32,3); u8g.drawPixel(6,31);}
  //------------- Affiche les repères et le nombre de Pas pour G et D---------------------
  Affiche_NB_Pas();} while(u8g.nextPage());}

void Incrementer_Pos_Tete() {Pos_tete++; if (Pos_tete > 56) Pos_tete = 1;}

void Initialise_le_BARILLET_Initial() {
  boolean Afficher = true; // Pour n'afficher la grille que si modification du contenu;
  boolean Fini = false; byte SAV_tete = Pos_tete; Pos_tete = 1;
  Barillet_Initial = true; Aff_Origine = false;  Nb_Pas = 1;
  while (!Fini) {
        if (Afficher) {Affiche_la_GRILLE(); Afficher = false;}
        //--------------- Traite la rotation du codeur incrémental -----------
        if (Rotation) {Afficher = true;
           if (Sens_positif) {Pos_tete = Pos_tete - Nb_Pas; if ((Pos_tete > 200) || (Pos_tete == 0)) Pos_tete = 56;}
                       else {Pos_tete = Pos_tete + Nb_Pas; if (Pos_tete > 56) Pos_tete = 1;} Rotation = false;}
        //---------------------- Traite le B.P. central ----------------------
        Teste_activation_BP_central(); if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}           
        //------------------------- Traite le petit clavier ------------------
        Tester_le_Clavier();
        if (Num_BP > 0) 
            switch (Num_BP) {
              case 1 : {BARILLET_Initial[Pos_tete - 1] = 0; Incrementer_Pos_Tete(); Afficher = true; break;}
              case 2 : {BARILLET_Initial[Pos_tete - 1] = 1; Incrementer_Pos_Tete(); Afficher = true; break;}
              case 3 : {BARILLET_Initial[Pos_tete - 1] = 2; Incrementer_Pos_Tete(); Afficher = true; break;}
              case 4 : {Nb_Pas = 1; Afficher = true; break;}
              case 5 : {Nb_Pas = 5; Afficher = true;}}}
  Pos_tete = SAV_tete; Aff_Origine = true;}

void Sauvegarder_le_BARILLET() {
  PTR = 295;
  for (byte I = 0; I < 56; I++) {EEPROM.write(PTR,BARILLET_Initial[I]); PTR++;}
  EEPROM.write(227,Pos_tete); EEPROM.write(228,Origine);}

void Recharger_le_BARILLET() {
  PTR = 295;
  for (byte I = 0; I < 56; I++) {BARILLET_Initial[I] = EEPROM.read(PTR); PTR++;}
  Pos_tete = EEPROM.read(227); Origine = EEPROM.read(228);}

void Sauvegarde_le_BARILLET() {
 if (Machine_ETENDUE) {BIP(); Attente_avec_texte(12,3);}
 else {Saisir_OUI_NON(2); if (OUI) Sauvegarder_le_BARILLET();}}
  
void Transfere_le_BARILLET() {
  if (Machine_ETENDUE) {BIP(); Attente_avec_texte(12,3);}
  else {Saisir_OUI_NON(3);
    if (OUI) {Recharger_le_BARILLET(); Barillet_Initial = true; Affiche_la_GRILLE(); Attendre_un_BP();}}}

void Saisie_Origine() {Saisir_un_nombre(2,56); Origine = Nombre;}

void Affiche_BIT_0(byte X, byte Y) {u8g.drawRFrame(X-2, Y-2,5,5,0);}

void Affiche_BIT_1(byte X, byte Y) {
  Affiche_BIT_0(X,Y); u8g.drawRFrame(X-3, Y-3,7,7,0); u8g.drawVLine(X-4,Y-3,7); u8g.drawVLine(X+4,Y-3,7);
  u8g.drawVLine(X-4,Y-3,7); u8g.drawVLine(X-5,Y-3,7); u8g.drawVLine(X+5,Y-3,7);}

void Affiche_reperes_decimaux() {
  // N'afficher le repère que si le pion est à l'état "B".
  if (Barillet_Initial) {if (BARILLET_Initial[0] == 0)Affiche_Repere(21, 8, 1);
     if (BARILLET_Initial[9] == 0) Affiche_Repere(33, 16, 10);
     if (BARILLET_Initial[19] == 0) Affiche_Repere(57, 24, 20);
     if (BARILLET_Initial[29] == 0) Affiche_Repere(81, 32, 30);
     if (BARILLET_Initial[39] == 0) Affiche_Repere(105, 40, 40);
     if (BARILLET_Initial[49] == 0) Affiche_Repere(33, 56, 50);}
  else {if (BARILLET[0] == 0)Affiche_Repere(21, 8, 1); if (BARILLET[9] == 0) Affiche_Repere(33, 16, 10);
     if (BARILLET[19] == 0) Affiche_Repere(57, 24, 20); if (BARILLET[29] == 0) Affiche_Repere(81, 32, 30);
     if (BARILLET[39] == 0) Affiche_Repere(105, 40, 40); if (BARILLET[49] == 0) Affiche_Repere(33, 56, 50);}}

// REMARQUE : Ci-dessous la séquence "Pos_FENETRE--; if (Pos_FENETRE == 0) Pos_FENETRE = 55;" est utilisée
//            trois fois. Pourtant faire une procédure consomme deux octets de plus.
void Recopie_les_pions_dans_la_fenetre() { 
  if (Barillet_Initial) {
       if(Pos_tete == 1) Pos_FENETRE = 55;
       else Pos_FENETRE = Pos_tete - 2;
       for (byte I=0; I<11; I++) {
            FENETRE[I] = BARILLET_Initial[Pos_FENETRE]; Pos_FENETRE++;
            if (Pos_FENETRE > 55) Pos_FENETRE = 0;}}
  else {if(POS_TETE == 1) Pos_FENETRE = 55;
       else Pos_FENETRE = POS_TETE - 2;
       for (byte I=0; I<11; I++) {
            FENETRE[I] = BARILLET[Pos_FENETRE]; Pos_FENETRE++;
            if (Pos_FENETRE > 55) Pos_FENETRE = 0;}}}

void Determine_la_position_origine_dans_FENETRE() {
  byte SAVorg = Origine;
  Pos_ORG = Pos_tete; 
  ORG_fenetre = 200; // Valeur arbitraire > 10;
  if (Barillet_Initial) {
      Pos_ORG = Pos_tete; Pos_ORG--; if(Pos_ORG == 0) Pos_ORG = 56;
      for (byte I=0; I<11; I++){
           if (Pos_ORG == Origine) ORG_fenetre = I; Pos_ORG++; if(Pos_ORG == 57) Pos_ORG = 1;}}
  else {Pos_ORG = POS_TETE; Pos_ORG--; if(Pos_ORG == 0) Pos_ORG = 56;
      for (byte I=0; I<11; I++){
           if (Pos_ORG == Origine) ORG_fenetre = I; Pos_ORG++; if(Pos_ORG == 57) Pos_ORG = 1;}}
  Origine = SAVorg; // Restituer la valeur initiale.
  X = 4 + (ORG_fenetre * 12);}

void Affiche_NB_cycles() {u8g.setPrintPos(18,62); u8g.print(Nb_Cycles_Horloge); u8g.print(" Cycles.");}

void Affiche_T_MicroP() {
  Aff_TEXTE_EEPROM(79,4); u8g.print(char(181)); u8g.print("P "); u8g.print(Duree_MicroP,1); u8g.print(" S");}

void Prepare_affichage_carrousel() {
  //---------------------- Affiche les lignes du plateau et la tête ----------------------
  u8g.drawHLine(0,25,128); u8g.drawHLine(0,33,128); u8g.drawRFrame(12,-3,9,13,3); u8g.drawHLine(12,0,9);
  //--------------------------- Affiche le petit E de la tête ----------------------------
  u8g.drawHLine(15,2,3); u8g.drawPixel(15,3); u8g.drawHLine(15,4,2); u8g.drawPixel(15,5); u8g.drawHLine(15,6,3);
  //-------------------------- Affiche les pions sous le plateau -------------------------
  for (byte X=2; X <130; X=X+12) u8g.drawRBox(X,34,5,3,0);
  //---------------------------- Affiche les pions de la fenêtre -------------------------
  Pos_FENETRE = 0; // Laisser cette initialisation ici et pas en tête de procédure dans la déclaration.
  for (byte X=4; X <130; X=X+12) {Affiche_PION(X, FENETRE[Pos_FENETRE]); Pos_FENETRE++;}}

void Affiche_le_BARILLET(){
  Recopie_les_pions_dans_la_fenetre();
  // Surtour placer l'instruction ci-dessus hors boucle d'affichage.
  //-------------------------------------------------------------------------------------- 
  u8g.firstPage(); do {// Attention : Chaque affichage balaye huit fois cette séquence de type "do".
  //--------------------------------------------------------------------------------------  
  Prepare_affichage_carrousel();
  //--------------------------------------------------------------------------------------
  u8g.setPrintPos(35,8);
  if (Barillet_Initial) u8g.print("INITIAL"); else Aff_TEXTE_EEPROM(0,7);
     Affiche_Origine(94,3); u8g.setPrintPos(101,8); u8g.print("= "); u8g.print(Origine); u8g.setPrintPos(15,49);
     if (T_MicroP) {Affiche_T_MicroP();}
     else {u8g.print("Tr"); u8g.print(char(233)); u8g.print("el "); Afficher_le_temps_reel_pour_executer();}
  Affiche_NB_cycles();
  //------------------- Affiche l'origine si dans la FENETRE visualisée ------------------
  Determine_la_position_origine_dans_FENETRE();
  if (ORG_fenetre < 12) Affiche_Origine(X,29);
  //------------- Affiche les repères et le nombre de Pas pour G et D --------------------
  Affiche_les_reperes(); Affiche_NB_Pas();}
  //--------------------------------------------------------------------------------------
  while(u8g.nextPage());}

void Affiche_le_plateau() {
  boolean Fini = false;
  boolean Afficher = true; // Pour n'afficher la grille que si modification du contenu;
  boolean Affiche_la_grille = false; // En entrée on affiche le carrousel graphique.
  Nb_Pas = 1; 
  while (!Fini) {
        if (Afficher) {if (Affiche_la_grille) Affiche_la_GRILLE(); else Affiche_le_BARILLET(); Afficher = false;}
        //--------------- Traite la rotation du codeur incrémental -----------
        if (Rotation) {INT_active = true; Rotation = false; Affiche_la_grille = !Affiche_la_grille; Afficher = true;}
        //---------------------- Traite le B.P. central ----------------------
        Teste_activation_BP_central();
        if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}
        //--------------------- Traite le petit clavier ----------------------
        Tester_le_Clavier();
        switch (Num_BP) {
           case 1 : {if (!Affiche_la_grille) {T_MicroP = !T_MicroP; Afficher = true;} // Temps µP ou temps réel.
                     else BIP(); break;}
           case 2 : {if (Nb_Pas == 5) Nb_Pas = 1; else Nb_Pas = 5; Afficher = true; break;}
           case 3 : {Barillet_Initial = !Barillet_Initial; Afficher = true; break;} // Plateau ou barillet initial.
           case 4 : {for (byte I = 0; I < Nb_Pas; I++) Tourne_a_gauche(); Afficher = true; break;}
           case 5 : for (byte I = 0; I < Nb_Pas; I++) Tourne_a_droite(); Afficher = true; }}}

void Force_tout_le_BARILLET_a_zero() {for (byte I = 0; I < 57; I++) BARILLET_Initial[I] = 0;}

void Force_la_plage_des_BITs(byte BIT) { // Force toute une ZONE du plateau à "B", "0" ou "1".
  for (byte I = (DEBUT - 1); I < FIN; I++) BARILLET_Initial[I] = BIT; }

void Aff_Etat(char BIT) {u8g.print(char(34)); u8g.print(BIT); u8g.print(char(34));}

void Sature_le_plateau() {
  byte Item = 1; boolean Fini = false; byte OCTET = 0; byte PAS = 1;
  DEBUT = 1; FIN = 56;
  while (!Fini) {
     u8g.firstPage();                                    
     do {if (OCTET != 3) { u8g.setPrintPos(90,23); u8g.print("P = "); u8g.print(PAS); u8g.drawRFrame(86,12,42,14,0);
     u8g.setPrintPos(3,15); u8g.print("DEBUT en : "); u8g.print(DEBUT);
     u8g.setPrintPos(15,31); u8g.print("FIN en : "); u8g.print(FIN);
     u8g.setPrintPos(3,47); u8g.print("Force les PIONS "); u8g.print(char(224)); SPC();}
     switch (OCTET) {
       case 0 : {Aff_Etat('B'); break;}
       case 1 : {Aff_Etat('0'); break;}
       case 2 : {Aff_Etat('1'); break;}
       case 3 : Afficher_un_texte(35, 38, 114, 5);}} while(u8g.nextPage());
        //--------------- Traite la rotation du codeur incrémental -----------
        if (Rotation) {INT_active = true; Rotation = false; // Réarmer et prise en compte du traitement.
           if (Item == 3) {if (Sens_positif) {DEBUT = DEBUT - PAS; if ((DEBUT == 0) || (DEBUT > 56)) DEBUT = 56;}
                          else {DEBUT = DEBUT + PAS; if (DEBUT > 56) DEBUT = 1;}}
           if (Item == 2) {if (Sens_positif) {FIN = FIN - PAS; if ((FIN == 0) || (FIN > 56)) FIN = 56;}
                           else {FIN = FIN + PAS; if (FIN > 56) FIN = 1;}}
           if (Item == 1) {if (Sens_positif) {OCTET--; if (OCTET == 255) OCTET = 3;}
                           else {OCTET++; if (OCTET == 4) OCTET = 0;} Rotation = false;}}
           if (DEBUT == FIN) {DEBUT = 1; FIN = 56;}
        //---------------------- Traite le B.P. central ----------------------
        Teste_activation_BP_central();
        if (BP_central_a_traiter) {
            Fini = true; BP_central_a_traiter = false; if (OCTET < 3) Force_la_plage_des_BITs(OCTET);}          
        //------------------------- Traite le petit clavier ------------------
        Tester_le_Clavier(); if (Num_BP>0) {
        switch (Num_BP) {
          case 1 : {Item = 1; break;}
          case 2 : {Item = 2; break;}
          case 3 : {Item = 3; break;}
          case 4 : {PAS = 1; break;}
          case 5 : PAS = 10;}}}
  Barillet_Initial = true; Affiche_la_GRILLE(); Sortie_par_le_BPC();}

void Saisie_Positionne_la_TETE() {
  Saisir_un_nombre(4, 56); // 4 pour le texte "Position Tête : "
  Pos_tete = Nombre; if (Pos_tete == 0) Pos_tete = 1;}

void Sortie_par_le_BPC() {boolean Fini = false;
  while (!Fini) {Teste_activation_BP_central(); if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;}}}

void RAZ() {boolean Fini = false; Saisir_OUI_NON(4); if (OUI) Force_tout_le_BARILLET_a_zero();}

void Recopier_le_plateau_dans_le_BARILLET() {
  boolean Fini = false; Saisir_OUI_NON(9);
  if (OUI) for (byte I=0; I<56; I++) BARILLET_Initial[I] = BARILLET[I]; Affiche_la_GRILLE(); Sortie_par_le_BPC();}

void Construire_le_Motif() {boolean Fini = false; byte BIT = 0; Saisir_OUI_NON(1);
  if (OUI) {for (byte I=0; I<56; I++) {BARILLET_Initial[I] = BIT; BIT++; if (BIT==3) BIT = 0;}
     Barillet_Initial = true; Affiche_la_GRILLE(); Sortie_par_le_BPC();}}

void Visualise_le_plateau() {
  Aff_pas = true; Affiche_le_plateau(); Aff_pas = false;}

//============================================ Traitement des OPTIONS =============================================

void Souligne_avec_egal(byte N) {for (byte I = 0; I<N; I++) u8g.print('=');}

void Option_PAUSE() {
  Saisir_OUI_NON(10); PAS_a_PAS = OUI; if (PAS_a_PAS) {Ralenti = false; Temps_reel = false;}}
  
void Option_BORNE() {
  Grand_sauts = true; Saisir_un_nombre(5,99999999); BORNE = Nombre; Grand_sauts = false; SURVEILLE = 0;}

void Option_JALONS() {
  Grand_sauts = true; Saisir_un_nombre(6,99999999); SURVEILLE = Nombre; Grand_sauts = false; BORNE = 0;}

void Option_Treel() {
  Saisir_OUI_NON(11); Temps_reel = OUI; if (Temps_reel) {Ralenti = false; PAS_a_PAS = false;}}
     
void Option_Ralenti() {
  Saisir_OUI_NON(12); Ralenti = OUI; if (Ralenti) {Temps_reel = false; PAS_a_PAS = false;}}

void Option_V_Maxi() {Saisir_OUI_NON(13); if (OUI) Booleens_en_mode_Aveugle();}

void Option_ETENDU() {
  Saisir_OUI_NON(14); Machine_ETENDUE = OUI; if (Machine_ETENDUE) {Num_PGM = 255; MAX = 20; Etendue = 120;}
                                             else {Impose_standard(); Purge_la_zone_programme();}}

void Affiche_Etat_RESET() {
    u8g.firstPage(); do { 
    u8g.setPrintPos(16,11); u8g.print("Sur RESET charge :"); u8g.setPrintPos(16,24); Souligne_avec_egal(18);
    u8g.setPrintPos(17,37); u8g.print("Le "); Aff_TEXTE_EEPROM(154,8); Aff_TEXTE_EEPROM(204,3);
    if (EEPROM.read(1021) & 1) Aff_TEXTE_EEPROM(73,3); else Aff_TEXTE_EEPROM(76,3);
    u8g.setPrintPos(17,50); u8g.print("Num "); Aff_TEXTE_EEPROM(168,9); SPC();
    if (Num_emplacement == 12) u8g.print("///");
    else {if (Num_emplacement == 11) Aff_TEXTE_EEPROM(76,3);
          else {if (Num_emplacement == 10) {u8g.print("20Tr"); u8g.setPrintPos(30,63);
                u8g.print("(PGM ETENDU.)");}
                else u8g.print(Num_emplacement);}}} while(u8g.nextPage());}
  
void Option_RESET() {
  boolean Fini = false; boolean Mdf_PGM = false;
  Num_emplacement = 12;
  while (!Fini) {
    Affiche_Etat_RESET();
    //--------------- Traite la rotation du codeur incrémental -----------
    if (Rotation) {INT_active = true; Rotation = false; // Réarmer et prise en compte du traitement.
       if (Sens_positif) {if (Num_emplacement > 0) Num_emplacement--; else Num_emplacement = 12;}
       else {Num_emplacement++; if (Num_emplacement == 13) Num_emplacement = 0; Mdf_PGM = true;}}
    //------------- Traite le B.P. central du codeur incrémental ---------
    Teste_activation_BP_central();
    if (BP_central_a_traiter) {Fini = true; BP_central_a_traiter = false;
       if ((Mdf_PGM) && (Num_emplacement != 12)) {
          if (Num_emplacement < 11) {
              EEPROM.write(1023,255);
              if (Num_emplacement == 10) {EEPROM.write(1022,255); EEPROM.write(1021,EEPROM.read(1021) & 254);}
              // Ci-dessus annule le rechargement d'un BARILLET éventuel car incompatible avec un PGM ÉTENDU.
              else EEPROM.write(1022,Num_emplacement);}
          else EEPROM.write(1023,0);}}    
    //------------------------- Traite le petit clavier ------------------
       Tester_le_Clavier(); if (Num_BP>0) {
       switch (Num_BP) {
         case 1 : {EEPROM.write(1021,EEPROM.read(1021) & 254); break;}
         case 2 : {EEPROM.write(1021,EEPROM.read(1021) & 254); break;}
         case 3 : {EEPROM.write(1021,EEPROM.read(1021) | 1); break;}
         case 4 : {EEPROM.write(1021,EEPROM.read(1021) & 254); break;}
         case 5 : {EEPROM.write(1021,EEPROM.read(1021) & 254);}}}}}

void Aff_titre() {u8g.setPrintPos(16,15); u8g.print("Saut sur la m"); u8g.print(char(234)); u8g.print("me");
    u8g.setPrintPos(34,28); u8g.print("TRANSITION"); u8g.setPrintPos(6,42); u8g.drawRFrame(0,0,128,36,3);}

void Analyse_les_sauts_de_TRANSITION_sur_elles_memes() {
  boolean Fini = false; byte Etendue = 66; if (Machine_ETENDUE) Etendue = 120; byte Pointeur; byte Transition = 1;
  byte Compteur = 0; boolean Present = false;
  if (Num_PGM == 0) {BIP(); Attente_avec_texte(35,1);}
  else {u8g.firstPage(); do {Aff_titre();} while(u8g.nextPage());
       while (!Fini) {
         for(byte Pointeur = 1; Pointeur < (Etendue + 1); Pointeur = Pointeur + 2) {
             Compteur++; if (Compteur == 4) {Compteur = 1; Transition++;}
             Fini = (Pointeur == (Etendue - 1));
             if (PROGRAMME[Pointeur] == Transition) {Present = true;
                u8g.firstPage(); do {Aff_titre(); u8g.setPrintPos(24,53); u8g.print("Appel en Tr"); u8g.print(Transition);
                u8g.print('-'); if (Compteur == 1) u8g.print('1'); if (Compteur == 2) u8g.print('0');
                if (Compteur == 3) u8g.print('B');} while(u8g.nextPage());
                Attendre_un_BP(); if (Num_BP == 3) {Fini = true; Pointeur = 119;}}}
         if ((!Present) && Fini) {u8g.firstPage(); do {Aff_titre();  u8g.setPrintPos(6,51);
         u8g.print("Pas d'appel suspect.");} while(u8g.nextPage()); Attendre_un_BP(); Num_BP = 0;}}}}

void Aff_titre_doublons() {u8g.setPrintPos(36,15); u8g.print("Doublons en"); u8g.setPrintPos(42,28);
  u8g.print(char(201)); u8g.print("CRITURE."); u8g.setPrintPos(6,42); u8g.drawRFrame(4,3,120,30,3);
  u8g.drawRFrame(0,0,128,36,5);}

void Analyse_les_ERCRITURES_en_doublon() {
  boolean Fini = false; byte Etendue = 65; if (Machine_ETENDUE) Etendue = 119; byte Pointeur; byte Transition = 1;
  byte Compteur = 0; boolean Present;
  if (Num_PGM == 0) {BIP(); Attente_avec_texte(35,1);}
  else {u8g.firstPage(); do {Aff_titre_doublons();} while(u8g.nextPage());
       while (!Fini) {
         for(byte Pointeur = 0; Pointeur < Etendue; Pointeur = Pointeur + 2) {
             Compteur++; if (Compteur == 4) {Compteur = 1; Transition++;};
             Fini = (Pointeur == (Etendue - 1)); OCTET = (PROGRAMME[Pointeur] & 7); Present = false;
             if ((OCTET == 1) && (Compteur == 3)) Present = true;
             if ((OCTET == 2) && (Compteur == 2)) Present = true;
             if ((OCTET == 4) && (Compteur == 1)) Present = true;
             if (Present) {
                u8g.firstPage(); do {Aff_titre_doublons(); u8g.setPrintPos(10,53); u8g.print("Doublon en Tr");
                u8g.print(Transition); u8g.print('-'); if (Compteur == 1) u8g.print('1');
                if (Compteur == 2) u8g.print('0'); if (Compteur == 3) u8g.print('B');} while(u8g.nextPage());
                Attendre_un_BP(); if (Num_BP == 3) {Fini = true; Pointeur = 119;}}
         if ((!Present) && Fini) {u8g.firstPage(); do {Aff_titre_doublons();  u8g.setPrintPos(12,51);
         u8g.print("ECR. inutile"); Aff_TEXTE_EEPROM(204,3); Aff_TEXTE_EEPROM(76,3);}
         while(u8g.nextPage()); Attendre_un_BP(); Num_BP = 0;}}}}}

void Inverser_les_Orientations_dans_le_PGM() {
  BIP(); if (Num_PGM == 0) Attente_avec_texte (35,1);
  else {Saisir_OUI_NON(16);
        if (OUI) {for (byte I = 0; I < Etendue; I = I + 2) {
                  OCTET = PROGRAMME[I] & 24;
                  if(OCTET > 0) {
                     if (OCTET == 8) PROGRAMME[I] = PROGRAMME[I] + 8; else PROGRAMME[I] = PROGRAMME[I] - 8;}}}}}

void Option_ANALYSE() {
  boolean Fini = false; Position_Index = 0;
  while (!Fini) {
     u8g.firstPage(); do {
     u8g.setPrintPos(10,11); u8g.print('V'); u8g.print(char(233)); u8g.print("rif NON F :"); AFF_OUI_NON(!Ignorer_non_FIN);
     u8g.setPrintPos(22,23); u8g.print("Permuter G et D");
     u8g.setPrintPos(10,35); u8g.print('V'); u8g.print(char(233)); u8g.print("rif ECR doublon");
     u8g.setPrintPos(10,47); u8g.print("Vf Tr / elle-m"); u8g.print(char(234)); u8g.print("me");
     u8g.setPrintPos(10,59); u8g.print("Info. RESET :"); AFF_OUI_NON(EEPROM.read(1021) & 4);
     Positionne_Curseur(116, 11 + Position_Index * 12 );} while(u8g.nextPage());
     //--------------- Traite la rotation du codeur incrémental -----------
     if (Rotation) TRAITE_LA_ROTATION_en_RETROGRADE(4);
     //---------------------- Traite le B.P. central ----------------------
     Teste_activation_BP_central(); if (BP_central_a_traiter) {BP_central_a_traiter = false; Fini = true;
        if (Position_Index == 1) Inverser_les_Orientations_dans_le_PGM();
        if (Position_Index == 2) Analyse_les_ERCRITURES_en_doublon();
        if (Position_Index == 3) Analyse_les_sauts_de_TRANSITION_sur_elles_memes();}
     //--------------------- Traite le petit clavier ----------------------
     Tester_le_Clavier();
     if (Num_BP > 0) {
        switch (Position_Index) {
            case 0 : {Ignorer_non_FIN = !Ignorer_non_FIN; break;}
            case 2 : {Analyse_les_ERCRITURES_en_doublon(); break;}
            case 3 : {Analyse_les_sauts_de_TRANSITION_sur_elles_memes(); break;}
            case 4 : {OCTET = EEPROM.read(1021) & 4;
                      if (OCTET == 4) EEPROM.write(1021, OCTET & 251); else EEPROM.write(1021, OCTET | 4);}}}}}

void Option_ACCUEIL() {Saisir_OUI_NON(15); OCTET = EEPROM.read(1021);
  if (OUI) EEPROM.write(1021, OCTET | 2); else EEPROM.write(1021, OCTET & 253);}

void Resumer_les_options() {
  byte Item = 0;
  boolean Fini = false; Verte(); Rouge();
  while (!Fini) {
    //------------------------------------------------------------------------------------------
    u8g.firstPage(); do {
    switch(Item) {
      case 0 : {u8g.setPrintPos(8,38); u8g.print("Retour MENU de BASE."); break;}
      case 1 : {u8g.setPrintPos(22,20); Aff_PAS_a_PAS();
                AFF_OUI_NON(PAS_a_PAS); u8g.setPrintPos(22,34); u8g.print("BORNE = ");
                if (BORNE == 0) Aff_TEXTE_EEPROM(76,3); else u8g.print(BORNE); u8g.setPrintPos(16,48);
                u8g.print("JALONS = "); if (SURVEILLE == 0) Aff_TEXTE_EEPROM(76,3); else u8g.print(SURVEILLE); break;}
      case 2 : {u8g.setPrintPos(28,20); u8g.print("Tmp R"); u8g.print(char(233)); u8g.print("el :");
                AFF_OUI_NON(Temps_reel); u8g.setPrintPos(34,34); u8g.print("Ralenti :"); AFF_OUI_NON(Ralenti);
                u8g.setPrintPos(16,48); u8g.print("V Maximale : ");
                if ((!PAS_a_PAS) && (!Temps_reel) && (!Ralenti)) Aff_TEXTE_EEPROM(73,3); else Aff_TEXTE_EEPROM(76,3); break;}
      case 3 : {u8g.setPrintPos(20,20); u8g.print("PGM"); Aff_TEXTE_EEPROM(219,7); Aff_TEXTE_EEPROM(204,2);
                AFF_OUI_NON(Machine_ETENDUE); u8g.setPrintPos(32,40); u8g.print("Portrait : ");
                if ((EEPROM.read(1021) & 2) > 0) u8g.print("OUI"); else  Aff_TEXTE_EEPROM(76,3); break;}
      case 4 : Affiche_Etat_RESET();}} while(u8g.nextPage());
    //------------------------------------------------------------------------------------------             
    if (Rotation) {INT_active = true; Rotation = false; // Réarmer et prise en compte du traitement.
       if (Sens_positif) {if (Item > 0) Item--; else Item = 4;}
       else {Item++; if (Item == 5) Item = 0;}}
    //------------------------------------------------------------------------------------------
    Teste_activation_BP_central();
    if (BP_central_a_traiter) {BP_central_a_traiter = false; Fini = true; FINI = (Item == 0);}}
    //------------------------------------------------------------------------------------------
  Eteindre_les_LEDs();}
                    
//===================================== Procédures du mode RUN =====================================                       
void Booleens_en_mode_Aveugle() {
  PAS_a_PAS = false; Ralenti = false; SURVEILLE = 0; Temps_reel = false; Theme = 0;}

void Attente_avec_texte(byte X, byte TEXTE) {
  u8g.firstPage(); do {u8g.setPrintPos(X,38); // Position du texte par défaut en hauteur au centre.
  switch(TEXTE) {
    case 1 : {u8g.print("Pas de PGM !"); break;}
    case 2 : {Aff_TEXTE_EEPROM(76,3); u8g.print(" F : Continuer ?"); break;}  // Ci-dessous : "Pas en PGM Étendu !".
    case 3 : u8g.print("Pas en "); Aff_TEXTE_EEPROM(133,3); Aff_TEXTE_EEPROM(219,7); u8g.print(" !");}}
  while(u8g.nextPage()); Attendre_un_BP();}

void Ecran_du_RUN() {
  if (EEPROM.read(1021) & 8) {
    u8g.firstPage(); do {
        u8g.setPrintPos(32,11); Aff_TEXTE_EEPROM(168,9); SPC(); u8g.print(Num_PGM);
        u8g.setPrintPos(44,23); u8g.print("BORNE :"); AFF_OUI_NON(BORNE > 0); u8g.setPrintPos(20,35);
        if (SURVEILLE > 0) {u8g.print("Surveille"); Aff_TEXTE_EEPROM(204,3); Aff_TEXTE_EEPROM(73,3);}
        else {Aff_PAS_a_PAS(); AFF_OUI_NON(PAS_a_PAS);}
        u8g.setPrintPos(44,47); u8g.print("Tr"); u8g.print(char(233)); u8g.print("el :"); AFF_OUI_NON(Temps_reel);
        u8g.setPrintPos(32,60); u8g.print("Ralenti :"); AFF_OUI_NON(Ralenti);}
    while(u8g.nextPage()); Attendre_un_BP();}}

void Marque_la_ligne_de_programme_comme_Utilisee() {
   byte Ligne = (Transition - 1) * 3; if (PION == 1) Ligne++; if (PION == 0) Ligne = Ligne + 2;
   Utilisee[Ligne] = true;} 

void Tester_la_presence_de_F() {
   if (!Ignorer_non_FIN) {
      F_existe = false;
      for(byte Pointeur = 1; Pointeur < (Etendue + 1); Pointeur = Pointeur + 2) {
          if (PROGRAMME[Pointeur] == 21) F_existe = true;}
      if (!F_existe) {BIP(); Attente_avec_texte(15,2); OUI = Num_BP == 3;}}}

void AFF_Ligne_Vide() {
  u8g.firstPage(); do {Afficher_un_texte(20, 23, 194, 11); u8g.print(Transition);
  u8g.setPrintPos(20,35); u8g.print("Ligne VIDE !"); u8g.setPrintPos(20,47); u8g.print("FIN du RUN.");}
  while(u8g.nextPage()); Alerte(); Attendre_un_BP();}

void Traite_le_Clavier_durant_le_RUN () {
  switch (Num_BP) {
     case  1 : {Ralenti = !Ralenti; if (Ralenti) {PAS_a_PAS = false; Temps_reel = false;} break;}
     case  2 : {T_mP = !T_mP; break;}
     case  3 : {Temps_reel = !Temps_reel; if (Temps_reel) {PAS_a_PAS = false; Ralenti = false;} break;}
     case  4 : {PAS_a_PAS = !PAS_a_PAS;
               if (PAS_a_PAS) {Ralenti = false; Temps_reel = false;} break;}
     case  5 : Booleens_en_mode_Aveugle(); Eteindre_OLED();}}

void Traiter_ECRITURE() {
  Etat = OCTET1 & 7;
  if (Etat == 1) {BARILLET[POS_TETE - 1] = 0; Duree_du_RUN = Duree_du_RUN + 1.8;
                  Duree_pour_un_cycle = Duree_pour_un_cycle + 1.8;}  // "B". 
  if (Etat == 2) {BARILLET[POS_TETE - 1] = 1; Duree_du_RUN = Duree_du_RUN + 1.65;
                  Duree_pour_un_cycle = Duree_pour_un_cycle + 1.65;} // "0".
  if (Etat == 4) {BARILLET[POS_TETE - 1] = 2; Duree_du_RUN = Duree_du_RUN + 1.4;
                  Duree_pour_un_cycle = Duree_pour_un_cycle + 1.4;}}  // "1".
  
void Traiter_la_ROTATION() {
  Etat = OCTET1 & 24;
  if (Etat == 8)  Tourne_a_gauche(); if (Etat == 16) Tourne_a_droite();
  //---------------------------------------------------------------------------------
  Duree_du_RUN = Duree_du_RUN + 0.27; Duree_pour_un_cycle = Duree_pour_un_cycle + 0.27;}

void Calcule_le_temps_pour_transiter() {
  byte Depart = Transition; byte Arrivee = OCTET2; byte NB_Transitions;
  if (Arrivee > Depart) NB_Transitions = Arrivee - Depart; else NB_Transitions = (11 - Depart) + Arrivee;
  Duree_du_RUN = Duree_du_RUN + (0.286 * NB_Transitions);
  Duree_pour_un_cycle = Duree_pour_un_cycle + (0.286 * NB_Transitions);}

void Affiche_NB_cycles_ou_TmP() {
  if (T_mP) {u8g.setPrintPos(18,62); Affiche_T_MicroP();} else Affiche_NB_cycles();}

void Affiche_la_fenetre() {
  Recopie_les_pions_dans_la_fenetre();
  Prepare_affichage_carrousel(); u8g.setPrintPos(35,8); Aff_TEXTE_EEPROM(0,7);
  //---------------------------- Affiche les pions de la fenêtre -------------------------
  Pos_FENETRE = 0; // Laisser cette initialisation ici et pas en tête de procédure dans la déclaration.
  for (byte X=4; X <130; X=X+12) {Affiche_PION(X, FENETRE[Pos_FENETRE]); Pos_FENETRE++;}
  //------------------- Affiche l'origine si dans la FENETRE visualisée ------------------
  Determine_la_position_origine_dans_FENETRE();
  if (ORG_fenetre < 12) Affiche_Origine(X,29);
  //--------------------------------------------------------------------------------------
  Affiche_les_reperes();}

void Aff_Etoile() {if (Etat == 0) u8g.print('*');}

void Affiche_instruction() {
  u8g.setPrintPos(18,49); u8g.print("Tr"); if(SAV_Transition < 10) u8g.print('0'); u8g.print(SAV_Transition); u8g.print('-');
  Etat = SAV_PION & 7; if (SAV_PION == 0) u8g.print('B'); if (PION == 1) u8g.print('0'); if (PION == 2) u8g.print('1');
  Aff_TEXTE_EEPROM(204,3); Etat = SAV_OCTET1 & 7; if (Etat == 1) u8g.print('B'); if (Etat == 2) u8g.print('0');
  if (Etat == 4) u8g.print('1'); Aff_Etoile(); SPC(); Etat = SAV_OCTET1 & 24; Aff_Etoile(); if (Etat == 8) u8g.print('G');
  if (Etat == 16) u8g.print('D'); SPC();
  if (SAV_OCTET2 > 0) {if (SAV_OCTET2 == 21) u8g.print('F'); else u8g.print(SAV_OCTET2);}
  else u8g.print('*');}

void Affiche_Ecran_du_RUN() {
  Recopie_les_pions_dans_la_fenetre(); Duree_MicroP = (millis() - Debut_du_RUN)/1000;
  // Surtour placer les instructions ci-dessus hors boucle d'affichage.
  //-------------------------------------------------------------------------------------- 
  if (Theme > 0) {
      u8g.firstPage(); do {
      if (PAS_a_PAS) {u8g.setPrintPos(95,8); u8g.print("PAUSE");}   
      switch (Theme) {
         case 1 : {Affiche_NB_cycles_ou_TmP(); Affiche_instruction(); Affiche_la_fenetre(); break;}
         case 2 : Affiche_la_GRILLE();}}
      while(u8g.nextPage());}}

void Teste_le_BPC() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;}}

void Traiter_le_programme() {
  FINI = false; // ATTENTION : Laisser cet identifiant en variable locale.
  Compteur_de_PAUSE = 0;
  while (!FINI) {// Ici commence un cycle d'HORLOGE.   
         Bleue(); delayMicroseconds(50); Eteindre_les_LEDs();
         //------------------ Traite les caractère frappés durant RUN --------------------
         Tester_le_Clavier(); if (Num_BP > 0) Traite_le_Clavier_durant_le_RUN();
         //-------------------------------------------------------------------------------
         if ((PAS_a_PAS) && (Nb_Cycles_Horloge>0)) {Num_BP = 0; Theme = 1; T_mP = false; Affiche_Ecran_du_RUN();
             while ((Num_BP == 0) && (!FINI)) {Teste_le_BPC(); if (FINI) Retour_direct = true; Tester_le_Clavier();
                   if (Num_BP == 4) PAS_a_PAS = false;} // Surveille le BPC et simultanément attend un BP du clavier.
             if (SURVEILLE > 0) PAS_a_PAS = false;}
         //------------ Si mode Surveillé fait croire à la frappe d'une PAUSE ------------
         Compteur_de_PAUSE++; if ((SURVEILLE > 0) && (Compteur_de_PAUSE == SURVEILLE)) {
                                 Compteur_de_PAUSE = 0; PAS_a_PAS = true;}
         //-------------------------------------------------------------------------------
         // Ci-dessous 2,53 Sec représente Un tour de la came d'HORLOGE et une LECTURE.
         Duree_du_RUN = Duree_du_RUN + 2.53; Duree_pour_un_cycle = 2.53;
         //--------------------------- Traite le B.P. central ----------------------------
         Teste_le_BPC();
         //------------------- Traite la rotation du codeur incrémental ------------------
         if (Rotation) {INT_active = true; Rotation = false; //Armer le comptage et prise en compte du traitement.
             if (Sens_positif) {if (Theme > 0) Theme--; else Theme = 2;}
             else {Theme++; if (Theme == 3) Theme = 0;}
             if (Theme == 0) Eteindre_OLED();}
                 // Ci-dessus l'affichage "de rien" permet d'effacer l'écran quand Theme == 0.
         //------------ LECTURE de l'état du PION situé sous la Tête de L/E. -------------
         PION = BARILLET[POS_TETE-1]; SAV_PION = PION; Nb_Cycles_Horloge++; // Laisser l'incrément Horloge ici.
         //-------------------- Pointer l'instruction dans la MATRICE --------------------
         Pointeur = (Transition - 1) * 6;
         if (PION == 1) Pointeur = Pointeur + 2; if (PION == 0) Pointeur = Pointeur + 4;
         Marque_la_ligne_de_programme_comme_Utilisee();
         //------------------------ Charge l'instruction pointée -------------------------
         OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur];
         SAV_OCTET1 = OCTET1; SAV_OCTET2 = OCTET2;
         //------------------- Vérifier si l'instruction n'est pas vide ------------------
         if ((OCTET1 == 0) && (OCTET2 == 0)) {AFF_Ligne_Vide(); FINI = true; Retour_direct = true;}
         if (!FINI) {
             Traiter_ECRITURE(); Traiter_la_ROTATION();
             //-------------------------- Traiter la TRANSITION --------------------------
             SAV_Transition = Transition;
             if (OCTET2 > 0) {if (OCTET2 == 21) {FINI = true; SURVEILLE = 0;} 
                              else {Calcule_le_temps_pour_transiter(); Transition = OCTET2;}}
         //-------------------------------------------------------------------------------
         if (Temps_reel) {Bleue(); Verte(); delay((int(Duree_pour_un_cycle) * 1000)); Eteindre_les_LEDs();}
         if (BORNE > 0) FINI = (Nb_Cycles_Horloge == BORNE);
         if (Ralenti) {for (byte I = 0; I < 12; I++) {
                       Bleue(); Rouge(); delay(16); Eteindre_les_LEDs(); delay(30);}}
         //-------------------------------------------------------------------------------
         Duree_MicroP = (millis() - Debut_du_RUN)/1000;
         Affiche_Ecran_du_RUN();}}} // Laisser après le test du clavier ou PAUSE ne
                                    // s'affichera pas en ouverture de PAS à PAS.

void RUN() {
  Retour_direct = false; Transition = 1; T_mP = false;
  for (byte I = 0; I < 60; I++) Utilisee[I] = false; // 60 est compatible avec PGM standard ou ÉTENDU.
  // NOTE : C'est uniquement dans Executer_le_PGM que peut se faire la vérification d'une ligne vide.
  //        quand elle est invoquée durant le déroulement du programme.
  if (Num_PGM == 0) {BIP(); Attente_avec_texte (35,1); Retour_direct = true;} // "Pas de PGM !".
  else {Duree_du_RUN = 0;
        //---------------------- Vérifie si le PGM contient au moins un "F" --------------------
        OUI = true; Tester_la_presence_de_F(); // Donc d'un 21 dans les transitions.
        if (OUI) {// Ici le booléen est représentatif d'un OUI à la question "On continue ?
            //----------------------------------------------------------------------------------
            Ecran_du_RUN();
            for (byte I=0; I<56; I++) BARILLET[I] = BARILLET_Initial[I]; POS_TETE = Pos_tete;
            //----------------------------------------------------------------------------------
            Debut_du_RUN = millis(); // Chronométrage pour la durée réelle d'exécution.
            Nb_Cycles_Horloge = 0; Theme = 1; Barillet_Initial = false; Traiter_le_programme();}
        else Retour_direct = true;}
        //--------------------------------------------------------------------------------------
  u8g.sleepOff(); Barillet_Initial = false; if (!Retour_direct) Visualise_le_plateau();}
  
//==================================================================================================

void Allume_eteint_ECRAN() {Ecran_noir = !Ecran_noir;
  if (Ecran_noir) {u8g.sleepOn(); Rouge(); Verte(); Bleue();} else {u8g.sleepOff(); Eteindre_les_LEDs();}}

void Calcule_le_nombre_de_trous() {
  byte Pointeur = 0; byte Nb_lignes = 33; if (Machine_ETENDUE) Nb_lignes = 60; Nb_Trous = 0;
  if (Num_PGM > 0) {
      for(byte Ligne = 0; Ligne < Nb_lignes; Ligne++) {
          OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur]; Pointeur++;
          if (OCTET1 & 7) Nb_Trous++; if (OCTET1 & 24) Nb_Trous++; if (OCTET2 > 0) Nb_Trous++;}}}

void MENU_des_OPTIONS() {
  Position_Index = 0; FINI = false;
  while (!FINI) {
  //------------------ Afficher le menu du BARILLET : ------------------
  u8g.firstPage(); do {
     u8g.setPrintPos(13,11); u8g.print("PAUSES    V Maxi.");
     u8g.setPrintPos(19,24); u8g.print("BORNE    "); Aff_TEXTE_EEPROM(219,7);
     u8g.setPrintPos(12,37); u8g.print("JALONS      RESET");
     u8g.setPrintPos(0,50); u8g.print("Tmp R"); u8g.print(char(233)); u8g.print("el "); u8g.print("   Analyse");     
     u8g.setPrintPos(6,63); u8g.print("Ralenti    Accueil"); Affiche_INDEX_10();} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) TRAITE_LA_ROTATION_en_RETROGRADE(9);
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {BP_central_a_traiter = false;
  switch (Position_Index) {
     case 0 : {Option_PAUSE(); break;}
     case 1 : {Option_BORNE(); break;}
     case 2 : {Option_JALONS(); break;}
     case 3 : {Option_Treel(); break;}
     case 4 : {Option_Ralenti(); break;}
     case 5 : {Option_V_Maxi(); break;}
     case 6 : {Option_ETENDU(); break;}
     case 7 : {Option_RESET(); break;}
     case 8 : {Option_ANALYSE(); break;}
     case 9 : Option_ACCUEIL();}
  Resumer_les_options();}}
  Position_Index = 0;}

void MENU_BARILLET() {
  Position_Index = 0; FINI = false;
  while (!FINI) {
  //------------------ Afficher le menu du BARILLET : ------------------
  u8g.firstPage();                       
  do {Afficher_un_texte(6, 11, 0, 18);   // "Affiche     PLATEAU"
      Afficher_un_texte(18, 24, 18, 16); // "SAUVE   RECHARGE"
      Afficher_un_texte(6, 36, 34, 18);  // "Arrange   Run >PLT"
      u8g.drawHLine(85,31,9); Afficher_un_texte(6, 49, 52, 18);  // "Origine   Pos.TETE"
      u8g.setPrintPos(30,63); Aff_TEXTE_EEPROM(70,3); u8g.setPrintPos(78,63); Aff_TEXTE_EEPROM(83,6); // "RAZ   Sature"
      Affiche_INDEX_10();} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) {TRAITE_LA_ROTATION_en_RETROGRADE(9);}
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {BP_central_a_traiter = false; FINI = true;
  switch (Position_Index) {
     case 0 : {Barillet_Initial = true; Visualise_le_plateau(); break;}
     case 1 : {Sauvegarde_le_BARILLET(); break;} // Sauvegarde le BARILLET.
     case 2 : {Aff_pas = true; Initialise_le_BARILLET_Initial(); Aff_pas = false; break;}
     case 3 : {Saisie_Origine(); break;}
     case 4 : {RAZ(); break;}
     case 5 : {Construire_le_Motif(); break;}
     case 6 : {Transfere_le_BARILLET(); break;} // Recharge le BARILLET.
     case 7 : {Recopier_le_plateau_dans_le_BARILLET(); break;}
     case 8 : {Saisie_Positionne_la_TETE(); break;}
     case 9 : Sature_le_plateau();}}}
  Position_Index = 0;}

void MENU_EEPROM() {
  boolean fini = false; Position_Index = 0; FINI = false;
  //------------------------ Indexer un algorithme -----------------------
  if (!Machine_ETENDUE) {
      //------------ Liste le contenu de l'EEPROM et indexe un fichier ------------
      while (!fini) {
         Liste_contenu_EEPROM(); if (Rotation) TRAITE_LA_ROTATION_en_RETROGRADE(9);
         Teste_activation_BP_central(); if (BP_central_a_traiter) fini = true;}
      //---------------------------------------------------------------------------
      Emplacement_Cible = Position_Index; BP_central_a_traiter = false; Position_Index = 0;}
  while (!FINI) {
     u8g.firstPage(); do {
     Afficher_un_texte(40, 11, 89, 7);  // "CHARGER"
     Afficher_un_texte(40, 23, 96, 11); // "SAUVEGARDER"
     Afficher_un_texte(40, 35, 107, 7); // "EFFACER"
     Afficher_un_texte(40, 48, 114, 5); // "Fuite"
     //-------------------- Affiche le texte >>> ÉTENDU <<< ---------------
     if (Machine_ETENDUE) {u8g.setPrintPos(16,61); u8g.print(">>>"); Aff_TEXTE_EEPROM(219, 8); u8g.print("<<<");}
     //-------------------- Visualise l'index à sa position ---------------
     Positionne_INDEX(30, 11 + Position_Index * 12 );} while(u8g.nextPage());
     //--------------- Traite la rotation du codeur incrémental -----------
     if (Rotation) TRAITE_LA_ROTATION_en_DIRECT(3);
     //---------------------- Traite le B.P. central ----------------------
     Teste_activation_BP_central();
     if (BP_central_a_traiter) {
         BP_central_a_traiter = false;
         switch (Position_Index) {
            case 0 : {Charger_un_emplacement(); break;}
            case 1 : {Sauvegarder_dans_un_emplacement(); break;}
            case 2 : {Effacer_un_PGM_en_EEPROM(); break;}
            case 3 : FINI = true;} FINI = true;}}
  Position_Index = 0;}

void MENU_PROGRAMME() {
  FINI = false; Position_Index = 0;
  while (!FINI) {
    //-------------------- Afficher le MENU PROGRAMME --------------------
    u8g.firstPage(); do {
      Afficher_un_texte(30, 11, 135, 8); // "MODIFIER"
      Afficher_un_texte(30, 23, 126, 7); // "RÉDIGER"
      Afficher_un_texte(30, 35, 107, 7); // "EFFACER"
      Afficher_un_texte(30, 48, 143, 8); // "FORMATER"
      Afficher_un_texte(30, 60, 114, 5); // "Fuite"
      Afficher_un_texte(93, 27, 133, 3); // "PGM"
      AFF_Num_PGM(93,39);  u8g.drawRFrame(88,16,27,26,0);
      Positionne_INDEX(21, 11 + Position_Index * 12 );} while(u8g.nextPage());
    //--------------- Traite la rotation du codeur incrémental -----------
    if (Rotation) TRAITE_LA_ROTATION_en_DIRECT(4);
    //---------------------- Traite le B.P. central ----------------------
    Teste_activation_BP_central();
    if (BP_central_a_traiter) {
       BP_central_a_traiter = false; FINI = false;
       switch (Position_Index) {
          case 0 : {MODIFIER_le_programme(); break;}
          case 1 : {REDIGER_le_programme(); break;}
          case 2 : {Renommer = true; Effacer_le_programme(); break;}
          case 3 : {Modifier_le_PGM_pour_recherche_standard_de_la_donnee(); break;}}
          FINI = true;}}
  Position_Index = 0;}

void MENU_de_BASE() {
  //--------------------- Afficher le MENU de BASE ---------------------
  Tension_lue = analogRead(Entree_Vmetre); // Mesure analogique [0 à 1023].
  Calcule_le_nombre_de_trous();
  u8g.firstPage(); do {
    u8g.setPrintPos(94,42); u8g.print((Tension_lue / 102.3),1); u8g.print('v');
    u8g.drawRFrame(0,0,128,64,10); AFF_Num_PGM(97, 21); u8g.drawRFrame(93,10,25,14,3); u8g.drawRFrame(91,8,29,18,5);
    Afficher_un_texte(30, 11, 151, 3); // "RUN"
    Afficher_un_texte(30, 23, 154, 8); // "BARILLET"
    Afficher_un_texte(30, 35, 162, 6); // "EEPROM"
    Afficher_un_texte(30, 47, 119, 7); // "OPTIONS"
    Afficher_un_texte(30, 59, 168, 9); // "PROGRAMME"
    if (Nb_Trous > 0) {u8g.setPrintPos(93,56); if (Nb_Trous > 0) u8g.print(Nb_Trous); u8g.print("Pr");}
    Positionne_INDEX(20, 11 + Position_Index * 12 );} while(u8g.nextPage());
  //--------------- Traite la rotation du codeur incrémental -----------
  if (Rotation) TRAITE_LA_ROTATION_en_DIRECT(4);
  //---------------------- Traite le B.P. central ----------------------
  Teste_activation_BP_central();
  if (BP_central_a_traiter) {
     BP_central_a_traiter = false;
     switch (Position_Index) {
        case 0 : {RUN(); break;}
        case 1 : {MENU_BARILLET(); break;}
        case 2 : {MENU_EEPROM(); break;}
        case 3 : {MENU_des_OPTIONS(); break;}
        case 4 : {MENU_PROGRAMME();}}}
   //--------------------- Traite le petit clavier ----------------------
   Tester_le_Clavier();
   switch (Num_BP) {
     case 2 : {Allume_eteint_ECRAN(); break;}
     case 5 : {Affiche_la_version();}}}

