//   Machine de Turing élémentaire. Aucune soudure à effectuer, sauf une sur D4 si vous désirez ajouter le
//   petit buzzer d'alerte sonore. (Prévu pour un bruiteur "actif") et une LED bleue sur la sortie D5 
//   qui s'allume durant le "RUN" et qui n'est pas vraiment indispensable.
//   L'intégralité du pilotage de cette matérialisation électronique d'une Machine de Turing se fait
//   avec le Moniteur de l'IDE d'Arduino.
//   Ce programme est compatible avec bien des versions de la série des modules Arduino, et en particulier
//   sur Arduino UNO et sur Arduino NANO.
  
//   >>>> Cette Machine virtuelle est une "fille" de la machine électromécanique dont le tutoriel de 
//   >>>> réalisation est disponible dans les "ouvrages" du site ROBOT MAKER.
//   >>>> De ce fait, la performance "combinatoire" sur RESET est celle de la machine matérielle :
//   >>>> Maximum 11 TRANSITIONS, et "Bande de papier" à 56 cases. (7 OCTETs)
//   >>>> BARILLET fait référence au plateau tournant sur la machine matérielle.
//   >>>> Le résultat visuel d'un "RUN" est identique à celui de la machine, c'est à dire que la tête de L/E
//   >>>> est immobile et c'est le BARILLET et ses repères qui se déplace sous la tête de L/E.
//   >>>> (D et G sont relatifs aux sens de décalage des pions sous la tête de L/E.)
//               
//   Branchements à effectuer :
//   --------------------------
//   Un petit BUZZER actif (Non impératif) est à brancher entre D4 et GND.
//   Une LED bleue avec 1kohm entre D5 et GND. 
//
//                                  PGM               RAM dynamique
//                          30720 Octets (100%)       840 Octets (41%)

#include <EEPROM.h>

//---------------- Définition des broches de sorties Binaires ---------------

#define BUZZER 4
#define LED_bleue 5
#define LED_Arduino 13

//------------------- Déclaration des constante du programme ----------------

#define CR 13 // Code du Retour Chariot qui vaut 13 codé en ASCII sous forme décimale.
#define Version_logiciel "> Version : Dimanche 18 Décembre 2022 <" //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Etoile '*' // Pour encadrer le résultat prévu à la compilation.

#define LGR_chaine 56 // Longueur maximale de la chaine de dialogue limitée par le buffeur du Moniteur série.

int Delai_voulu = 100; // Durée en mS éventuellement corrigé de la fréquence XTAL.
unsigned long Temps_Ecoule_depuis_RESET = 0;

//----------------------- Déclaration des variables -------------------------

char Caractere, Etat;

boolean Aff_Barillet_Initial, Accepte_CR_seul, Charger_le_Barillet, Commande, Decalage, F_existe, FINI,
        Format_Graphique, Fuite, Ignorer_les_lignes_NON_utilisees, Ignorer_non_FIN, Lister_a_chaque_cycle,
        Machine_ETENDUE, Mode_modifier, Options, PAS_a_PAS, Place_possible, Premier_O, Ralenti, Retour_MNU, Sauter,
        Surveille, Temps_reel, Visualiser_Barillet; 

byte BIT, Configuration_du_BARILLET, Etendue, Filtre, Index, LGR_traitee, Ligne_du_PGM, Longueur_du_TEXTE, Nb_TRN, Nb_Trous,
     Num_PGM, OCTET, OCTET1, OCTET2, Origine, ORIGINE, Pointeur, Pos_Tete, POS_TETE, Pos_Tete_Initiale, PION,
     Pos_BIT, Transition;
     
int Emplacement_Cible, PTR, PTR_source, SAV_Source;

float Debut_du_RUN, Duree_du_RUN, Duree_pour_un_cycle;

// Duree_du_RUN est la durée calculée en secondes pour effectuer le même traitement sur la Machine réelle.
// Bien que pouvant compter jusqu'à 10^38 le travail sur les réels en C+ d'Arduino ne fait que 6 à 7 chiffres
// significatifs, décimales comprises. (Et passer en double précision n'arrange pas ce fait. Donc, si la durée
// calculée dépasse 2 ANS, les valeurs des jours, des heures et des secondes ne sont plus fiables.

unsigned long BORNE, Compteur_de_PAUSE, NB_cycles_entre_PAUSES, Nb_Cycles_Horloge, Nombre_saisi;
  
char Tampon[LGR_chaine+1];
String Chaine_Memorisee; // Tampon mémoire pour la chaine entrante.

boolean Utilisee[60]; // Pour pouvoir tester sur un PGM Étendu.
byte PROGRAMME[120]; // 120 au lieu de 66 pour Machine ETENDUE virtuelle à 20 transitions.
byte BARILLET_Initial[56]; // Si BIT > 1 : Etat "B";
byte BARILLET[56];
byte Graduations_du_BARILLET[56];

char Modele_des_graduations[56] = {char(135), ' ', ' ', ' ', char(166), ' ', ' ', ' ', ' ', '1', ' ', ' ',
                                   ' ', ' ', char(166), ' ', ' ', ' ', ' ', '2', ' ', ' ', ' ', ' ', char(166),
                                   ' ', ' ', ' ', ' ', '3', ' ', ' ', ' ', ' ', char(166), ' ', ' ', ' ', ' ',
                                   '4', ' ', ' ', ' ', ' ', char(166), ' ', ' ', ' ', ' ', '5', ' ', ' ', ' ', 
                                   ' ', char(166), ' '};
char MODELE_des_GRADUATIONS[56];

char Caracteres_valides[] = {'B', 'G', 'D', 'T', 'F', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'};

//==================================================================================================

void setup() {Serial.begin(115200);
  Chaine_Memorisee.reserve(LGR_chaine+1); // Reserve LGR_chaine octets pour la chaine.
  pinMode(LED_Arduino, OUTPUT); pinMode(LED_bleue, OUTPUT); digitalWrite(LED_bleue, LOW); 
  pinMode(BUZZER, OUTPUT); digitalWrite(BUZZER, LOW); TEXTE_du_RESET();
  //------------------------------- Initialisation des variables   //-------------------------------
  Effacer_le_BARILLET_Initial(); INITIALISER_le_plateau_de_la_machine();  
  Accepte_CR_seul = false; BORNE = 0; Machine_ETENDUE = false; Etendue = 66; Nb_TRN = 11; Options = false;
  if (EEPROM.read(1023) == 1) {Nombre_saisi = EEPROM.read(1022); 
  if((Nombre_saisi == 14) && ( EEPROM.read(871) == 254)) {Machine_ETENDUE = true; Etendue = 120; Nb_TRN = 20;}
  Tranfere_le_PGM_depuis_EEPROM();
  if (Machine_ETENDUE) {
     Serial.println(); N_SPC(17); Serial.print(F("Remarque : C'est un programme de type Machine ETENDUE."));}
     if (EEPROM.read(1021) == 1) {
         Serial.println(); N_SPC(17); Serial.print(F("Configuration BARILLET ")); E_accent();
         Serial.print(F("galement charg")); E_accent(); Serial.print(F("e.")); Serial.println();
         Charger_un_Barillet_sur_RESET();}
     else Serial.println();}
  else Effacer_le_programme();
  Commande = true; Format_Graphique = false; Ignorer_les_lignes_NON_utilisees = true; Ignorer_non_FIN = true;
  Mode_modifier = false; Retour_MNU = true; Booleens_en_mode_Aveugle();
  //------------------------------------------------------------------------------------------------
  Aff_Barillet_Initial = true;  Afficher_le_BARILLET();}

void loop() {// Traite le MENU de BASE.
  while (Serial.available()) Caractere = Serial.read(); // Vider le tampon USB.
  Attendre_une_chaine_sur_RX(); Traite_la_ligne_du_MENU_de_BASE();} 

//======================================= Routines d'affichage =====================================

void TEXTE_du_RESET() {Serial.println(); Serial.println(); N_SPC(23); Serial.println(F(Version_logiciel)); 
  Serial.println(); MENU_de_BASE();
  if (EEPROM.read(1023) == 1) {Serial.println(); N_SPC(17);
     Serial.print(F("L'algorithme d'emplacement EEPROM num")); E_accent(); Serial.print(F("ro "));
     Serial.print(EEPROM.read(1022)); Serial.print(F(" est charg")); E_accent(); Serial.print('.');}}

// NOTE : D'une façon générale il n'est pas rentable en termes de taille occupée de créer des
//        procédures pour afficher des chaînes identiques plusieurs fois. Donc, quand l'occasion s'en
//        présente, tester la pertinence d'un sous-programme d'affichage.

void Aff_ATTENTION() {Serial.print(F("ATTENTION : "));}
void E_accent() {Serial.print(char(233));}
void A_accent() {Serial.print(char(224));}                     
void SPC() {Serial.print(' ');}
void Deux_SPC() {SPC(); SPC();}
void Fleche() {Serial.print(F(" -> "));}
void Affiche_cote_B_cote() {Serial.print(char(34)); Serial.print('B'); Serial.print(char(34));}
void Debut_cadre(byte N) {Serial.println(); N_SPC(8); for (byte I = 0; I < N; I++) Serial.print(Etoile);
                          Serial.println(); N_SPC(8); Serial.print(Etoile);}
void Fin_cadre(byte N) {Serial.println(Etoile); N_SPC(8); for (byte I = 0; I < N; I++) Serial.print(Etoile);}
void N_etoiles(byte Nb) {for (byte I = 0; I < Nb; I++) Serial.print(Etoile);}
void N_SPC(byte Nb) {for (byte I = 0; I < Nb; I++) Serial.print(' ');} 
void Retour_cadre() {Serial.println(Etoile); SPC(); Serial.print(Etoile); Serial.print(' ');}
void Recup_Etat() {Serial.print('R'); E_accent(); Serial.print(F("cup")); E_accent(); Serial.print(F("rer l'ETAT "));}
void Reserve_aux_programmeurs() {
  Serial.print(F("(R")); E_accent(); Serial.print(F("serv")); E_accent(); Serial.print(F(" aux programmeurs.)"));}
  
void MENU_de_BASE() {
  SPC(); N_etoiles(87); Retour_cadre();
  Serial.println(F("     Virgule ou ? : Rappel de ce MENU.     *     O : Menu des OPTIONS. (O ou Q.)     *"));
  SPC(); N_etoiles(87); Retour_cadre(); Serial.print(F("A : Affiche le BARILLET.")); N_SPC(19); 
  Serial.print(F("* C : Charger un PGM depuis l'EEPROM.     ")); Retour_cadre();
  Serial.print(F("B : Sauvegarder la config. du BARILLET.    * D : D")); E_accent();
  Serial.print(F("placer un PGM en EEPROM."));N_SPC(10); Retour_cadre();
  Serial.print(F("F : Forcer l'Etat sur un BIT du BARILLET.  * G : Gommer un Emplacement en EEPROM.    "));
  Retour_cadre(); Serial.print(F("I : Initialiser le BARILLET.")); N_SPC(15);
  Serial.print(F("* P : Purger totalement l'EEPROM.")); N_SPC(9); Retour_cadre();
  Serial.print(F("Z : ")); Recup_Etat();
  Serial.print(F("du Carrousel.         * S : Sauvegarder le PGM en EEPROM.       "));
  Retour_cadre(); Serial.print(F("  (Le Carrousel devient BARILLET initial)  * T : Transf")); E_accent();
  Serial.print(F("rer une config. du BARILLET. ")); Retour_cadre();
  Serial.print('Z'); E_accent(); Serial.print(F("ro : D")); E_accent();
  Serial.print(F("placer l'Origine sur le BARILLET. * U : Utilisation de l'EEPROM.")); N_SPC(12); Retour_cadre();
  Serial.print(F("> ou & : Rotation ")); A_accent(); Serial.print(F(" Droite du BARILLET.    ")); N_etoiles(42);
  Retour_cadre(); Serial.print(F("< : Rotation "));A_accent(); Serial.print(F(" gauche du BARILLET.")); N_SPC(9);
  Serial.print(F("* W : (Word) Listage du contenu EEPROM.   ")); Retour_cadre();
  Serial.print(F("'=' Repositionne la t")); Serial.print(char(234)); Serial.print(F("te de L/E.           *        "));
  Reserve_aux_programmeurs(); N_SPC(6); Serial.println('*');
  SPC(); N_etoiles(88); Serial.println(); Serial.print(F(" * E : Ecrire le programme.")); N_SPC(19);
  Serial.print(F("*  R : RUN.  (Ex")); E_accent(); Serial.print(F("cuter le programme.)     ")); Retour_cadre();
  Serial.print(F("K : (KILL) Effacer le programme.")); N_SPC(11);
  Serial.print(F("*      Options : G, L, P, Q, R, T et V.   ")); Retour_cadre();
  Serial.print(F("L : Lister le programme.")); N_SPC(19); Serial.print(F("*  V : Voir le r")); E_accent(); 
  Serial.print(F("sultat d'un traitement.  ")); Retour_cadre(); Serial.print(F("M : Modifier le programme."));
  N_SPC(17); N_etoiles(42); Retour_cadre();
  Serial.print(F("N : Nombre de trous dans la GRILLE.        *  Minuscules converties en MAJUSCULES.   "));
  Retour_cadre(); Serial.print(F("X : Changer le Num")); E_accent();
  Serial.print(F("ro de r")); E_accent(); Serial.print('f');  E_accent(); Serial.print(F("rence du PGM. *  R"));
  E_accent(); Serial.print(F("p. ")); A_accent(); Serial.print(F(" O/N autre que 'o' devient 'N'.  "));
  Serial.println('*'); SPC(); N_etoiles(88); Serial.println();}

void AFF_restrictions_du_mode_ETENDU() {
  Serial.println(); Serial.print(F("    ")); N_etoiles(33); Serial.println();
  Serial.print(F("    * Mode MACHINE ETENDUE ")); A_accent(); Serial.println(F(" 20 Tr. *"));
  Serial.println(F("    * Valides : C, B, S, T, U et W. *"));
  Serial.print(F("    "));N_etoiles(33); Serial.println(); Serial.println();}
  
//============================ Routines d'aide au développement du programme =======================
void Liste_le_contenu_EEPROM() {
  Serial.println(F("  Listage EEPROM en HEXADECIMAL."));
  PTR = 0; Lister_un_BLOC();Lister_un_BLOC(); Lister_un_BLOC();Lister_un_BLOC(); Revenir_au_MENU_de_BASE();}
  
void Lister_un_BLOC() {
  Lister_ENTETE();
  byte Compteur_lignes = 0;
  while (Compteur_lignes < 16)
    {Lister_une_LIGNE(); Compteur_lignes++; }}
  
void Lister_une_LIGNE() {   
  byte Compteur_Octets = 0; Serial.print(' ');
  if (PTR<10) {Serial.print('0');};
  if (PTR<100) {Serial.print('0');};
  if (PTR<1000) {Serial.print('0');};
  Serial.print(PTR); Serial.print("  ");  
  while (Compteur_Octets < 16) {
    Lister_OCTET(); Compteur_Octets++; }
  Serial.println(); }
  
void Lister_ENTETE() {
   Serial.println();
   Serial.println(" ADRS  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F");}
   
byte Lire_Octet_EEPROM(int ADRESSE) {return eeprom_read_byte((unsigned char*) ADRESSE);}

void Ecrire_Octet_EEPROM(int ADRESSE, byte Octet) {eeprom_write_byte((unsigned char*) ADRESSE, Octet);}
   
void Lister_OCTET() {
  byte octet = Lire_Octet_EEPROM(PTR);
  if (octet < 16) Serial.print('0'); 
  Serial.print(octet,HEX);
  Serial.print(' '); PTR++; }
  
//====================================== Routines de servitude =====================================

void Inverser_LED_Arduino() {digitalWrite(LED_Arduino,!digitalRead(LED_Arduino));}

void BIP() {
   for (int I=0; I < 700; I++) {
        digitalWrite(BUZZER, HIGH); delayMicroseconds(70); digitalWrite(BUZZER, LOW); delayMicroseconds(130);}}

// NOTE : la fonction tone() et noTone consomme à elle seule la bagatelle de 4% de mémoire de
//        programme soit plus de 1400 octets. Il donc de loin préférable à crer sa propre routine.

void Traite_erreur(byte Type_Erreur) {// Inutile de mettre "incorrecte." en procédure.
   Serial.println(); SPC(); Fleche(); BIP();
   switch (Type_Erreur) {
      case  1 : {Serial.println(F("Ligne vide.")); break;}
      case  2 : {Serial.println(F(" LGR > 1.")); break;}
      case  3 : {Serial.print(F("Caract")); Serial.print(char(232));
                 Serial.println(F("re NON valide.")); Ligne_du_PGM--; break;}
      case  4 : {Serial.print(F("Pas de PGM.")); break;}
      case  5 : {Serial.print(F("F ne termine pas la ligne.")); break;}
      case  6 : {Serial.print(F("Transition incorrecte.")); Ligne_du_PGM--; break;}
      case  7 : {Serial.print(F("ECRITURES contradictoires.")); Ligne_du_PGM--; break;}
      case  8 : {Serial.print(F("ROTATIONS contradictoires.")); Ligne_du_PGM--; break;}
      case  9 : {Serial.print(F("F avec une Transition.")); Ligne_du_PGM--; break;}
      case 10 : {Serial.print(F("Remarque : Ecriture surabondante.")); break;}
      case 11 : {Serial.print(F("Remarque : Saut sur la Transition actuelle.")); break;}
      case 12 : {Aff_ATTENTION(); Serial.print(F("Pas de F dans le PGM.")); break;}
      case 13 : {Serial.print(F("Numero de ligne incorrect."));}}}
      
void Attendre_une_chaine_sur_RX() {// Cette procédure attend une chaine sur la ligne USB.
  Longueur_du_TEXTE = 0; Chaine_Memorisee = ""; // Vider le tampon USB pour une nouvelle capture.
  Caractere = '-'; // Caractère arbitraire pour "effacer" l'ancien CR.
  Tampon[0] = '\0'; // Présise un tampon vide.
  if (Commande) {Serial.println(); Serial.print(F("  Commande")); Fleche();}
  else {if (Options) Serial.print(F("  Option")); Fleche();}
  //----------------------------------------------------------------------------------------
  while (Caractere != CR) {// CR : Retour Chariot qui sera généré par le Moniteur de l'IDE.
    //================= Chronométrage pour le délai écoulé ================
    if (millis() - Temps_Ecoule_depuis_RESET > Delai_voulu)  {
        Temps_Ecoule_depuis_RESET = millis(); Inverser_LED_Arduino();}
    //==================== Traite la saisie d'une chaîne ==================
    if (Serial.available()) {Caractere = Serial.read(); // Puiser un caractère dans le FIFO.
    if (Caractere == '&') Caractere = '>'; // transcode pour la rotation du BARILLET à droite.
    // Puis concaténer ce caractère sauf le CR :
    if ((Longueur_du_TEXTE < LGR_chaine) && (Caractere != CR)) {
       Chaine_Memorisee = Chaine_Memorisee + Caractere; Longueur_du_TEXTE++;}}}
       Chaine_Memorisee.toUpperCase(); // Forcer la chaine en MAJUSCULES.
       if(Chaine_Memorisee[0] == ',') Chaine_Memorisee[0] = '?';
  for (byte I = 0; I < Longueur_du_TEXTE; I++) Tampon[I] = Chaine_Memorisee[I];
  //---------------------------------------------------------------------------------------- 
  Chaine_Memorisee[LGR_chaine] = '\0';
  Tampon[Longueur_du_TEXTE] = '\0'; //" \0 " pour marquer la fin de chaîne.
  //--------------------------- Affiche la chaîne saisie -----------------------------------
  if (Longueur_du_TEXTE > 0) {Serial.print(F("[")); Serial.print(Tampon); Serial.println(F("]"));} 
  //----------------------------------------------------------------------------------------
  if ((Longueur_du_TEXTE == 0) && (!Accepte_CR_seul)) Traite_erreur(1); // Erreur si validation sur ligne vide.
  if (Longueur_du_TEXTE == LGR_chaine) {
      Debut_cadre(25); Serial.print(F(" Attention : LGR > 56. ")); Fin_cadre(25);}
  digitalWrite(LED_Arduino, LOW);}

void OUI_NON() {
  Attendre_une_chaine_sur_RX(); Caractere = Chaine_Memorisee[0];
  if (Caractere != 'O') Caractere = 'N';}

void Appel_a_OUI_NON() {Serial.print(F(" ? (O/N)")); Commande = false; OUI_NON();}

void Saisir_un_nombre() {
  Nombre_saisi = 0;
  while (Nombre_saisi == 0) {
         Attendre_une_chaine_sur_RX(); Nombre_saisi = atol(Tampon); if (Nombre_saisi == 0) BIP();}}

void Passe_a_la_ligne_suivante() {
  switch (Etat) {
     case  '1' : {Etat = '0'; break;}
     case  '0' : {Etat = 'B'; break;}
     case  'B' : {Etat = '1'; Transition++; break;}}}
     
void Erreur_de_syntaxe(byte Erreur) {Traite_erreur(Erreur);}

void Traiter_la_ligne_saisie() {
  // Ci-dessous Tampon[Index] ='0' remplace chaque caractère du début par un zéro
  // pour que Nombre_saisi = atoi(Tampon) puisse transformer le nombre de "T".
  if (Caractere == 'B') {if (OCTET1 > 0) {Erreur_de_syntaxe(7); OCTET1 = 0; Sauter = true;}
                         else {if (Etat == 'B') Traite_erreur(10);
                               OCTET1 = OCTET1 | 1; Tampon[Index] ='0';}}
  if (Caractere == '0') {if (OCTET1 > 0) {Erreur_de_syntaxe(7); OCTET1 = 0; Sauter = true;}
                         else {if (Etat == '0') Traite_erreur(10);
                               OCTET1 = OCTET1 | 2; Tampon[Index] ='0';}}
  if (Caractere == '1') {if (OCTET1 > 0) {Erreur_de_syntaxe(7); OCTET1 = 0; Sauter = true;}
                         else {if (Etat == '1') Traite_erreur(10);
                               OCTET1 = OCTET1 | 4; Tampon[Index] ='0';}}
  if (Caractere == 'G') {Filtre = OCTET1 & 16; if (Filtre > 0) {Erreur_de_syntaxe(8); OCTET1 = 0; Sauter = true;}
                                               else {OCTET1 = OCTET1 | 8; Tampon[Index] ='0';}}
  if (Caractere == 'D') {Filtre = OCTET1 & 8; if (Filtre > 0) {Erreur_de_syntaxe(8); OCTET1 = 0; Sauter = true;}
                                              else {OCTET1 = OCTET1 | 16; Tampon[Index] ='0';}}
  if (Caractere == 'T') {Tampon[Index] ='0'; Nombre_saisi = atoi(Tampon); LGR_traitee = 1;
                         // Ci-dessus LGR_traitee = 1 pour faire sortir suite au traitement de 'T'.
                         if ((Nombre_saisi < 1) || (Nombre_saisi > Nb_TRN)) {Erreur_de_syntaxe(6); Sauter = true;}
                         else {OCTET2 = Nombre_saisi;
                               if (Nombre_saisi == Transition) Traite_erreur(11);}}                        
  if (Caractere == 'F') {if ((Longueur_du_TEXTE - ++Index) > 0) Erreur_de_syntaxe(5); else OCTET2 = Nb_TRN + 1;}}

void Analyse_syntaxique() {
  boolean F_present = false; boolean T_present = false; Sauter = false; boolean VALIDE;
  LGR_traitee = Longueur_du_TEXTE; Index = 0; OCTET1 = 0; OCTET2 = 0;
  while (LGR_traitee > 0) { // Scanne caractère par caractère depuis "la gauche".
    Caractere = Chaine_Memorisee[Index]; VALIDE = false;
    //---------------------------- Vérifie la validité du caractère -----------------------------
    for (byte I=0; I < 15; I++) {
         if (Caractere == Caracteres_valides[I]) VALIDE = true;}     
    if (!VALIDE) {Traite_erreur(3); Sauter = true;} 
    //------------------------------ Vérifie T et F non simultanés ------------------------------
    for (byte I = 0; I < Longueur_du_TEXTE; I++) if(Chaine_Memorisee[I] == 'F') F_present = true;
    for (byte I = 0; I < Longueur_du_TEXTE; I++) if(Chaine_Memorisee[I] == 'T') T_present = true;
    if (F_present && T_present) {Erreur_de_syntaxe(9); Sauter = true; LGR_traitee = 1;}
    // Ci-dessus LGR_traitee = 1; une fois décrémentée plus avant fait sortir de la boucle d'analyse.
    //-------------------------------------------------------------------------------------------
    if (!Sauter) Traiter_la_ligne_saisie();
    LGR_traitee--; Index++;}}

void Affiche_PION() {
  switch (PION) {
     case  2 : {Serial.print('B'); break;}
     case  0 : {Serial.print('0'); break;}
     case  1 :  Serial.print('1');}}

void Revenir_au_MENU_de_BASE() {
  Commande = true; Accepte_CR_seul = false; Serial.println(); Serial.println(F("  Retour au MENU de BASE."));}

void Effacer_le_programme() {
  if (! Machine_ETENDUE) Num_PGM = 0; for(byte Pointeur = 0; Pointeur < Etendue; Pointeur++) PROGRAMME[Pointeur] = 0;}
  
void Sauvegarder_la_configiration_du_BARILLET() {
  boolean Existe; byte Placement;  Existe = false; int SAV_PTR;
  BIP(); Serial.println(); Serial.print(F("  Stocker la Configuration Barillet en EEPROM")); Appel_a_OUI_NON();
  if (Caractere == 'O') {
     //----------------------- Purger une configuration éventuelle -----------------------
       // Impératif, car un emplacement vide qui la précède serait utilisé pour sauvegarder
       // et on se trouverait avec deux configurations dans l'EEPROM.
       PTR = 0;
       for(byte I = 1; I < 16; I++) {
           OCTET = EEPROM.read(PTR); if (OCTET == 255) EEPROM.write(PTR, 0); Pointer_le_PGM_suivant();}
       //----------------- Rechercher le premier emplacement de programme libre ------------
       PTR = 0;
       for(byte I = 1; I < 16; I++) {
           OCTET = EEPROM.read(PTR); if (OCTET == 0) {Existe = true; Placement = I; I = 16; SAV_PTR = PTR;}
           Pointer_le_PGM_suivant();}
       //-----------------------------------------------------------------------------------
       if (Existe) {
           Serial.print(F("  BARILLET en emplacement "));
           Serial.print(Placement); Serial.print(F(" (PTR ")); Serial.print(SAV_PTR); SPC(); A_accent(); SPC();
           PTR = SAV_PTR; EEPROM.write(PTR, 255); PTR++;
           for (byte Bit = 0; Bit < 56; Bit++) {EEPROM.write(PTR, BARILLET_Initial[Bit]); PTR++;}
           EEPROM.write(PTR, Pos_Tete); PTR++; EEPROM.write(PTR, Origine); Serial.print(PTR); Serial.println(')');} 
       else { BIP(); Serial.println(); Serial.println(F("  Aucun emplacement disponible."));}}
       //-----------------------------------------------------------------------------------
  Retour_au_menu_de_BASE_avec_affichage_Utilisation_EEPROM();}

void Charger_un_Barillet_sur_RESET() {
  boolean Existe = false;
  //---------------------- Rechercher si une configuration existe ---------------------
  PTR = 0;
  for(byte I = 1; I < 16; I++) {
      OCTET = EEPROM.read(PTR);
      if (OCTET == 255) {Existe = true; I = 16;} else Pointer_le_PGM_suivant();}
  //-----------------------------------------------------------------------------------
  if (Existe) {
      PTR++; // Saute le 255 qui en num de PGM signale la présence d'une configuration de BARILLET.
      for (byte Bit = 0; Bit < 56; Bit++) {BARILLET_Initial[Bit] = EEPROM.read(PTR); PTR++;}
      Pos_Tete = EEPROM.read(PTR); PTR++; Origine = EEPROM.read(PTR);}}
  
void Translater_la_configuration_du_BARILLET() {
  boolean Existe = false; int SAV_PTR;
  //---------------------- Rechercher si une configuration existe ---------------------
  PTR = 0;
  for(byte I = 1; I < 16; I++) {
      OCTET = EEPROM.read(PTR);
      if (OCTET == 255) {Existe = true; I = 16;} else Pointer_le_PGM_suivant();}
  //-----------------------------------------------------------------------------------
  if (Existe) {SAV_PTR = PTR;
      PTR = SAV_PTR; Serial.println(); Serial.println(F("  Configuration du BARILLET par l'EEPROM."));
      PTR++; // Saute le 255 qui en num de PGM signale la présence d'une configuration de BARILLET.
      for (byte Bit = 0; Bit < 56; Bit++) {BARILLET_Initial[Bit] = EEPROM.read(PTR); PTR++;}
      Pos_Tete = EEPROM.read(PTR); PTR++; Origine = EEPROM.read(PTR); Retour_MNU = false;
      Aff_Barillet_Initial = true;  Afficher_le_BARILLET();}
  else {Serial.println(); Serial.print(F("Pas de config. BARILLET en EEPROM.")); BIP();} 
  //-----------------------------------------------------------------------------------
  Revenir_au_MENU_de_BASE();}
  
void Affiche_la_ligne_des_etats_des_pions(){
  Deux_SPC();
  if (Format_Graphique) {
     for (byte I=0; I < 56; I++) {
          if (Aff_Barillet_Initial) {
              if (BARILLET_Initial[I] == 1) Serial.print('*'); else Serial.print(' ');}
          else {switch (BARILLET[I]) {
                  case  1 : {Serial.print('*'); break;}
                  case  0 : {Serial.print(' '); break;}
                  case  2 : {Serial.print(' '); break;}}}}
     Serial.println();   Deux_SPC();
     for (byte I=0; I < 56; I++) {
          if (Aff_Barillet_Initial) {
              if (BARILLET_Initial[I] == 2) Serial.print('-'); else Serial.print('*');}
          else {if (BARILLET[I] == 2) Serial.print('-'); else Serial.print('*');}}
     Serial.println();}
  else {for (byte I=0; I < 56; I++) {
             if (Aff_Barillet_Initial) {
                 if (BARILLET_Initial[I] < 2) Serial.print(BARILLET_Initial[I]); else Serial.print('-');}
             else {if (BARILLET[I] < 2) Serial.print(BARILLET[I]); else Serial.print('-');}}
        Serial.println();}}
  
void Afficher_le_BARILLET() {
  boolean Jalon = false; char Marque; byte Compteur = 1; byte Dizaines = 0;
  Serial.println();
  //----------------------- Affiche la ligne des Etats des PIONS ----------------------
  Affiche_la_ligne_des_etats_des_pions(); Deux_SPC();
  //---------------------------- Trace la ligne des repères ---------------------------
  if (Aff_Barillet_Initial) {
      for (byte I=0; I < 56; I++) {
           if ((I == Origine) || (I == Pos_Tete)) {if (I == Pos_Tete) Serial.print(char(141));
                                                   else Serial.print(char(149));}
           else Serial.print(Modele_des_graduations[I]);} } //Serial.println();}
       //------------------------------------------------------------------------------
   else {for (byte I=0; I < 56; I++) {
              if ((I == ORIGINE) || (I == POS_TETE)) {if (I == POS_TETE) Serial.print(char(141));
                                                      else Serial.print(char(149));}
              else Serial.print(MODELE_des_GRADUATIONS[I]);} } //Serial.println();}

       //------------------------------------------------------------------------------
  Serial.println(); if (Retour_MNU) Revenir_au_MENU_de_BASE();}

void Voir_le_resultat_de_RUN() {
  Serial.println(); Serial.print(F("  R")); E_accent(); Serial.println(F("sultat d'un traitement :"));
  Aff_Barillet_Initial = false; Afficher_le_BARILLET();}
 
void Tranfere_le_PGM_depuis_EEPROM() {
  byte Nb = 34;
  if (Machine_ETENDUE) {PTR = 871; Nb = 61;} else PTR = (Nombre_saisi - 1) * 67;
  Pointeur = 0; Num_PGM = EEPROM.read(PTR);  PTR++;
  for (byte I = 1; I < Nb; I++) {
       PROGRAMME[Pointeur] = EEPROM.read(PTR); PTR++; Pointeur++;
       PROGRAMME[Pointeur] = EEPROM.read(PTR); PTR++; Pointeur++;}} 

void Charger_le_PGM() {
  if (Machine_ETENDUE) {
      Tranfere_le_PGM_depuis_EEPROM(); Serial.println(); Serial.print(F("  Programme ETENDU charg")); E_accent();
      Serial.println('.');}
  else {Retour_MNU = false; Utilisation_EEPROM(); Serial.println();
        Serial.print(F("  Charger l'emplacement EEPROM Num ? (Fuite si > 15.)")); Commande = false; Saisir_un_nombre();
        if (Nombre_saisi < 16) {
            PTR = (Nombre_saisi - 1) * 67; Pointeur = 0;
            if (EEPROM.read(PTR) == 0) {BIP(); Serial.print(F("  Emplacement vide."));}
            else {if (EEPROM.read(PTR) == 255) {
                      BIP(); Serial.print(F("  L'emplacement contient une configuration BARILLET."));}
                  else Tranfere_le_PGM_depuis_EEPROM();}}}
   Revenir_au_MENU_de_BASE();}

void Deplace_Emplacement() {for(byte I = 1; I < 68; I++) {
       OCTET = EEPROM.read(PTR_source); EEPROM.write(PTR, OCTET);
       PTR++; PTR_source++;} EEPROM.write(SAV_Source, 0); Fuite = false;}

void Traiter_le_Deplacement() {// ATTENTION : Ecrase l'emplacement cible.
  // Déplace sans vérifier si l'emplacement est vide.
  // Efface l'emplacement cible.
  FINI = false; // En entrée de procédure PTR pointe la cible. 
  if (SAV_Source == PTR) {
     Serial.println(); Serial.println(F("Cible = Source Interdit !"));BIP();}
  else {OCTET = EEPROM.read(PTR);
       if (OCTET == 255) {
           BIP(); Serial.println(F("L'emplacement cible est une configuration de BARILLET."));
           Serial.print(F("On persiste")); Appel_a_OUI_NON();
           if (Caractere == 'O') {FINI = true; Deplace_Emplacement(); EEPROM.write(1021,255);}FINI = true;}
       if (!FINI) {
          if (Emplacement_Cible == EEPROM.read(1022) && EEPROM.read(1023) == 1){   
              Serial.println(); BIP(); Serial.print(F("Emplacement utilis")); E_accent();
              Serial.println(F(" pour charger un PGM au RESET.")); Serial.print(F("On persiste"));
              Appel_a_OUI_NON(); if (Caractere == 'O') {FINI = true; Deplace_Emplacement();} FINI = true;}}
       if (!FINI) Deplace_Emplacement();}}

void Deplacer_un_programme() {
  Commande = false; Fuite = true;
  Retour_MNU = false; Utilisation_EEPROM(); Serial.println(); Serial.print(F("  Emplacement "));
  A_accent(); Serial.print(F(" d")); E_accent(); Serial.print(F("placer (Fuite si > 15.)")); Commande = false;
  Saisir_un_nombre();
  if (Nombre_saisi < 16) {
     PTR_source = (Nombre_saisi - 1) * 67;  SAV_Source = PTR_source;
     Serial.print(F("  Emplacement cible (Fuite si > 15.)")); Saisir_un_nombre();
     if (Nombre_saisi < 16) {
         Emplacement_Cible = Nombre_saisi; PTR = (Nombre_saisi - 1) * 67; Traiter_le_Deplacement();}}
  if (!Fuite) Retour_au_menu_de_BASE_avec_affichage_Utilisation_EEPROM();
  Revenir_au_MENU_de_BASE();}

void AFF_Ligne_Vide() {
  Serial.println(); Serial.println(); Serial.print(F("  Appel de Tr")); Serial.print(Transition); Serial.print('-');
  Affiche_PION(); Serial.println(F(" qui est une ligne VIDE !")); Serial.print(F("  Sortie de l'Ex")); E_accent();
  Serial.print(F("cution.")); BIP(); delay(300); BIP(); delay(300); BIP(); BIP(); delay(300); BIP();}

void Liste_action_Ecriture() {
  if (Etat == 0) Serial.print(F("    -------")); else Serial.print(F("    Ecrit "));
  if (Etat == 1) Serial.print('B'); if (Etat == 2) Serial.print('0'); if (Etat == 4) Serial.print('1');}

void Traiter_ECRITURE() {
  Etat = OCTET1 & 7;
  if (Etat == 1) {BARILLET[POS_TETE] = 2; Duree_du_RUN = Duree_du_RUN + 1.8;
                  Duree_pour_un_cycle = Duree_pour_un_cycle + 1.8;}  // "B".
  if (Etat == 2) {BARILLET[POS_TETE] = 0; Duree_du_RUN = Duree_du_RUN + 1.65;
                  Duree_pour_un_cycle = Duree_pour_un_cycle + 1.65;} // "0".
  if (Etat == 4) {BARILLET[POS_TETE] = 1; Duree_du_RUN = Duree_du_RUN + 1.4;
                  Duree_pour_un_cycle = Duree_pour_un_cycle + 1.4;}  // "1".
  if (Lister_a_chaque_cycle) Liste_action_Ecriture();}
  
void Affiche_la_rotation() {
  if (Etat == 0) Serial.print(F("  ----------")); else Serial.print(F("  Rotation "));
  if (Etat == 8) Serial.print('G'); if (Etat == 16) Serial.print('D');}

void Rotation_du_plateau_a_GAUCHE() {
  byte BIT; BIT = BARILLET[0];
  for (byte I = 0; I < 55; I++) BARILLET[I] = BARILLET[I+1]; BARILLET[55] = BIT; BIT = MODELE_des_GRADUATIONS[0];
  for (byte I = 0; I < 55; I++) MODELE_des_GRADUATIONS[I] = MODELE_des_GRADUATIONS[I+1];
       MODELE_des_GRADUATIONS[55] = BIT; if (ORIGINE > 0) ORIGINE--; else  ORIGINE = 55;}

void Rotation_du_plateau_a_DROITE() {
  byte BIT; BIT = BARILLET[55];
  for (byte I = 54; I > 0; I--) BARILLET[I+1] = BARILLET[I]; BARILLET[1] = BARILLET[0]; BARILLET[0] = BIT;
  BIT = MODELE_des_GRADUATIONS[55];
  for (byte I = 54; I > 0; I--) MODELE_des_GRADUATIONS[I+1] = MODELE_des_GRADUATIONS[I];
  MODELE_des_GRADUATIONS[1] = MODELE_des_GRADUATIONS[0]; MODELE_des_GRADUATIONS[0] = BIT;
  ORIGINE++; if (ORIGINE == 56) ORIGINE = 0;}
  
void Traiter_la_ROTATION() {
  Etat = OCTET1 & 24;
  if (Lister_a_chaque_cycle) Affiche_la_rotation();
  //------------------ Affichage pour des MOUVEMENTS de type Machine ----------------
  // ATTENTION : Comme ici c'est la tête de L/E qui se déplace, G et D ont un effet inverse
  //             à celui sur la machine matérielle pour obtenir le même effet visuel.              
       if (Etat == 8)  Rotation_du_plateau_a_GAUCHE();
       if (Etat == 16) Rotation_du_plateau_a_DROITE();
  //---------------------------------------------------------------------------------
  Duree_du_RUN = Duree_du_RUN + 0.27; Duree_pour_un_cycle = Duree_pour_un_cycle + 0.27;} 

void Affiche_la_Tr_sui_de_Pos_Tete() {
  if (Lister_a_chaque_cycle){
     if ((OCTET2 == 0) || (Transition == (Nb_TRN + 1))) Serial.print(F("    --------------"));
     else {Serial.print(F("    Tr demand")); E_accent(); Serial.print(F("e "));
           if (Transition < 10) Serial.print('0'); Serial.print(Transition);} 
     Serial.print(F("    (Nb cycles = ")); Serial.print(Nb_Cycles_Horloge); Serial.print(')');}}

void AFF_Tete_et_Transition() {
  if (Lister_a_chaque_cycle) Affiche_la_Tr_sui_de_Pos_Tete();}

void INITIALISER_le_plateau_de_la_machine() {
  POS_TETE = Pos_Tete; ORIGINE = Origine; for (byte I = 0; I < 56; I++) BARILLET[I] = BARILLET_Initial[I];
  for (byte I=0; I < 56; I++) MODELE_des_GRADUATIONS[I] = Modele_des_graduations[I];}

void Liste_le_debut_ligne_de_PGM() {
  if (Lister_a_chaque_cycle){
     Serial.println(); Serial.print(F("  Lecture ")); Affiche_PION(); Serial.print(F(" : Tr"));
     if(Transition < 10) Serial.print('0'); Serial.print(Transition); Serial.print('-'); Affiche_PION();}}

void Traiter_la_PAUSE() {
  Serial.println(); Serial.println(); Deux_SPC(); Serial.print(Nb_Cycles_Horloge);
  Serial.println(F(" cycles d'HORLOGE :")); Serial.print(F("  PAUSE : Valider pour reprendre l'Ex")); E_accent();
  Serial.print(F("cution.")); Caractere = Serial.read(); Accepte_CR_seul = true; Retour_MNU = false; Commande = false;
  Aff_Barillet_Initial = false; Pos_Tete = POS_TETE; Afficher_le_BARILLET(); Attendre_une_chaine_sur_RX(); 
  if (Chaine_Memorisee[0] == 'P') {Surveille = false; PAS_a_PAS = true; Lister_a_chaque_cycle = true;}}

void Marque_la_ligne_de_programme_comme_Utilisee() {
   byte Ligne = (Transition - 1) * 3; if (PION == 0) Ligne++; if (PION == 2) Ligne = Ligne + 2;
   Utilisee[Ligne] = true;}

void Verifie_si_une_ligne_NON_vide_est_Inutilisee() {
  boolean Ligne_Non_vide_Inutilisee = false; Pointeur = 0; boolean Vide; byte Cible; byte MAX = 33; if (Machine_ETENDUE) MAX = 60;
  for (byte I = 0; I < MAX; I++) {
       OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur]; Pointeur++;
       Vide = ((OCTET1 == 0) && (OCTET2 == 0));
       if (!Vide && !Utilisee[I]) {Ligne_Non_vide_Inutilisee = true; Cible = I + 1 ;}}
  if (Ligne_Non_vide_Inutilisee) {   
      BIP(); Serial.println(); Serial.print(F("REMARQUE : L'instruction "));  Serial.print(Cible);
      Serial.print(F(" NON VIDE n'est pas utilis")); E_accent(); Serial.println(F("e."));}}

void Calcule_le_temps_pour_transiter() {
  byte Depart = Transition; byte Arrivee = OCTET2; byte NB_Transitions;
  if (Arrivee > Depart) NB_Transitions = Arrivee - Depart; else NB_Transitions = (11 - Depart) + Arrivee;
  Duree_du_RUN = Duree_du_RUN + (0.286 * NB_Transitions);
  Duree_pour_un_cycle = Duree_pour_un_cycle + (0.286 * NB_Transitions);}

void Afficher_le_temps_reel_pour_executer() {
  float Temps_qui_reste, Quotient; Temps_qui_reste = Duree_du_RUN;
  Serial.print(Duree_du_RUN,0); Serial.println(F(" Secondes.")); Deux_SPC(); Serial.print(F("Soit : "));
  //---------------------- Déterminer si dépasse un AN soit 31536000 s ---------------------- 
  if (Temps_qui_reste > 31536000) {
     Quotient = Temps_qui_reste / 31536000; Serial.print(floor(Quotient),0); Serial.print(F(" ANNEES "));
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 31536000);}
  //----------------------- Déterminer si dépasse un JOUR soit 86400 s ----------------------
  if (Temps_qui_reste > 86400) {
     Quotient = Temps_qui_reste / 86400; Serial.print(Quotient,0); Serial.print(F(" JOURS "));
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 86400);}
  //---------------------- Déterminer si dépasse une HEURE soit 3600 s ----------------------
  if (Temps_qui_reste > 3600) {
     Quotient = Temps_qui_reste / 3600; Serial.print(Quotient,0); Serial.print(F(" HEURES "));
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 3600);}
  //----------------------- Déterminer si dépasse une minute soit 60 s ----------------------
  if (Temps_qui_reste > 60) {
     Quotient = Temps_qui_reste / 60; Serial.print(Quotient,0); Serial.print(F(" Minutes "));
     Temps_qui_reste = Temps_qui_reste - (floor(Quotient) * 60);}
  //------------------------------- Afficher le reste des secondes --------------------------
  if (Duree_du_RUN > 60) {Serial.print(Temps_qui_reste,0); Serial.println(F(" Secondes."));}}

void Previent_du_mode_PAS_a_PAS() {
  if (PAS_a_PAS) {
      Serial.println(); SPC(); N_etoiles(24); Retour_cadre(); 
      Serial.print(F("   MODE PAS a PAS.    ")); Retour_cadre();
      Serial.println(F(">>> ABANDON avec 'P'. *")); SPC(); N_etoiles(25); Serial.println();}}

void Affiche_les_caracteres_valides() {
  Serial.print(F("Caract")); Serial.print(char(232)); Serial.print(F("res possibles : ?, G, L, P, Q, R, T et V."));}

void Texte_Caractere() {
  if (Caractere != 0) {
      Serial.println(); Serial.print(F("  Caract")); Serial.print(char(232)); Serial.print(F("re "));
      Serial.print(char(91)); Serial.print(char(Caractere)); Serial.print(char(93));}}
  
void Caractere_incorrect_durant_le_RUN() {
  BIP(); Serial.print(F(" NON valide."));}

void Traite_les_Options_durant_le_RUN() {
  if (Caractere != '-') Texte_Caractere(); // Affiche le caractère frappé durant le RUN.
  if (Caractere == ',') Caractere = '?';
  if ((!Premier_O) && (Caractere == 79)) Caractere = '+'; // Caractère quelconque pour déclencher un BIP.
  switch (Caractere) {
     //--------- Ne pas faire BIT si validation à vide" ---------
     case  79 : {Premier_O = false; break;}
     case  13 : {break;}
     case  45 : {break;} // Pour ne pas faire BIP sur le caractère de prise en compte.
     case   0 : {break;}
     //----------------------------------------------------------
     case '?' : {Serial.println(); Deux_SPC(); Affiche_les_caracteres_valides(); Serial.println(); break;}
     case 'G' : {Format_Graphique = !Format_Graphique; break;}
     case 'L' : {Lister_a_chaque_cycle = !Lister_a_chaque_cycle; break;}
     case 'P' : {PAS_a_PAS = !PAS_a_PAS;
                 if (PAS_a_PAS) {Surveille = false; Lister_a_chaque_cycle = true; Commande = false;}
                 else Commande = true; break;}
     case 'Q' : {FINI = true; break;}
     case 'R' : {Ralenti = !Ralenti; if (Ralenti) {Temps_reel = false; PAS_a_PAS = false;} break;} 
     case 'T' : {Temps_reel = !Temps_reel; if (Temps_reel) Ralenti = false; break;}
     case 'V' : {Visualiser_Barillet = !Visualiser_Barillet; break;}
     default  : Caractere_incorrect_durant_le_RUN();}
  Caractere = '-';} // Caractere quelconque pour prise en compte.
  
void Traiter_le_programme() {
  char Inutile;
  FINI = false; // ATTENTION : Laisser cet identifiant en variable locale.
  Compteur_de_PAUSE = 0;         // Ci-dessous 2,53 Sec représente Un tour de la came d'HORLOGE et une LECTURE.
  if (PAS_a_PAS) {Aff_Barillet_Initial = false; Afficher_le_BARILLET(); while (!Serial.available()) {};
         Caractere = Serial.read();}
  while (!FINI) { Nb_Cycles_Horloge++; // Ici commence un cycle d'HORLOGE.
         Compteur_de_PAUSE++; Duree_du_RUN = Duree_du_RUN + 2.53; Duree_pour_un_cycle = 2.53;
         if (Serial.available()) {Caractere = Serial.read();
         if ((Caractere > 96) && (Caractere < 123)) Caractere = Caractere & 223; // 223 pour annuler le BIT 5.
         // Ci-dessus : Annuler le BIT 5 consiste à forcer MAJUSCULE.
         Inutile = Serial.read();} // Purger le Tampon Série du CR.
         //------------------ Traite les caractère frappés durant RUN --------------------
         Traite_les_Options_durant_le_RUN(); // Traite les booléens de ?, G, L, P, Q, R, T, V.
         //------------ Si mode Surveillé fait croire à la frappe d'une PAUSE ------------
         if ((Surveille) && (NB_cycles_entre_PAUSES == Compteur_de_PAUSE)) {
            Compteur_de_PAUSE = 0; Caractere = 'P';}
         //------- PAUSE manuelle si caractère 'P' et abandon du Temps réel si 'T' -------
         if (Caractere == 'P') Traiter_la_PAUSE();
         digitalWrite(LED_bleue, HIGH); delayMicroseconds(50); digitalWrite(LED_bleue, LOW);
         //------------ LECTURE de l'état du PION situé sous la Tête de L/E. -------------
         PION = BARILLET[POS_TETE];
         //-------------------- Liste le traitement de l'instruction ---------------------
         if (Lister_a_chaque_cycle) Liste_le_debut_ligne_de_PGM();
         //-------------------- Pointer l'instruction dans la MATRICE --------------------
         Pointeur = (Transition - 1) * 6;  Marque_la_ligne_de_programme_comme_Utilisee();
         if (PION == 0) Pointeur = Pointeur + 2; if (PION == 2) Pointeur = Pointeur + 4;
         //------------------------ Charge l'instruction pointée -------------------------
         OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur];
         //------------------- Vérifier si l'instruction n'est pas vide ------------------
         if ((OCTET1 == 0) && (OCTET2 == 0)) {BIP(); AFF_Ligne_Vide(); FINI = true;}
         if (!FINI) {
              Traiter_ECRITURE(); Traiter_la_ROTATION();
              //------------------------- Traiter la TRANSITION --------------------------
              if (OCTET2 > 0) {if (OCTET2 == (Nb_TRN + 1)) {FINI = true; Surveille = false;} 
                               else {Calcule_le_temps_pour_transiter(); Transition = OCTET2;}}
              AFF_Tete_et_Transition();
              if (Temps_reel) {digitalWrite(LED_bleue, HIGH); delay((int(Duree_pour_un_cycle) * 1000));
                               digitalWrite(LED_bleue, LOW);}
         if (BORNE > 0) FINI = (Nb_Cycles_Horloge == BORNE);
         if (Chaine_Memorisee[0] == 'Q') FINI = true;
         if (Visualiser_Barillet) {Aff_Barillet_Initial = false; Afficher_le_BARILLET();}
         if (Ralenti) {for (byte I = 0; I < 12; I++) {
                            digitalWrite(LED_bleue, HIGH); delay(16); digitalWrite(LED_bleue, LOW); delay(25);}}
         if (PAS_a_PAS) {
             Retour_MNU = false; Accepte_CR_seul = true; Attendre_une_chaine_sur_RX();
             Caractere = Chaine_Memorisee[0]; Traite_les_Options_durant_le_RUN();}}}} // Traite G, L, P, Q, R, T, V.

void Tester_la_presence_de_F() {
   if (!Ignorer_non_FIN) {
      F_existe = false;
      for(byte Pointeur = 1; Pointeur < (Etendue + 1); Pointeur = Pointeur + 2) {
          if (PROGRAMME[Pointeur] == (Nb_TRN + 1)) F_existe = true;}
      if (!F_existe) Traite_erreur(12);}}

void RUN() {// Ci-dessous : 60 pour compatibilité avec PGM standard ou ÉTENDU.
  Transition = 1; for (byte I = 0; I < 60; I++) Utilisee[I] = false; Pos_Tete_Initiale = Pos_Tete; Premier_O = true;
  // NOTE : C'est uniquement dans Executer_le_PGM que peut se faire la vérification d'une ligne vide.
  //        quand elle est invoquée durant le déroulement du programme.
  if (Num_PGM == 0) Traite_erreur(4);
  else {Duree_du_RUN = 0;
        //---------------------- Vérifie si le PGM contient au moins un "F" --------------------
        Tester_la_presence_de_F(); Caractere = 'O'; // Pour entrer dans le if si F_xiste = true.
        if ((!F_existe) && (!Ignorer_non_FIN)) {Serial.println(); Serial.print(F("  On continue ")); Appel_a_OUI_NON();}
        if (Caractere == 'O') {digitalWrite(LED_Arduino,LOW);
            //------------- Prévenir du temps de réaction de T, Q, L si Temps_reel) ------------
            if (Temps_reel) {
               Serial.println(); Serial.print(F("  ATTENTION Temps r")); E_accent();
               Serial.print(F("el : Environ 4S de d")); E_accent(); Serial.print(F("lai de r")); E_accent();
               Serial.println(F("action au clavier."));}
            //---------------------------------------------------------------------------------- 
            Serial.println(); Serial.print(F("  Ex")); E_accent(); Serial.print(F("cution du Programme. ("));
            Affiche_les_caracteres_valides(); Serial.println(')');
            Previent_du_mode_PAS_a_PAS(); INITIALISER_le_plateau_de_la_machine(); Ralenti = false;
            //----------------------------------------------------------------------------------
            Debut_du_RUN = millis(); // Chronométrage pour la durée réelle d'exécution.
            Nb_Cycles_Horloge = 0; Retour_MNU = false; Traiter_le_programme();
            //---------------------- Afficher l'état Initial de la machine ---------------------
            Pos_Tete = Pos_Tete_Initiale; Aff_Barillet_Initial = true;
            Serial.println(); false; Serial.println(); Serial.print(F("  Etat initial de la Machine :"));
            Retour_MNU = false;  Afficher_le_BARILLET();
            //---------------------- Afficher l'état Final de de la machine --------------------
            Serial.println(); Serial.print(F("  Traitement en ")); Serial.print(Nb_Cycles_Horloge);
            Serial.println(F(" cycles d'HORLOGE.")); Serial.print(F("  Dur")); E_accent();
            Serial.print(F("e du traitement : ")); Serial.print((millis() - Debut_du_RUN)/1000,2);
            Serial.println(F(" secondes. (PAUSES comprises.) "));
            Serial.print(F("  Temps d'Ex")); E_accent(); Serial.print(F("cution pour la Machine mat"));
            E_accent(); Serial.print(F("rielle : ")); Afficher_le_temps_reel_pour_executer(); Serial.println();
            Serial.print(F("  Etat final de la Machine :")); Serial.println(); Aff_Barillet_Initial = false;
            Retour_MNU = true; Pos_Tete = POS_TETE; Pos_Tete = Pos_Tete_Initiale; Afficher_le_BARILLET();}
  if (!Ignorer_les_lignes_NON_utilisees) Verifie_si_une_ligne_NON_vide_est_Inutilisee();}}   

void Effacer_le_PGM() {
  BIP(); Serial.print(F("  Effacer le PGM")); Appel_a_OUI_NON();
  if (Caractere == 'O') {Effacer_le_programme(); Serial.println(); Serial.print(F("  Programme effac"));
                        E_accent(); Serial.print('.');} Revenir_au_MENU_de_BASE();}

void Convertir_le_caractere() {
  switch (Caractere) {
     case  '0' : {BIT = 0; break;}
     case  '1' : {BIT = 1; break;}
     default   : {BIT = 2;}}}

void Forcer_Etat_sur_un_BIT() {
  Serial.println(); Serial.print(F("  Position du BIT ? (>0 et <57) ")); Commande = false; Saisir_un_nombre();
  Pos_BIT = Nombre_saisi - 1;
  if ((Nombre_saisi > 0) && (Nombre_saisi < 57)) {
     Serial.println(); Serial.print(F("  Etat du BIT ?")); Attendre_une_chaine_sur_RX();
     Caractere = Chaine_Memorisee[0]; Convertir_le_caractere(); BARILLET_Initial[Pos_BIT] = BIT;
     Aff_Barillet_Initial = false; Retour_MNU = true; Aff_Barillet_Initial = true; Afficher_le_BARILLET();}
  else Revenir_au_MENU_de_BASE();}

void Effacer_le_BARILLET_Initial() {
     for (byte I = 0; I < 56; I++) BARILLET_Initial[I] = 2; // La valeur > 1 correspond à l'Etat "B".
     Pos_Tete = 0; Origine = 0;}

void Initialiser_le_BARILLET() {
  Commande = false; FINI = false;
  Serial.println(); Retour_MNU = false; Aff_Barillet_Initial = true; Afficher_le_BARILLET();
  BIP(); Serial.print(F("  Effacer tous les pions actuels")); Appel_a_OUI_NON();
  if (Caractere == 'O') Effacer_le_BARILLET_Initial(); Serial.println();
  Serial.print(F("  Position de l'ORIGINE")); Saisir_un_nombre();
  if (Nombre_saisi > 56) FINI = true; else Origine = Nombre_saisi - 1;
  if (!FINI) {
     Serial.print(F("  Position de la T")); Serial.print(char(234)); Serial.print(F("te de L/E"));
     Saisir_un_nombre(); Pos_Tete = Nombre_saisi - 1; 
     //------------------ Saisir la configuration des pions ------------------
     Serial.print(F("  Position du pion de GAUCHE")); Saisir_un_nombre(); Pos_BIT = Nombre_saisi - 1; 
     Serial.println(); Serial.print(F("  Tout caract")); Serial.print(char(232)); Serial.print(F("re diff"));
     E_accent(); Serial.print(F("rent de [1,0,B] sera forc")); E_accent(); SPC(); A_accent(); SPC();
     Affiche_cote_B_cote(); Serial.println('.'); Serial.print(F("  Configuration des donn")); E_accent();
     Serial.print(F("es")); Attendre_une_chaine_sur_RX(); LGR_traitee = Longueur_du_TEXTE; Index = 0;
     while (LGR_traitee > 0) { // Scanne caractère par caractère depuis "la gauche vers la "droite".
        Caractere = Chaine_Memorisee[Index]; Convertir_le_caractere();
        BARILLET_Initial[Pos_BIT] = BIT; LGR_traitee--; Index++; Pos_BIT++; if (Pos_BIT == 56) Pos_BIT = 0;}}
  Retour_MNU = false; Afficher_le_BARILLET();
  //-----------------------------------------------------------------------
  Retour_MNU = true; Revenir_au_MENU_de_BASE();}   

void Affiche_la_reference_du_PGM() {
  if (Num_PGM < 100) Serial.print('0'); if (Num_PGM < 10) Serial.print('0'); Serial.print(Num_PGM);
  Serial.print(' ');}

void Afficher_la_feuille_perforee() {
  byte Compteur_par_3 = 1; byte Compteur_des_TR = 1; Pointeur = 0;
  Serial.println(); Serial.println(F("  TRn-L  B01  GD  T1 T2 T3 T4 T5 T6 T7 T8 T9 10 11  F"));
  Serial.print(F("  ----------- Programme Utilisateur n")); Serial.print(char(186)); Affiche_la_reference_du_PGM();
  Serial.print(F("-----------"));
  for (byte Ligne = 1; Ligne < 34; Ligne++) {
       Serial.println(); Serial.print(F("  T")); if (Compteur_des_TR < 10) Serial.print('R');
       Serial.print(Compteur_des_TR); Serial.print('-');
       switch(Compteur_par_3) {
           case  1 : {Serial.print('1'); Compteur_par_3 = 2; break;}
           case  2 : {Serial.print('0'); Compteur_par_3 = 3; break;}
           case  3 : {Serial.print('B'); Compteur_par_3 = 1; Compteur_des_TR++; break;}}
       Serial.print(F("  ")); 
       //------------------ Tracer les perforations de la grille de programme ------------------
       OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur]; Pointeur++;
       //-------------------------- Décode l'ECRITURE -----------------------------
       if (OCTET1 & 1) Serial.print(char(149)); else Serial.print('o');
       if (OCTET1 & 2) Serial.print(char(149)); else Serial.print('o');
       if (OCTET1 & 4) Serial.print(char(149)); else Serial.print('o'); Serial.print(F("  "));
       //------------------------- Décode la ROTATION -----------------------------
       if (OCTET1 & 8) Serial.print(char(149)); else Serial.print('o');
       if (OCTET1 & 16) Serial.print(char(149)); else Serial.print('o'); Serial.print(F("   "));
       //----------------------- Décode les Transitions ---------------------------
       for (byte I = 1; I < 13; I++) {
           if (OCTET2 == I) Serial.print(char(149)); else Serial.print('o'); Serial.print(F("  "));}}
  Serial.println();}

void Afficher_le_nombre_de_trous() {
  byte Nb_lignes = 33; if (Machine_ETENDUE) Nb_lignes = 60;
  if (Num_PGM == 0) Traite_erreur(4);
  else {if (!Machine_ETENDUE) Afficher_la_feuille_perforee();
        byte Pointeur = 0; Nb_Trous = 0;
        for(byte Ligne = 0; Ligne < Nb_lignes; Ligne++) {
            OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur]; Pointeur++;
            if (OCTET1 & 7) Nb_Trous++; if (OCTET1 & 24) Nb_Trous++; if (OCTET2 > 0) Nb_Trous++;}
        Serial.println(); Serial.print(F("  Nombre de trous dans la feuille perfor")); E_accent();
        Serial.print(F("e : ")); Serial.print(Nb_Trous); Serial.println('.');}}

void Affiche_le_debut_de_la_ligne() {
  Serial.print(F("   Ligne ")); if (Ligne_du_PGM < 10) Serial.print('0');
  Serial.print(Ligne_du_PGM); Serial.print(F("   ")); Serial.print('T');
  if (Transition < 10) Serial.print('r'); Serial.print(Transition); Serial.print('-');
  Serial.print(Etat); Serial.print(F(" : "));}

void Lister_le_PGM() {
  byte NB_lignes; if (Machine_ETENDUE) NB_lignes = 60; else NB_lignes = 33;
  if (Num_PGM == 0) Traite_erreur(4);
  else {byte Pointeur = 0; Etat = '1'; Transition = 1; Ligne_du_PGM = 1;
        //------------------------ Afficher le cadre et le Num du PGM. -----------------
        Debut_cadre(17); Serial.print(F(" PROGRAMME ")); Affiche_la_reference_du_PGM(); Fin_cadre(17);
        //----------------------------- Liste les lignes du PGM. -----------------------
        for(byte Ligne = 0; Ligne < NB_lignes; Ligne++) {
            OCTET1 = PROGRAMME[Pointeur]; Pointeur++; OCTET2 = PROGRAMME[Pointeur]; Pointeur++;
            Serial.println(); Affiche_le_debut_de_la_ligne();
            //-------------------------- Décode l'ECRITURE -----------------------------
            if (OCTET1 & 1) Serial.print('B'); if (OCTET1 & 2) Serial.print('0');
            if (OCTET1 & 4) Serial.print('1'); if (!(OCTET1 & 7)) Serial.print('.'); Serial.print(' ');
            //------------------------- Décode la ROTATION -----------------------------
            if (OCTET1 & 8) Serial.print('G'); if (OCTET1 & 16) Serial.print('D');
            if (!(OCTET1 & 24)) Serial.print('.'); Serial.print(' ');
            //------------------------ Décode la TRANSITION ----------------------------
            if (OCTET2 != (Nb_TRN + 1)) {if (OCTET2 == 0) { Serial.print(F(".. "));}
                               else {if (OCTET2 < 10) Serial.print('0'); Serial.print(OCTET2); Serial.print(' ');}}
            else { Serial.print(F(".. "));}
            //----------------------- Décode le "F" éventuel ---------------------------
            if (OCTET2 == (Nb_TRN + 1)) Serial.print('F'); else Serial.print('.');    
        //------------------------------------------------------------------------------
        Passe_a_la_ligne_suivante(); Ligne_du_PGM++;}}
   Serial.println(); Revenir_au_MENU_de_BASE();}

void Debut_ligne() {Serial.println(); Deux_SPC(); Serial.print(Etoile); SPC();}

void Fin_de_ligne(byte Nb_espaces) {N_SPC(Nb_espaces); Serial.print(Etoile); Debut_ligne();}

void Longue_Separation() {Serial.println(); Deux_SPC(); N_etoiles(65); Debut_ligne();}

void Separation_menu_Editeur() {Serial.println(); Deux_SPC(); N_etoiles(38); Debut_ligne();}

void Pointe_la_ligne_precedente() {
  if (Ligne_du_PGM != 2 ) {
     Ligne_du_PGM = Ligne_du_PGM - 2; Pointeur = Pointeur - 2;
     switch (Etat) {
       case  '1' : {Etat = 'B';  Transition--; break;}
       case  '0' : {Etat = '1'; break;}
       case  'B' : Etat = '0';}}
  else Ligne_du_PGM--;}

void Afficher_les_commandes_Editeur() {
  Separation_menu_Editeur(); 
  Serial.print(F("  Commandes de l'EDITEUR de PGM.   *")); Separation_menu_Editeur();
  Serial.print(F("Ecriture PUIS Rotation en premier.")); Fin_de_ligne(1); Serial.print(F("Transition indiqu"));
  E_accent(); Serial.print(F("e par 'T'")); Fin_de_ligne(8); Serial.print(F("T et F s'excluent mutuellement."));
  Fin_de_ligne(4);
  Serial.print(F("Exemples : BDT3 ou GF ou 0T11 ...  *")); Separation_menu_Editeur();
  Serial.print(F("? ou Virgule : Rappel de ce menu.")); Fin_de_ligne(2);
  Serial.print(F("Q : Sortie de l'EDITEUR de PGM.")); Fin_de_ligne(4);
  Serial.print(F("Valide seul : Ligne suivante.")); Fin_de_ligne(6);
  Serial.print(F("< : Ligne pr")); E_accent(); Serial.print('c'); E_accent(); Serial.print(F("dente."));
  N_SPC(14); Serial.println('*'); Deux_SPC(); N_etoiles(38); Serial.println();
  Pointe_la_ligne_precedente(); Pointeur = Pointeur + 2; Ligne_du_PGM++;}

void Saisir_une_ligne_de_programme() {
  Affiche_le_debut_de_la_ligne(); Ligne_du_PGM++;
  Commande = false; Attendre_une_chaine_sur_RX();
  if (Chaine_Memorisee[0] == 0) Serial.println();
  if (Chaine_Memorisee[0] == '?') Afficher_les_commandes_Editeur();
  if (Chaine_Memorisee[0] == '<') Pointe_la_ligne_precedente();
  else {FINI = (Chaine_Memorisee[0] == 'Q');
        if ((Longueur_du_TEXTE == 0) && (Mode_modifier)) {
             FINI = true; PROGRAMME[Pointeur] = 0; Pointeur++; PROGRAMME[Pointeur] = 0;}
        if (Chaine_Memorisee[0] != '?'){  
           if (!FINI) {if (Longueur_du_TEXTE == 0) Pointeur = Pointeur + 2;
                       if ((Longueur_du_TEXTE > 0) && (!FINI)) {Analyse_syntaxique();
                          if (!Sauter) { //Sauvegarde l'instruction.
                             PROGRAMME[Pointeur] = OCTET1; Pointeur++; PROGRAMME[Pointeur] = OCTET2; Pointeur++;}}}}}}

void Modifier_le_PGM() {
  FINI = false; Commande = false; Mode_modifier = true;
  byte Nb_ligne_MAX = 33; byte Ligne = 1; byte Compteur = 0; byte  Decalage = 0;
  if (Machine_ETENDUE) {Nb_TRN = 20; Nb_ligne_MAX = 60;} else Nb_TRN = 11; 
  if (Num_PGM == 0) Traite_erreur(4);
  else {//-------------------- Déterminer le num de Tr et l'Etat --------------------
        Serial.println(); Serial.print(F("  Modifier la ligne ")); Saisir_un_nombre();
        if (Nombre_saisi > Nb_ligne_MAX) {Traite_erreur(13); FINI = true;}
        if (!FINI) {Transition = 1; Etat = '1'; Ligne_du_PGM = Nombre_saisi;
           while (Ligne < Nombre_saisi) {
                  Compteur++; if (Compteur == 3) {Compteur = 0; Transition++;} Ligne++; 
                  switch (Etat) {
                    case  '1' : {Etat = '0'; Decalage = 2; break;}
                    case  '0' : {Etat = 'B'; Decalage = 4; break;}
                    case  'B' : Etat = '1';  Decalage = 0;}}}}
        //---------------------------------------------------------------------------             
        if (!FINI) {Pointeur = ((Transition - 1) * 6) + Decalage; Saisir_une_ligne_de_programme();}
        Mode_modifier = false; Tester_la_presence_de_F(); Revenir_au_MENU_de_BASE();} 

void Pointer_le_PGM_suivant() {PTR = PTR + 67;}

void AFF_etat_BORNE() {
  if (BORNE > 0) {Serial.print(F("  Sortie de PGM ")); A_accent(); SPC(); Serial.print(BORNE);
      Serial.println(F(" cycles d'HORLOGE."));}
  else {Serial.print(F("  Borne de Sortie anticip")); E_accent(); Serial.println(F("e du PGM : NON."));}}

void Saisie_de_la_borne() {
  Serial.println(); Serial.print(F("  Nombre de cycles d'HORLOGE pour sortir du PGM ? (Valeur 1 pour fuir.)"));
  Commande = false; Saisir_un_nombre(); BORNE = Nombre_saisi;
  if (BORNE == 1) BORNE = 0;}

void Mode_surveille_le_deroulement() {
  Serial.println(); Serial.print(F("  Surveiller le d")); E_accent(); Serial.print(F("roulement du PGM "));
  Appel_a_OUI_NON();
  if (Caractere == 'O') {
      Serial.println(); Serial.print(F("  Nombre de cycles d'HORLOGE entre chaque PAUSE ? "));
      Commande = false; Saisir_un_nombre(); NB_cycles_entre_PAUSES = Nombre_saisi; Surveille = true;
      if (Surveille) {Lister_a_chaque_cycle = false; PAS_a_PAS = false; BORNE = 0;}}}

void Annule_un_chargement_sur_RESET() {EEPROM.write(1023, 255); EEPROM.write(1021, 255);}

void Charger_un_PGM_EEPROM_au_RESET() {
  FINI = false;
  Retour_MNU = false; Utilisation_EEPROM();
  //------------------------------- Saisir l'Emplacement désiré ------------------------------
  Serial.println(); Serial.print(F("  Emplacement ")); A_accent();
  Serial.print(F(" invoquer sur un RESET (N > 15 pour annuler.) ? "));
  Commande = false; Saisir_un_nombre();
  if (Nombre_saisi < 16) {
     //-------------------------- Vérifier le contenu de l'emplacement --------------------------
     PTR = (Nombre_saisi - 1) * 67; OCTET = EEPROM.read(PTR);
     //---------------------- Vérifie si ce n'est pas un programme ETENDU -----------------------
      if (OCTET == 254) {
         Serial.println(); Serial.println(F("  Remarque : Chargement du programme Etendu.")); BIP(); Serial.println();}
     //------------------ Vérifie si ce n'est pas une configuration de BARILLET -----------------
     if (OCTET == 255) {Serial.println(); Serial.println();
         Serial.println(F("  Config. de BARILLET.")); BIP(); FINI = true;}
     //----------------------- Vérifie si ce n'est pas un Emplacement vide ----------------------
     if (OCTET == 0) {BIP(); Serial.println(F("  Emplacement vide.")); FINI = true;}
     //------------------------------------------------------------------------------------------
     if (!FINI) {if (Nombre_saisi < 16) {EEPROM.write(1022, Nombre_saisi); EEPROM.write(1023, 1);}
         else {FINI = true; Annule_un_chargement_sur_RESET();}}
     //------------- Saisie pour charger éventuellement l'initialisation du BARILLET ------------
     if (!FINI) {
        Serial.print(F("  Charger aussi l'Etat du BARILLET "));
        Appel_a_OUI_NON(); if (Caractere == 'O') EEPROM.write(1021, 1);
                           else EEPROM.write(1021, 255);}
     else {Annule_un_chargement_sur_RESET();}}
  else {Annule_un_chargement_sur_RESET();}}

void AFF_OUI_NON (boolean OUI) {if (OUI) Serial.println(F(" : OUI.")); else Serial.println(F(" : NON."));}

void Fin_de_ligne_longue_separation(byte Nb_espaces) {N_SPC(Nb_espaces); Serial.print(Etoile); Longue_Separation();}

void Menu_des_OPTIONS() {
  Longue_Separation(); Serial.print(F("? : Rappel de ce menu des OPTIONS. (? ou Virgule.)            *"));
  Debut_ligne(); Serial.print(F("Q : Quitter le menu des OPTIONS et reveniir au MENU de BASE."));
  Fin_de_ligne(2); Serial.print(F("R : R"));
  E_accent(); Serial.print(F("sum")); E_accent(); Serial.print(F(" des OPTIONS.")); Fin_de_ligne_longue_separation(39);
  N_SPC(21); Serial.print(F("Options pour le RUN.")); Fin_de_ligne_longue_separation(21);
  Serial.print(F("A : Mode Aveugle. (Annule Listages, Surveillance, PAS ")); A_accent();
  Serial.print(F(" PAS.) *"));  Debut_ligne(); Serial.print(F("B : Borne. (Sortir du PGM ")); A_accent();
  Serial.print(F(" N cycles d'HORLOGE.)")); Fin_de_ligne(14); Serial.print(F("F : Ignorer ")); Serial.print(char(91));
  Serial.print(F("Pas de FIN dans le PGM")); Serial.print(char(93)); Serial.print('.'); Fin_de_ligne(25);
  Serial.print(F("G : Format d'affichage Graphique.")); Fin_de_ligne(29);
  Serial.print(F("L : Listage des instructions durant le RUN.")); Fin_de_ligne(19);
  Serial.print(F("P : PAS ")); A_accent(); Serial.print(F(" PAS durant le RUN.")); Fin_de_ligne(34);
  Serial.print(F("S : Surveille le d")); E_accent(); Serial.print(F("roulement du PGM.")); Fin_de_ligne(26);
  Serial.print(F("T : Ex")); E_accent(); Serial.print(F("cution en TEMPS r")); E_accent();
  Serial.print(F("el de la Machine mat")); E_accent(); Serial.print(F("rielle.         *")); Debut_ligne();
  Serial.print(F("V : Visualiser l'Affichage du Barillet.")); Fin_de_ligne_longue_separation(23);
  N_SPC(11); Serial.print(F("Options pour la R")); E_accent(); Serial.print(F("daction du PROGRAMME."));
  Fin_de_ligne_longue_separation(12);
  Serial.print(F("M : Modifier le PGM pour recherche standard de la donn")); E_accent();
  Serial.print(F("e.")); Fin_de_ligne(5); Serial.print(F("N : Ignorer les lignes NON utilis")); E_accent();
  Serial.print(F("es.")); Fin_de_ligne(25); Serial.print(F("O : Inverser les ORIENTATIONS de D et G dans le programme."));
  Fin_de_ligne_longue_separation(4); N_SPC(22); Serial.print(F("Options Diverses.")); Fin_de_ligne_longue_separation(23);
  Serial.print(F("C : Charger un PGM depuis l'EEPROM sur RESET.")); Fin_de_ligne(17);
  Serial.print(F("E : Inverser l'option Machine ETENDUE.)")); Fin_de_ligne(23);
  Serial.print(F("I : Initialiser tout le barillet avec des B,0 ou 1.")); Fin_de_ligne_longue_separation(11);
  N_SPC(13); Serial.print(F("Options r")); E_accent(); Serial.print(F("serv")); E_accent();
  Serial.print(F("es aux programmeurs.")); Fin_de_ligne_longue_separation(14);
  Serial.print(F("$ : Listage des codes ASCII.  ")); Fin_de_ligne(32);
  Serial.println(F("W : Affiche la Place disponible en PILE dynamique.            *")); Deux_SPC(); N_etoiles(65);
  Serial.println(); Serial.println();}
  
void Resumer_les_OPTIONS() {
  if (EEPROM.read(1023) != 255) {Serial.print(F("  Sur RESET charge le PGM d'Emplacement "));
      Serial.print(EEPROM.read(1022)); Serial.print('.');
      if (EEPROM.read(1023) == 254) {Serial.print(F(" Remarque : Chargement du programme Etendu."));  
      Serial.println();}}
  if (EEPROM.read(1021) != 255) {Serial.println(); Serial.print(F("  Charge ")); E_accent();
       Serial.println(F("galement une Configuration Barillet.")); Serial.println();}
  else Serial.println();
  if (Machine_ETENDUE) AFF_restrictions_du_mode_ETENDU(); else Serial.println(F("  Machine ETENDUE : NON."));  
  Serial.print(F("  Mode PAS ")); A_accent(); Serial.print(F(" PAS")); AFF_OUI_NON(PAS_a_PAS);
  Serial.print(F("  Lister ")); A_accent(); Serial.print(F(" chaque cycle"));
  AFF_OUI_NON(Lister_a_chaque_cycle); Serial.print(F("  Visualiser le BARILLET durant le RUN"));
  AFF_OUI_NON(Visualiser_Barillet); Serial.print(F("  Format d'affichage Graphique")); AFF_OUI_NON(Format_Graphique);
  Serial.print(F("  Ignorer les lignes NON utilis")); E_accent(); Serial.print(F("es"));
  AFF_OUI_NON(Ignorer_les_lignes_NON_utilisees);
  Serial.print(F("  F : Ignorer ")); Serial.print(char(91)); Serial.print(F("Pas de FIN dans le PGM"));
  Serial.print(char(93)); AFF_OUI_NON(Ignorer_non_FIN); AFF_etat_BORNE();
  if (Surveille) { Serial.print(F("  PAUSE tous les ")); Serial.print(NB_cycles_entre_PAUSES);
                   Serial.println(F(" cycles d'HORLOGE."));}
  else Serial.println(F("  Surveiller le PGM : NON."));
  Serial.print(F("  Temps Machine R")); E_accent(); Serial.print(F("el respect")); E_accent();
  AFF_OUI_NON(Temps_reel); Serial.println();}

void Booleens_en_mode_Aveugle() {
  Lister_a_chaque_cycle = false; NB_cycles_entre_PAUSES = 0; PAS_a_PAS = false; Surveille = false;
  Temps_reel = false; Visualiser_Barillet = false;}

void Passer_en_options_pour_AVEUGLE() {
  BIP(); Serial.println(); Serial.print(F("  Mode AVEUGLE")); Appel_a_OUI_NON();
  if (Caractere == 'O') {Booleens_en_mode_Aveugle();}}

void Ecrit_56_etats(byte BIT) {for (byte I = 0; I < 56; I++) BARILLET_Initial[I] = BIT;}

void Initialiser_tout_le_BARILLET() {
  Serial.println(); Serial.print(F("  > Initialiser tout le barillet avec (B,0,1 ou N) "));
  Accepte_CR_seul = true; Commande = false; Attendre_une_chaine_sur_RX();
  switch (Chaine_Memorisee[0]) {
     case  '1' : {Ecrit_56_etats(1); FINI = true; break;}
     case  '0' : {Ecrit_56_etats(0); FINI = true; break;}
     case  'B' : {Ecrit_56_etats(2); FINI = true; break;}
     default   : Caractere = '$';}
  if (Caractere != '$') {
      Serial.println(); Serial.print(F("  BARILLET initialis")); E_accent(); Serial.print(F(" avec des "));
      Serial.print(Chaine_Memorisee[0]); Serial.println('.');}}

void Verifier_place_possible() {
  Place_possible = true;
  if (Machine_ETENDUE) {
      OCTET1 = PROGRAMME[114]; OCTET2 = PROGRAMME[115]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[116]; OCTET2 = PROGRAMME[117]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
      OCTET1 = PROGRAMME[118]; OCTET2 = PROGRAMME[119]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}   
  else {OCTET1 = PROGRAMME[60]; OCTET2 = PROGRAMME[61]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[62]; OCTET2 = PROGRAMME[63]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;
        OCTET1 = PROGRAMME[64]; OCTET2 = PROGRAMME[65]; if((OCTET1 > 0) || (OCTET2 > 0)) Place_possible = false;}
  if (!Place_possible) {BIP();  Serial.println(); Serial.println(F("  Place insuffisante."));}}

void Modifier_le_PGM_pour_recherche_standard_de_la_donnee() {
  byte Debut_translation;
  if (Num_PGM == 0) Traite_erreur(4);
  else {if (Machine_ETENDUE) Debut_translation = 113; else Debut_translation = 59;
        BIP(); Serial.println(); Serial.print(F("  FORMATER le programme en standard")); Appel_a_OUI_NON();
        if (Caractere == 'O') { 
            Verifier_place_possible();
            if (Place_possible) {
                //-------------------- Si Machine ETENDUE préparer pour "F" -------------------
                if(Machine_ETENDUE)
                   for (byte I = 1; I < 120; I = I + 2) {if (PROGRAMME[I] == 12) PROGRAMME[I] = 21;}
                //----------------- Décaler d'une transition les instructions -----------------             
                    for (byte I = Debut_translation; I > 0; I--) PROGRAMME[I + 6] = PROGRAMME[I];
                    PROGRAMME[7] = PROGRAMME[1]; PROGRAMME[6] = PROGRAMME[0];            
                //--------- Remplacer les trois premières lignes par la vectorisation ---------
                    PROGRAMME[0] = 0; PROGRAMME[1] = 2; PROGRAMME[2] = 0; PROGRAMME[3] = 0;
                    PROGRAMME[4] = 8; PROGRAMME[5] = 0;
                //------------------ Recalculer les adresse des translations ------------------
                for (byte I = 7; I < Etendue; I = I + 2) {
                     OCTET = PROGRAMME[I]; if((OCTET > 0) && (OCTET != (Nb_TRN + 1)))
                                              {OCTET++; PROGRAMME[I] = OCTET;}}}}}}

void Inverser_les_Orientations_dans_le_PGM() {
  if (Num_PGM == 0) Traite_erreur(4);
  else {BIP(); Serial.println(); Serial.print(F("  Inverser G et D dans le programme")); Appel_a_OUI_NON();
        if (Caractere == 'O') { 
            for (byte I = 0; I < Etendue; I = I + 2) {
                 OCTET = PROGRAMME[I] & 24;
                 if(OCTET > 0) {
                     if (OCTET == 8) PROGRAMME[I] = PROGRAMME[I] + 8; else PROGRAMME[I] = PROGRAMME[I] - 8;}}}}}
                     
void Choisir_Mode_Machine_Etendue() {
  byte Nb_Etoiles = 64;
  if (!Machine_ETENDUE) { 
     BIP(); Serial.println(); Serial.print(F("  Passer en MACHINE ")); Serial.print(char(201));
     Serial.print(F("TENDUE")); Appel_a_OUI_NON();
     if (Caractere == 'O') {
        //---------------- Avertit si un programme est présent ----------------
        if (Num_PGM > 0) {
            BIP(); Serial.println(); Deux_SPC(); if(Num_PGM > 9) Nb_Etoiles++; if(Num_PGM > 99) Nb_Etoiles++;
            N_etoiles(Nb_Etoiles); Serial.println(); Serial.print(F("  * Le programme n")); Serial.print(char(176));
            Serial.print(Num_PGM); Serial.print(F(" est pr")); E_accent(); Serial.print(F("sent et transform"));
            E_accent(); Serial.println(F(" au format ETENDU. *")); Deux_SPC(); N_etoiles(Nb_Etoiles); Serial.println();
            for(byte Pointeur = 1; Pointeur < (67); Pointeur = Pointeur + 2) {
               if (PROGRAMME[Pointeur] == 12) PROGRAMME[Pointeur] = 21;}}
        //---------------------------------------------------------------------
        Machine_ETENDUE = true; Etendue = 120; Nb_TRN = 20; Num_PGM = 254;}}
  else {Machine_ETENDUE = false;  Serial.println(); Serial.print(F("  Mode ETENDU annul")); E_accent();
       Serial.println('.'); Etendue = 66; Nb_TRN = 11;}}

//============ Ci-dessous fonction pour afficher la place disponible ============
int SRAM_LIBRE() { // Fonction qui retourne la taille de SRAM disponible.
    extern int __heap_start, *__brkval; // Déclaration des deux pointeurs dédiés.
    byte BIDON; // Dernière variable allouée, donc occupe le "haut" du TAS.
    if (__brkval == 0) {return (int) &BIDON -(int) &__heap_start;}
    else {return (int) &BIDON -(int) __brkval;}}
//===============================================================================

void Place_disponible_sur_la_PILE() {// vérifier la non "collision de PILE".
  Serial.println(); Serial.print(F("  >> Il reste actuellement "));
  Serial.print(SRAM_LIBRE()); Serial.print(F(" octets dans la PILE de la m")); E_accent();
  Serial.print(F("moire dynamique. <<")); Serial.println(); Serial.println();}

void Lister_les_codes_ASCII() {// Routine de servitude pour le programmeur.
  byte Compteur = 0;
  Serial.println(); Serial.print(F("  Listage des codes ASCII.  "));
  Serial.println(); Serial.println(F("  ========================"));
  for (byte I=33; I<255; I++) { 
        Serial.print(F("  Caractere = ")); if (I < 100) Serial.print('0'); 
        Serial.print(I); Serial.print(F(" : ")); Serial.print(char(I));
        Compteur++; if (Compteur == 4) {Serial.println(); Compteur = 0;}}
  Serial.println(); Serial.println();}
  
void OPTIONS() {
  Menu_des_OPTIONS(); Accepte_CR_seul = true; Caractere = '+'; // Caractère quelconque sauf 'Q'.
  while (Caractere != 'Q') {
    Attendre_une_chaine_sur_RX(); Caractere = Chaine_Memorisee[0]; if (Caractere == '<') Caractere = '>';
    switch (Caractere) {
       case  '?' : {Menu_des_OPTIONS(); break;}
       case  '$' : {Lister_les_codes_ASCII(); break;} 
       case  'A' : {Options = false; Passer_en_options_pour_AVEUGLE(); Options = true; break;}
       case  'B' : {Options = false; Saisie_de_la_borne(); Options = true; break;}
       case  'C' : {Options = false; Charger_un_PGM_EEPROM_au_RESET(); Options = true; break;}
       case  'E' : {Options = false; Choisir_Mode_Machine_Etendue(); Options = true; break;}      
       case  'F' : {Ignorer_non_FIN = !Ignorer_non_FIN; break;}
       case  'G' : {Format_Graphique = !Format_Graphique; break;}
       case  'I' : {Options = false; Initialiser_tout_le_BARILLET(); Options = true; break;}
       case  'L' : {Lister_a_chaque_cycle = !Lister_a_chaque_cycle; Surveille = false; break;}
       case  'M' : {Options = false; Modifier_le_PGM_pour_recherche_standard_de_la_donnee(); Options = true; break;}
       case  'N' : {Ignorer_les_lignes_NON_utilisees = !Ignorer_les_lignes_NON_utilisees; break;}
       case  'O' : {Options = false; Inverser_les_Orientations_dans_le_PGM(); Options = true; break;}
       case  'P' : {PAS_a_PAS = !PAS_a_PAS; 
                    if (PAS_a_PAS) {Surveille = false; Lister_a_chaque_cycle = true; break;}}
       case  'Q' : {Commande = true; Options = false; break;}
       case  'R' : {break;} // ne rien faire si 'R' car affiche l'état après chaque commande d'option.
       case  'S' : { Options = false; Mode_surveille_le_deroulement(); Options = true; break;}
       case  'T' : {Temps_reel = !Temps_reel; Lister_a_chaque_cycle = true; break;}
       case  'V' : {Visualiser_Barillet = !Visualiser_Barillet; break;}
       case  'W' : {Place_disponible_sur_la_PILE(); break;}
       default  : {Traite_erreur(3);}}
    if ((Caractere != '$') && (Caractere != 'Q')) Resumer_les_OPTIONS();}
  Revenir_au_MENU_de_BASE();}   

void Purger_EEPROM() {
  BIP(); Serial.println(); Serial.print(F("  Purger l'EEPROM")); Appel_a_OUI_NON();
  if (Caractere == 'O') {
      BIP(); Serial.println(); Serial.print(F("  Effacer vraiment tous les Programmes")); Appel_a_OUI_NON();
      if (Caractere == 'O') {
          PTR = 0; for(byte I = 1; I < 16; I++) {EEPROM.write(PTR, 0); Pointer_le_PGM_suivant();}}
          EEPROM.write(1021, 255); EEPROM.write(1023, 255);}
  Revenir_au_MENU_de_BASE();}
  
void Saisir_un_Num_de_PGM() {
  Nombre_saisi = 254;
  while ((Nombre_saisi < 1) || (Nombre_saisi > 253)) { 
         Commande = false; Saisir_un_nombre(); Num_PGM = Nombre_saisi;
         if (Nombre_saisi > 253) {BIP(); Serial.println(); Serial.println(F("  Nb < 254 !"));}}}

void Changer_le_Num_du_Programme() {
  if (Machine_ETENDUE) {
      Num_PGM = 254; BIP(); Serial.println(); Serial.print(F("  Machine ETENDUE. (Reference forc"));
      E_accent(); Serial.print(F("e ")); A_accent(); Serial.print(F(" 254.)"));}
  else {Serial.println(); Serial.print(F("  Nouveau Num")); E_accent();
        Serial.print(F("ro pour le PGM  (Num < 254)")); Saisir_un_Num_de_PGM();}
  Revenir_au_MENU_de_BASE();}

void Ecrire_le_PGM() {
  FINI = false; Pointeur = 0; BIP(); Serial.print(F("  Effacer le programme actuel")); Appel_a_OUI_NON();
  if (Caractere == 'O') Effacer_le_programme();
  Etat = '1'; Transition = 1; Accepte_CR_seul = true; Ligne_du_PGM = 1;
  if (! Machine_ETENDUE) {
     Serial.print(F("  Num")); E_accent(); Serial.print(F("ro de r")); E_accent();
     Serial.print(F("ference du programme (Num < 254) ")); Saisir_un_Num_de_PGM();}
  //--------------------- Saisie des lignes de programme ---------------------        
  while (!FINI) {Saisir_une_ligne_de_programme();
                 if ((!Sauter) && (Chaine_Memorisee[0] != ' ') && (Chaine_Memorisee[0] != '<'))
                    Passe_a_la_ligne_suivante();
                 if (Transition == (Nb_TRN + 1)) FINI = true;} 
  //--------------------------------------------------------------------------
  Tester_la_presence_de_F();
  Revenir_au_MENU_de_BASE();}

void Ecrit_le_PGM_en_EEPROM(byte Nb) {
  EEPROM.write(PTR, Num_PGM); Pointeur = 0; PTR++;
  for (byte I = 1; I < Nb; I++) {
       OCTET = PROGRAMME[Pointeur]; EEPROM.write(PTR, OCTET); Pointeur++; PTR++;
       OCTET = PROGRAMME[Pointeur]; EEPROM.write(PTR, OCTET); Pointeur++; PTR++;}
  Fuite = false;}

void Prevenir_si_Emplacement_RESET() {
  Serial.println(); BIP(); Serial.print(F("  Emplacement utilis")); E_accent();
  Serial.println(F(" pour charger un PGM au RESET.")); Serial.print(F("  On persiste"));
  Appel_a_OUI_NON();}

void Prevenir_de_config_Barillet() {
    Serial.println(); Serial.println(); Serial.println(F("  Emplacement d'une Configuration de BARILLET."));
    BIP(); Serial.print(F("  On persiste")); Appel_a_OUI_NON();}
 
void SAV_PGM_11TR() {
  FINI = false; Fuite = true;
  Commande = false; Serial.println(); Serial.print(F("  Ecrire le programme en Emplacement num ? (Fuite si > 15;)"));
  Saisir_un_nombre();
  if (Nombre_saisi < 16) {
      PTR = (Nombre_saisi - 1) * 67;
      OCTET = EEPROM.read(PTR);
      //-------------- Vérifie si le programme n'est pas chargé sur RESET -------------
      if ((EEPROM.read(1022) == Nombre_saisi) && (EEPROM.read(1023) == 1)) {
         Prevenir_si_Emplacement_RESET(); if (Caractere == 'O') Ecrit_le_PGM_en_EEPROM(34); FINI = true;}
      if (!FINI) {
          //----------- Vérifie si ce n'est pas une configuration de BARILLET ---------
          if (OCTET == 255) {Prevenir_de_config_Barillet();
                             if (Caractere == 'O') { 
                                Ecrit_le_PGM_en_EEPROM(34); FINI = true; EEPROM.write(1021,255);}
                             else FINI = true;}
      //---------------------- Vérifie si l'emplacement est libre ---------------------
      if (!FINI) {if (OCTET != 0) {Serial.println(); Serial.print(F("  Emplacement occup")); E_accent();
                  Serial.print(F(". On persiste")); BIP(); Appel_a_OUI_NON();
                  if (Caractere == 'O') {FINI = true; Ecrit_le_PGM_en_EEPROM(34);}
                  else FINI = true;}
      else Ecrit_le_PGM_en_EEPROM(34);}}}
  if (!Fuite) Retour_au_menu_de_BASE_avec_affichage_Utilisation_EEPROM();
  else Revenir_au_MENU_de_BASE();}   

void SAV_PGM_20TR() {
  BIP(); Serial.println(); Serial.print(F("  Ecraser les deux emplacements 14 et 15 en EEPROM")); Appel_a_OUI_NON();
  if (Caractere == 'O') {
      PTR = 871; Ecrit_le_PGM_en_EEPROM(61);
      Retour_au_menu_de_BASE_avec_affichage_Utilisation_EEPROM();}}
  
void Sauver_le_PGM() {
  Retour_MNU = false; Utilisation_EEPROM();
  if (Num_PGM == 0) Traite_erreur(4); else {if (Machine_ETENDUE) SAV_PGM_20TR(); else SAV_PGM_11TR();}}

void Utilisation_EEPROM() {
  boolean Etendu = false; PTR = 0; 
  Serial.println(); Serial.println(F("     Programmes en EEPROM :"));
                    Serial.println(F("     ======================"));
  for(byte I = 1; I < 16; I++) {
      Serial.print(F("  Emplacement ")); if (I < 10) Serial.print('0'); Serial.print(I);  Fleche();
      OCTET = EEPROM.read(PTR);
      //-------------------------- Afficher la référence du PGM --------------------------
      if (OCTET == 255) Serial.println(F("Configuration BARILLET."));
      else {Serial.print(F("Programme "));
            if ((I == 14) && (OCTET == 254)) Etendu = true;
            if ((I == 15) && Etendu) Serial.print(F("Etendu."));
            else {if (OCTET == 0) Serial.print(F("###"));
                  else {if (OCTET < 100) Serial.print('0'); 
                        if (OCTET < 10) Serial.print('0'); Serial.print(OCTET);}
                  if ((EEPROM.read(1022) == I) && (EEPROM.read(1023) == 1)) {Serial.print(F(" (Charg"));
                       E_accent(); Serial.println(F(" au RESET.)"));}         
                  else Serial.println();}}
      Pointer_le_PGM_suivant();}
  Serial.println(); if (Retour_MNU) Revenir_au_MENU_de_BASE();}

void Retour_au_menu_de_BASE_avec_affichage_Utilisation_EEPROM() {
  Utilisation_EEPROM(); Revenir_au_MENU_de_BASE();}

void Gommer_un_PGM_EEPROM() {
  Fuite = true; FINI = false; Retour_MNU = false; Utilisation_EEPROM(); Serial.println();
  Serial.print(F("  Emplacement ")); A_accent(); Serial.print(F(" lib")); E_accent();
  Serial.print(F("rer en EEPROM (Fuite si > 15.) ")); Commande = false; Saisir_un_nombre();
  if (Nombre_saisi < 16) {PTR = (Nombre_saisi - 1) * 67;
  //---------------- Traiter le cas ou cet Emplacement est utilisé par le RESET ----------------
  if ((Nombre_saisi == EEPROM.read(1022)) && (EEPROM.read(1023) == 1)){Prevenir_si_Emplacement_RESET();
     if (Caractere == 'O') {EEPROM.write(1023, 255);  EEPROM.write(1021, 255); EEPROM.write(PTR,0); Fuite = false;}
     FINI = true;}
  //---------- Traiter le cas ou cet Emplacement contient une configuration de BARILLET --------
   if (!FINI) {if (EEPROM.read(PTR) == 255) {Prevenir_de_config_Barillet();
               if (Caractere == 'O') {
                  FINI = true; EEPROM.write(PTR,0); EEPROM.write(1021,255);} FINI = true; Fuite = false;}}
  //----------------- Traiter le cas d'emplacement du PGM ou d'emplacement vide ----------------
  if (!FINI) {EEPROM.write(PTR,0); Fuite = false;}} // Cas différent de "RESET" ou "BARILLET".
  if (!Fuite) Retour_au_menu_de_BASE_avec_affichage_Utilisation_EEPROM();
  else Revenir_au_MENU_de_BASE();}

void Repositionner_la_tete_de_L_E() {
  Aff_Barillet_Initial = true; Retour_MNU = false; Afficher_le_BARILLET();
  Serial.println(); Serial.print(F("  Position de la T")); Serial.print(char(234)); Serial.print(F("te de L/E"));
   Commande = false; Saisir_un_nombre(); Pos_Tete = Nombre_saisi - 1;
  Aff_Barillet_Initial = true; Afficher_le_BARILLET(); Revenir_au_MENU_de_BASE();}

void Repositionner_Origine() {
  Aff_Barillet_Initial = true; Afficher_le_BARILLET();
  Serial.println(); Serial.print(F("  Position de L'Origine")); Commande = false; Saisir_un_nombre();
  Origine = Nombre_saisi - 1; Aff_Barillet_Initial = true; Afficher_le_BARILLET(); Revenir_au_MENU_de_BASE();}

void Rotation_manuelle_du_Barillet_a_Droite() {
  Rotation_du_plateau_a_DROITE(); Aff_Barillet_Initial = false; Afficher_le_BARILLET();}

void Rotation_manuelle_du_Barillet_a_Gauche() {
  Rotation_du_plateau_a_GAUCHE(); Aff_Barillet_Initial = false; Afficher_le_BARILLET();}

void Restriction_EEPROM() {AFF_restrictions_du_mode_ETENDU(); BIP();}

void Recuperer_Etat_du_caroussel() {
  Serial.println(); Deux_SPC(); Recup_Etat(); Serial.print(F("actuel du carrousel"));
  Appel_a_OUI_NON();
  if (Caractere == 'O') {
      Pos_Tete = POS_TETE; Origine = ORIGINE; for (byte I = 0; I < 56; I++) BARILLET_Initial[I] = BARILLET[I];
      for (byte I=0; I < 56; I++) Modele_des_graduations[I] = MODELE_des_GRADUATIONS[I];}
  Aff_Barillet_Initial = true; Afficher_le_BARILLET(); Revenir_au_MENU_de_BASE();}
     
void Traite_la_ligne_du_MENU_de_BASE() { // Traite la ligne reçue sur USB.
  //------------------------------ Repérer les lignes de consignes -------------------------
  if (Longueur_du_TEXTE > 1) Traite_erreur(2);
  else {Caractere = Chaine_Memorisee[0];
       switch (Caractere) {
          case  '?' : {Serial.println(); MENU_de_BASE(); break;}
          case  '=' : {Repositionner_la_tete_de_L_E(); break;}
          case  '0' : {Repositionner_Origine(); break;}
          case  'A' : {Aff_Barillet_Initial = true; Afficher_le_BARILLET(); break;}
          case  'B' : {Sauvegarder_la_configiration_du_BARILLET(); break;}
          case  'C' : {Charger_le_PGM(); break;}
          case  'D' : {if (Machine_ETENDUE) Restriction_EEPROM(); else Deplacer_un_programme(); break;}
          case  'E' : {Ecrire_le_PGM(); break;}
          case  'F' : {Forcer_Etat_sur_un_BIT(); break;}
          case  'G' : {if (Machine_ETENDUE) Restriction_EEPROM(); else Gommer_un_PGM_EEPROM(); break;}
          case  'I' : {Initialiser_le_BARILLET(); break;}
          case  'K' : {Effacer_le_PGM(); break;}
          case  'L' : {Lister_le_PGM(); break;}
          case  'M' : {if (Num_PGM == 0) Traite_erreur(4); else Modifier_le_PGM(); break;}
          case  'N' : {Afficher_le_nombre_de_trous(); break;}
          case  'O' : {Commande = false; Options = true; OPTIONS(); break;}
          case  'P' : {if (Machine_ETENDUE) Restriction_EEPROM(); else Purger_EEPROM(); break;}
          case  'Q' : {Commande = false; Options = true; OPTIONS(); break;}
          case  'R' : {RUN(); break;}
          case  'S' : {Sauver_le_PGM(); break;}
          case  'T' : {Translater_la_configuration_du_BARILLET(); break;}
          case  'U' : {Utilisation_EEPROM(); break;}
          case  'V' : {Voir_le_resultat_de_RUN(); break;}
          case  'W' : {Liste_le_contenu_EEPROM(); break;}
          case  'Z' : {Recuperer_Etat_du_caroussel(); break;}
          case  'X' : {if (Num_PGM == 0) Traite_erreur(4); else Changer_le_Num_du_Programme(); break;}
          case  '>' : {Rotation_manuelle_du_Barillet_a_Droite(); break;}
          case  '<' : {Rotation_manuelle_du_Barillet_a_Gauche(); break;}
          default  : {if (Longueur_du_TEXTE > 0) Traite_erreur(3);}}}}

